require "behaviours/wander"
require "behaviours/faceentity"
require "behaviours/chaseandattack"
require "behaviours/panic"
require "behaviours/follow"
require "behaviours/attackwall"

require "behaviours/runaway"
require "behaviours/doaction"
require "behaviours/standstill"


local PrinceBrain = Class(Brain, function(self, inst)
    Brain._ctor(self, inst)
end)

--Prince will help chop, mine and fight.


--Prince should stay really close!
local MIN_FOLLOW_DIST = 0.5
local TARGET_FOLLOW_DIST = 1.25
local MAX_FOLLOW_DIST = 1.5

local START_FACE_DIST = 6
local KEEP_FACE_DIST = 8

local KEEP_WORKING_DIST = 3
local SEE_WORK_DIST = 5


local function FindLightSourceInPockets(inst)

    --if not inst.LightWatcher:IsInLight() then
    inst.components.container:Close()


    local item = inst.components.container:FindItem(function(lightsource)
        return lightsource.prefab == "torch"
        or (lightsource.prefab == "lantern" and lightsource.components.fueled and not lightsource.components.fueled:IsEmpty())
        or lightsource.prefab == "bottlelantern"
    end)

        --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
        if not inst:HasTag("armingprince") then
            inst:AddTag("armingprince")
        end

        if item then

            local itemremoval = inst.components.container:GetItemSlot(item)

                --Weirdly, tools don't label themselves as hand equippables...
                    local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                        if currenthand and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch"
                        or (currenthand and currenthand.prefab == "lantern" and currenthand.components.fueled and currenthand.components.fueled:IsEmpty()) then

                                inst.components.inventory:RemoveItem(currenthand)--:Remove() 
                                inst.components.container:GiveItem(currenthand)
                                inst.components.container:RemoveItemBySlot(itemremoval)--:Remove()  
                                --inst.components.inventory:GiveItem(item)
                                print("Prince is providing light!")
                                inst.components.inventory:Equip(item)
                        end

                        if not currenthand then
                            inst.components.container:RemoveItemBySlot(itemremoval)--:Remove()  
                            --inst.components.inventory:GiveItem(item)
                            print("Prince is providing light!")
                            inst.components.inventory:Equip(item)
                        end



        end

        if inst:HasTag("armingprince") then
            inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
        end


    --end


end


local function HandNotOccupied(inst)

    local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
    return (currenthand and not currenthand.components.waterproofer and not currenthand.components.fishingrod 
    and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch" and not currenthand.components.tool)
    or not currenthand    

end

local function StayDryHand(inst)

    local container = inst.components.container
    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)

        if item then
                
                --Weirdly, tools don't label themselves as hand equippables...
                if item.components.equippable and item.components.waterproofer and item.components.equippable.equipslot ~= EQUIPSLOTS.BODY and item.components.equippable.equipslot ~= EQUIPSLOTS.HEAD then
                    local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                        if currenthand and not currenthand.components.waterproofer
                        or (currenthand and currenthand.components.waterproofer and currenthand.components.waterproofer.effectiveness < item.components.waterproofer.effectiveness) then
                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currenthand)--:Remove()  
                            inst.components.container:GiveItem(currenthand)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay dry!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currenthand then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay dry!")
                            inst.components.inventory:Equip(item)
                        end
    
                end

        end

    end

end


local function StayDry(inst)

    local container = inst.components.container
    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)

        --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
        inst:AddTag("armingprince")

        if item then

                if item.components.equippable and item.components.waterproofer and item.components.equippable.equipslot == EQUIPSLOTS.BODY then
                    local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                        if (currentbody and currentbody.prefab ~= "armorsnurtleshell" and currentbody.prefab ~= "amulet" and not currentbody.components.waterproofer) 
                        or (currentbody and currentbody.components.waterproofer and currentbody.components.waterproofer.effectiveness < item.components.waterproofer.effectiveness) then
                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currentbody)--:Remove()  
                            inst.components.container:GiveItem(currentbody)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay dry!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currentbody then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay dry!")
                            inst.components.inventory:Equip(item)
                        end

                end



                if item.components.equippable and item.components.waterproofer and item.components.equippable.equipslot == EQUIPSLOTS.HEAD then
                    local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                        if (currenthead and not currenthead.components.waterproofer)
                        or (currenthead and currenthead.components.waterproofer and currenthead.components.waterproofer.effectiveness < item.components.waterproofer.effectiveness) then
                            if (currenthead.prefab ~= "icehat" and currenthead.prefab ~= "watermelonhat" and currenthead.prefab ~= "bushhat") then

                                inst.components.container:Close()
                                inst.components.inventory:RemoveItem(currenthead)--:Remove()  
                                inst.components.container:GiveItem(currenthead)
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                print("Prince is trying to stay dry!")
                                inst.components.inventory:Equip(item)
                            end
                        end

                        if not currenthead then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay dry!")
                            inst.components.inventory:Equip(item)
                        end
        

                end

        end

        --When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
        inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
            
    end

end

local function GetWarm(inst)

    local container = inst.components.container
    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)

        --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
        inst:AddTag("armingprince")

        if item then


                --Weirdly, Prince will equip an umbrella to stay warm, even though umbrellas are for summer shade...
                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot ~= EQUIPSLOTS.BODY and item.components.equippable.equipslot ~= EQUIPSLOTS.HEAD then
                    local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                        if currenthand and not currenthand.components.insulator and not currenthand.components.fishingrod and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch" and not currenthand.components.tool then
                                inst.components.container:Close()
                                inst.components.inventory:RemoveItem(currenthand)--:Remove()  
                                inst.components.container:GiveItem(currenthand)
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                print("Prince is trying to stay warm!")
                                inst.components.inventory:Equip(item)
                        end

                        if not currenthand then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end
    
                end


                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot == EQUIPSLOTS.BODY then
                    local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                        if (currentbody and currentbody.prefab ~= "armorsnurtleshell" and not currentbody.components.insulator) then
                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currentbody)--:Remove()  
                            inst.components.container:GiveItem(currentbody)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currentbody then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end
        
                end



                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot == EQUIPSLOTS.HEAD then
                    local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                        if (currenthead and not currenthead.components.insulator and currenthead.prefab ~= "bushhat") then

                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currenthead)--:Remove()  
                            inst.components.container:GiveItem(currenthead)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currenthead then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end
        

                end   

        end

        --When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
        inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
            
    end

end


local function GetWarmGiant(inst)

    local container = inst.components.container
    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)

        --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
        inst:AddTag("armingprince")

        if item then


                --Weirdly, Prince will equip an umbrella to stay warm, even though umbrellas are for summer shade...
                if item.components.equippable and item.components.insulator and (item.components.insulator.type == "WINTER" and item.components.insulator.type ~= "SUMMER") and item.components.equippable.equipslot ~= EQUIPSLOTS.BODY and item.components.equippable.equipslot ~= EQUIPSLOTS.HEAD then
                    local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                        if (currenthand and not currenthand.components.insulator and not currenthand.components.fishingrod and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch" and not currenthand.components.tool)
                        or (currenthand and currenthand.components.insulator and currenthand.components.insulator.type ~= "WINTER" and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch" and not currenthand.components.tool) then
                                inst.components.container:Close()
                                inst.components.inventory:RemoveItem(currenthand)--:Remove()  
                                inst.components.container:GiveItem(currenthand)
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                print("Prince is trying to stay warm!")
                                inst.components.inventory:Equip(item)
                        end

                        if not currenthand then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end
    
                end


                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot == EQUIPSLOTS.BODY and item.components.insulator.type == "WINTER" then
                    local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                        if (currentbody and currentbody.prefab ~= "armorsnurtleshell" and not currentbody.components.insulator)
                        or (currentbody and currentbody.components.insulator and currentbody.components.insulator.type ~= "WINTER") then
                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currentbody)--:Remove()  
                            inst.components.container:GiveItem(currentbody)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currentbody then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end
        
                end



                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot == EQUIPSLOTS.HEAD and item.components.insulator.type == "WINTER" then
                    local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                        if (currenthead and not currenthead.components.insulator and currenthead.prefab ~= "bushhat")
                        or (currenthead and currenthead.components.insulator and currenthead.components.insulator.type ~= "WINTER") then

                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currenthead)--:Remove()  
                            inst.components.container:GiveItem(currenthead)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currenthead then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay warm!")
                            inst.components.inventory:Equip(item)
                        end
        

                end

        end

        --When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
        inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)

    end

end


local function GetCool(inst)

    local container = inst.components.container
    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)

        --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
        inst:AddTag("armingprince")

        if item then


                --Weirdly, tools don't label themselves as hand equippables...
                if item.components.equippable and item.components.insulator and (item.components.insulator.type == "SUMMER" and item.components.insulator.type ~= "WINTER") and item.components.equippable.equipslot ~= EQUIPSLOTS.BODY and item.components.equippable.equipslot ~= EQUIPSLOTS.HEAD then
                    local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                        if (currenthand and not currenthand.components.insulator and not currenthand.components.fishingrod and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch" and not currenthand.components.tool)
                        or (currenthand and currenthand.components.insulator and currenthand.components.insulator.type ~= "SUMMER" and currenthand.prefab ~= "lantern" and currenthand.prefab ~= "bottlelantern" and currenthand.prefab ~= "torch" and not currenthand.components.tool) then
                                inst.components.container:Close()
                                inst.components.inventory:RemoveItem(currenthand)--:Remove()  
                                inst.components.container:GiveItem(currenthand)
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                print("Prince is trying to stay cool!")
                                inst.components.inventory:Equip(item)
                        end

                        if not currenthand then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay cool!")
                            inst.components.inventory:Equip(item)
                        end
    
                end


                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot == EQUIPSLOTS.BODY and item.components.insulator.type == "SUMMER" then
                    local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                        if (currentbody and currentbody.prefab ~= "armorsnurtleshell" and not currentbody.components.insulator and not currentbody.tool)
                        or (currentbody and currentbody.components.insulator and currentbody.components.insulator.type ~= "SUMMER" and not currentbody.tool) then
                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currentbody)--:Remove()  
                            inst.components.container:GiveItem(currentbody)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay cool!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currentbody then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay cool!")
                            inst.components.inventory:Equip(item)
                        end
        
                end



                if item.components.equippable and item.components.insulator and item.components.equippable.equipslot == EQUIPSLOTS.HEAD and item.components.insulator.type == "SUMMER" then
                    local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                        if (currenthead and currenthead.prefab ~= "bushhat" and not currenthead.components.insulator and not currenthead.tool)
                        or (currenthead and currenthead.components.insulator and currenthead.components.insulator.type ~= "SUMMER" and not currenthead.tool) then

                            inst.components.container:Close()
                            inst.components.inventory:RemoveItem(currenthead)--:Remove()  
                            inst.components.container:GiveItem(currenthead)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay cool!")
                            inst.components.inventory:Equip(item)
                        end

                        if not currenthead then
                            inst.components.container:Close()
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            print("Prince is trying to stay cool!")
                            inst.components.inventory:Equip(item)
                        end
        

                end


    

        end

        --When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
        inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
            
    end

end

--Prince will wear protective gear if there is an earthquake!
local function EarthquakeSuitUp(inst)

    inst.components.container:Close()
    local container = inst.components.container
    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)

        --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
        inst:AddTag("armingprince")

        if item then

                if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.BODY and item.components.armor then
                    local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                        if currentbody and not currentbody.components.armor then

                            print("Prince is protecting himself from an earthquake!")
                            inst.components.inventory:RemoveItem(currentbody)--:Remove()  
                            inst.components.container:GiveItem(currentbody)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            inst.components.inventory:Equip(item)
                        end

                        if not currentbody then
                            print("Prince is protecting himself from an earthquake!")
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            inst.components.inventory:Equip(item)
                        end
        
                end



                if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.HEAD and item.components.armor then
                    local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                        if (currenthead and not currenthead.components.armor) and (currenthead and currenthead.prefab ~= "minerhat") then

                            print("Prince is protecting himself from an earthquake!")
                            inst.components.inventory:RemoveItem(currenthead)--:Remove()  
                            inst.components.container:GiveItem(currenthead)
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            inst.components.inventory:Equip(item)
                        end

                        if not currenthead then
                            print("Prince is protecting himself from an earthquake!")
                            inst.components.container:RemoveItemBySlot(i)--:Remove()  
                            inst.components.inventory:Equip(item)
                        end
        

                end
    

        end

        --When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
        inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)

    end

end


local function EatFoodInInventory(inst)

    local food = inst.components.container:FindItem(function(item)
        return item.components.edible and inst.components.eater:CanEat(item) and item.components.edible.healthvalue >= 0 and item.components.edible.hungervalue > 0 and item.components.edible.foodtype ~= "SEEDS" --If it's not poisonous, not a seed, and is edible, Prince will eat it.
    end)
    if food and inst.components.hunger:GetPercent() < 0.8 and not inst.components.combat.target and not inst.sg:HasStateTag("busy") and not inst:HasTag("hiding") then
        if not inst:HasTag("hungryprince") then
            inst:AddTag("hungryprince") --This tag is so Prince can keep eating until he's full!
        end
        print("Prince is feeding himself!")
        return BufferedAction(inst, food, ACTIONS.EAT)
    end
end

local function IsNearLightning(inst)
    return FindEntity(inst, 2, function(target) return target.prefab == "lightning" end)
end


local function ShouldRunAway(inst, target)
    return inst.components.follower and not inst.components.follower.leader and not inst.sg:HasStateTag("hiding")
end


local function HasStateTags(inst, tags)
    for k,v in pairs(tags) do
        if inst.sg:HasStateTag(v) then
            return true
        end
    end
end

local function KeepWorkingAction(inst, actiontags)
    local currenttool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
    return inst.components.follower.leader and inst.components.follower.leader:GetDistanceSqToInst(inst) <= 3 and
    HasStateTags(inst.components.follower.leader, actiontags) and not inst.sg:HasStateTag("toolbroken")
end

local function StartWorkingCondition(inst, actiontags)
    local currenttool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
    return inst.components.follower.leader and HasStateTags(inst.components.follower.leader, actiontags) and not HasStateTags(inst, actiontags)
end

local function FindObjectToWorkAction(inst, action)
    if inst.sg:HasStateTag("working") and not inst.sg:HasStateTag("fishing") then
        return 
    end

    local notags = {"FX", "NOCLICK", "DECOR","INLIMBO"}
    local target = FindEntity(inst.components.follower.leader, SEE_WORK_DIST, function(item) return item.components.workable and item.components.workable.action == action end)
    if target then

        --Get Prince to equip the right tool for the job!
        local tool = inst.components.container:FindItem(function(equippable)
            return equippable.components.tool and equippable.components.tool:CanDoAction(action)
        end)
        local currenttool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)


        if not tool and not currenttool
        or (currenttool and currenttool.components.tool and not currenttool.components.tool:CanDoAction(action) and not tool)
        or (currenttool and currenttool.prefab == "lantern") or (currenttool and currenttool.prefab == "torch") then
            --print("Prince doesn't have the tool to help out!")
            return
        end


        if tool or (currenttool and currenttool.components.tool and currenttool.components.tool:CanDoAction(action)) then

            inst.components.container:Close()

            local actiontool = inst.components.container:GetItemSlot(tool)

            local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if (currenthand and not currenthand.components.tool) or (currenthand and currenthand.components.tool and not currenthand.components.tool:CanDoAction(action)) then

                inst.components.inventory:RemoveItem(currenthand)--:Remove()  
                inst.components.container:GiveItem(currenthand)
                inst.components.container:RemoveItemBySlot(actiontool)--:Remove()  
                inst.components.inventory:Equip(tool)
            end

            if not currenthand then
                inst.components.container:RemoveItemBySlot(actiontool)--:Remove()  
                inst.components.inventory:Equip(tool)
            end

            if not inst:HasTag("canwork") then
                inst:AddTag("canwork")
            end

            --print("Prince is going to help work!")
            return BufferedAction(inst, target, action)
        end

    end

end

local function FindObjectToFishAction(inst, action)
    if inst.sg:HasStateTag("working") and not inst.sg:HasStateTag("fishing") then
        return 
    end

    local notags = {"FX", "NOCLICK", "DECOR","INLIMBO"}
    local target = FindEntity(inst.components.follower.leader, SEE_WORK_DIST, function(item) return item.components.fishable and not item.components.flotsamfisher end)
    if target then

        --Get Prince to equip the right tool for the job!
        local tool = inst.components.container:FindItem(function(equippable)
            return equippable.components.fishingrod
        end)
        local currenttool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)


        if not tool and not currenttool
        or (currenttool and not currenttool.components.fishingrod and not tool)
        or (currenttool and currenttool.prefab == "torch") then
            --print("Prince doesn't have a fishing rod!")
            return
        end


        if tool or (currenttool and currenttool.components.fishingrod) then

            inst.components.container:Close()

            local actiontool = inst.components.container:GetItemSlot(tool)

            local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if (currenthand and not currenthand.components.fishingrod) then

                if currenthand.prefab == "lantern" then
                    inst.components.inventory:RemoveItem(currenthand)
                    inst.components.inventory:DropItem(currenthand, true, true) 
                else
                    inst.components.inventory:RemoveItem(currenthand)--:Remove()  
                    inst.components.container:GiveItem(currenthand)
                end
                inst.components.container:RemoveItemBySlot(actiontool)--:Remove()  
                inst.components.inventory:Equip(tool)
            end

            if not currenthand then
                inst.components.container:RemoveItemBySlot(actiontool)--:Remove()  
                inst.components.inventory:Equip(tool)
            end

            if not inst:HasTag("canwork") then
                inst:AddTag("canwork")
            end

            --print("Prince is going to help work!")
            return BufferedAction(inst, target, action, tool)
        end

    end

end


local function CurrentSpot(inst)
    if not inst.components.follower.leader and not inst.components.combat.target then
        return Vector3(inst.Transform:GetWorldPosition() )
    end
end

local function GetLeader(inst)
    return inst.components.follower.leader 
end

local function GetFaceTargetFn(inst)
    local target = GetClosestInstWithTag("player", inst, START_FACE_DIST)
    if target and not target:HasTag("notarget") then
        return target
    end
end

local function KeepFaceTargetFn(inst, target)
    return inst:IsNear(target, KEEP_FACE_DIST) and not target:HasTag("notarget")
end

function PrinceBrain:OnStart()
    local root = PriorityNode(
    {

        --It burns!
        ChattyNode(self.inst, STRINGS.CHARACTERS.PRINCE.PANICFIRE,
            WhileNode(function() return self.inst.components.health.takingfiredamage end, "OnFire",
                    Panic(self.inst)
                )
            ),


        --Shocking!
        WhileNode(function() return IsNearLightning(self.inst) end, "Struck by lightning!",
                DoAction(self.inst, function() return self.inst:PushEvent("lightningstrike") end)
            ),


        --Prince will try to keep himself dry during the rain!
        WhileNode(function() return IsDLCEnabled(REIGN_OF_GIANTS) and GetSeasonManager():GetSeason() ~= SEASONS.WINTER and GetSeasonManager():GetSeason() ~= SEASONS.SUMMER
        and ((GetSeasonManager():GetSeason() == SEASONS.SPRING and GetSeasonManager().percent_season > 0.15) or GetSeasonManager():GetSeason() ~= SEASONS.SPRING)
        and self.inst.components.moisture.moisture > 0
        and ((GetWorld().components.quaker and not GetWorld().components.quaker.quake) or not GetWorld().components.quaker)
        and not self.inst.components.container:IsOpen() and not self.inst.components.combat.target end, "Getting wet!",
                DoAction(self.inst, function() return StayDry(self.inst) end)
            ),
        --For lag reasons, Prince's dry hand equippable will be a different function!
        WhileNode(function() return IsDLCEnabled(REIGN_OF_GIANTS) and GetSeasonManager():GetSeason() ~= SEASONS.WINTER and GetSeasonManager():GetSeason() ~= SEASONS.SUMMER
        and ((GetSeasonManager():GetSeason() == SEASONS.SPRING and GetSeasonManager().percent_season > 0.15) or GetSeasonManager():GetSeason() ~= SEASONS.SPRING)
        and ((GetWorld().components.quaker and not GetWorld().components.quaker.quake) or not GetWorld().components.quaker)
        and self.inst.components.moisture.moisture > 0 and not self.inst.components.container:IsOpen() and not self.inst.components.combat.target 
        and HandNotOccupied(self.inst) end, "Getting wet!",
                DoAction(self.inst, function() return StayDryHand(self.inst) end)
            ),
        --Prince will find warm clothes in the cold!
        WhileNode(function() return not IsDLCEnabled(REIGN_OF_GIANTS) and GetSeasonManager():GetSeason() == SEASONS.WINTER
        and ((GetWorld().components.quaker and not GetWorld().components.quaker.quake) or not GetWorld().components.quaker)
        and not self.inst.components.container:IsOpen() and not self.inst.components.combat.target end, "Getting cold!",
                DoAction(self.inst, function() return GetWarm(self.inst) end)
            ),
        --Prince will find warm clothes in the cold! Reign of Giants version!
        WhileNode(function() return IsDLCEnabled(REIGN_OF_GIANTS) and (GetSeasonManager():GetSeason() == SEASONS.WINTER
        or (GetSeasonManager():GetSeason() == SEASONS.SPRING and GetSeasonManager().percent_season <= 0.15))
        and ((GetWorld().components.quaker and not GetWorld().components.quaker.quake) or not GetWorld().components.quaker)
        and not self.inst.components.container:IsOpen() and not self.inst.components.combat.target end, "Getting cold!",
                DoAction(self.inst, function() return GetWarmGiant(self.inst) end)
            ),
        --Prince will find cool clothes in the heat!
        WhileNode(function() return IsDLCEnabled(REIGN_OF_GIANTS) and GetSeasonManager():GetSeason() == SEASONS.SUMMER
        and ((GetWorld().components.quaker and not GetWorld().components.quaker.quake) or not GetWorld().components.quaker)
        and not self.inst.components.container:IsOpen() and not self.inst.components.combat.target end, "Getting hot!",
                DoAction(self.inst, function() return GetCool(self.inst) end)
            ),


        --Prince will wear armor to protect himself when an earthquake starts!
        WhileNode(function() return GetWorld().components.quaker and GetWorld().components.quaker.quake end, "Earthquake!",
                DoAction(self.inst, function() return EarthquakeSuitUp(self.inst) end)
            ),

                  
        --Prince will help work on things!
        WhileNode(function() return StartWorkingCondition(self.inst, {"chopping", "prechop"}) and 
        KeepWorkingAction(self.inst, {"chopping", "prechop"}) end, "keep chopping",
            LoopNode{ 
                ChattyNode(self.inst, STRINGS.CHARACTERS.PRINCE.WORKASSIST,
                    DoAction(self.inst, function() return FindObjectToWorkAction(self.inst, ACTIONS.CHOP) end)
                    )
                }
            ),
        WhileNode(function() return StartWorkingCondition(self.inst, {"mining", "premine"}) and 
        KeepWorkingAction(self.inst, {"mining", "premine"}) end, "keep mining",    
            LoopNode{ 
                ChattyNode(self.inst, STRINGS.CHARACTERS.PRINCE.WORKASSIST,               
                    DoAction(self.inst, function() return FindObjectToWorkAction(self.inst, ACTIONS.MINE) end)
                    )
                }
            ),
        WhileNode(function() return StartWorkingCondition(self.inst, {"fishing"}) and 
        KeepWorkingAction(self.inst, {"fishing"}) end, "Start fishing!",    
            --LoopNode{ 
                --ChattyNode(self.inst, STRINGS.CHARACTERS.PRINCE.WORKASSIST,               
                    DoAction(self.inst, function() return FindObjectToFishAction(self.inst, ACTIONS.FISH) end)
                    --)
                --}
            ),


        --Prince will provide light if it's dark!
        IfNode(function() return not self.inst.LightWatcher:IsInLight() end, "It's dark!",
            DoAction(self.inst, function() return FindLightSourceInPockets(self.inst) end)
        ),


        --Prince will try to feed himself if he's hungry and has edible food!
        WhileNode(function() return (self.inst.components.hunger and self.inst.components.hunger:GetPercent() < TUNING.HUNGRY_THRESH) or self.inst:HasTag("hungryprince") end, "hungry",
        DoAction(self.inst, function() return EatFoodInInventory(self.inst, ACTIONS.EAT) end)
        ),


        --Prince should only fight things under the company of Princess!
        ChattyNode(self.inst, STRINGS.CHARACTERS.PRINCE.BATTLECRY,
            WhileNode(function() return self.inst.components.follower.leader end, "attack",
                    ChaseAndAttack(self.inst, 5)
                )
            ),
        ChattyNode(self.inst, STRINGS.CHARACTERS.PRINCE.PANICRETREAT,
            WhileNode( function() return self.inst.components.combat.target end, "Panic",
                RunAway(self.inst, function() return self.inst.components.combat.target end, 10, 15, function(target) return ShouldRunAway(self.inst, target) end)
                )
            ),


        WhileNode(function() return GetLeader(self.inst) end, "has leader",
            --Prince should stop following Princess if she gets too far!
            Follow(self.inst, function() return self.inst.components.follower and self.inst.components.follower.leader end, MIN_FOLLOW_DIST, TARGET_FOLLOW_DIST, MAX_FOLLOW_DIST),            
            FaceEntity(self.inst, GetFaceTargetFn, KeepFaceTargetFn )
            ),


        --If Prince isn't occupied, he'll stay put!
        WhileNode(function() return not self.inst.components.health.takingfiredamage and not self.inst.components.combat.target and not self.inst.components.follower.leader end, "Wait here",
            StandStill(self.inst)
            ),


    }, .25)
    
    self.bt = BT(self.inst, root)    
end

return PrinceBrain