local easing = require "easing"


local PrincePersister = Class(function(self, inst)
    self.inst = inst
    self.inst:StartUpdatingComponent(self)

    self.nearbyentities = {}
	--self.nearbyentitycount = 0

    --self.task = self.inst:DoPeriodicTask(1, function() self:CheckIfNearPrince(1) end)
end)


function PrincePersister:EntityGone(entity)

	for k,v in pairs(self.nearbyentities) do
		if v and entity == v then
			table.remove(self.nearbyentities, k)
			return
		end
	end

end


function PrincePersister:OnSave()
    if self.noserial then return end
    
    local refs = {}
    local data = {
		nearbyentities={},
	}

    for k,v in pairs(self.nearbyentities) do
		table.insert(data.nearbyentities, v.GUID)
		table.insert(refs, v.GUID)
    end

	
    return data, refs
end


function PrincePersister:OnProgress()
	self.noserial = true
end


function PrincePersister:LoadPostPass(newents, savedata)

    if savedata.nearbyentities then
        for k,v in pairs(savedata.nearbyentities) do
            local child = newents[v]
            if child and child.entity.components.prince_anti_persister and child.entity.components.prince_anti_persister.NearPrince then
				child = child.entity
				table.insert(self.nearbyentities, child)
				child:ListenForEvent( "onremove", function() self:EntityGone(child) end, child )	
            end
        end
    end
    
end


function PrincePersister:CheckIfNearPrince()

	local x,y,z = self.inst.Transform:GetWorldPosition() 
    local ents = TheSim:FindEntities(x,y,z, 50)
    for k,v in pairs(ents) do 

        if v 
        and v.prefab 
        and (v.components.health or v.components.pickable or v.components.inventoryitem or v.components.combat or v.Physics or v.Light)
        and (v.components.prince_anti_persister and not v.components.prince_anti_persister.NearPrince)
        and v ~= GetPlayer() 
        and v.prefab ~= "prince" 
        and v ~= self.inst
    	and not v:HasTag("ground") 
    	and not v:HasTag("minimap") 
        then
    		v:PushEvent("NearPrince")
    		table.insert(self.nearbyentities, v)
    	end

    end


    if (IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC)) then
        if self.inst.components.follower.leader then
            if (self.inst.components.follower.leader.components.mapwrapper and self.inst.components.follower.leader.components.mapwrapper.state == 2) 
            or (self.inst.components.follower.leader.components.mapwrapper and self.inst.components.follower.leader.components.mapwrapper.state == 3) then
                return
            end
        end
    end




    --If Princess wanders too far, Prince will lose track of her!
    local princess = FindEntity(self.inst, 10, function(ent)
        return ent:HasTag("liarprincess") or ent == GetPlayer()
    end)
    if not princess then
        if self.inst.components.follower and self.inst.components.follower.leader and not self.inst.sg:HasStateTag("dead") then
            --if strayprincess and strayprincess.is_teleporting ~= true and not strayprincess.components.health:IsDead() then
            print("Prince can't find his way!")
                if GetPlayer().prefab == "princess" then
                    local lostprince = {"Where did you go, Princess?", "I seem to have lost my way...", "Princess? Are you still there?", "Princess, I think I've lost you."}
                    self.inst.components.talker:Say(lostprince[math.random(1, #lostprince)])
                elseif GetPlayer().prefab ~= "princess" then
                    local lostprince = {"I seem to have lost my way...", "Where did you go?", "Hello? Are you still there?", "I think I've lost you."}
                    self.inst.components.talker:Say(lostprince[math.random(1, #lostprince)])
                end
            self.inst.components.follower:StopFollowing()
        end
    end




    --Get Prince to warp with Princess through the Wooden Thing!
    local princesstraveler = GetPlayer()
    if princesstraveler.is_teleporting == true then

        for k,v in pairs(princesstraveler.components.leader.followers) do
                if not k:HasTag("blindprince") then
                    k.components.follower:StopFollowing()
            end
        end

        if self.inst.components.follower and not self.inst.components.follower.leader then
            princesstraveler.components.leader:AddFollower(self.inst)
        end

            self.inst:DoTaskInTime(1, function()
                if not self.inst:HasTag("teleporting") then

                    local container = self.inst.components.container
                    for i = 1, container:GetNumSlots() do
                        local item = container:GetItemInSlot(i)

                        if item then

                            if item:HasTag("irreplaceable") and not item:HasTag("princecloak") then

                                self.inst.components.container:DropItem(item)

                            end

                        end

                    end


                    if SaveGameIndex:GetCurrentMode() == "adventure" then
                        self.inst.components.container:Close()
                        local handslot = self.inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                        if handslot then
                            self.inst.components.inventory:RemoveItem(handslot)--:Remove()  
                            self.inst.components.container:GiveItem(handslot)
                        end

                        local bodyslot = self.inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                        if bodyslot then
                            self.inst.components.inventory:RemoveItem(bodyslot)--:Remove()  
                            self.inst.components.container:GiveItem(bodyslot)
                        end

                        local headslot = self.inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                        if headslot then
                            self.inst.components.inventory:RemoveItem(headslot)--:Remove()  
                            self.inst.components.container:GiveItem(headslot)
                        end

                        self.inst.components.container:DropEverything()
                    end


                    self.inst.sg:GoToState("teleportato_teleport")
                end
            end)
            self.inst:DoTaskInTime(5, function()

                if not self.inst:HasTag("registered") then

                    SaveGameIndex:GetSaveFollowers(GetPlayer()) 
                    --Just in case!
                    self.inst:AddTag("registered")
                end

            end)

    end


end


function PrincePersister:CheckIfFarFromPrince()

    for k,v in pairs(self.nearbyentities) do 

        if v and v.components.prince_anti_persister and v.components.prince_anti_persister.NearPrince and v:IsValid() then
        	local dist = v:GetDistanceSqToInst(self.inst)
        	if dist > 50*50 then
    			v:PushEvent("FarFromPrince")
    			self:EntityGone(v)
    		end
    	end

    end

end


function PrincePersister:KeepPrinceAwake()

    if self.inst:IsAsleep() then
        
        --print("Come back to us Prince!")
    end

end


function PrincePersister:OnUpdate(dt)

	self:CheckIfNearPrince()	
	self:CheckIfFarFromPrince()	
    --self:KeepPrinceAwake()

end

return PrincePersister
