require "brains/princebrain"
require "stategraphs/SGprince"


--For Prince's worm attacks!
require("fa_periodicthreats")

require "widgets/badge"
local hungerbadge = require "widgets/hungerbadge"
local healthbadge = require "widgets/healthbadge"
local sanitybadge = require "widgets/sanitybadge"

local easing = require "easing"

--local Widget = require "widgets/widget"
--local EquipSlot = require "widgets/equipslot"

local assets = 
{
	Asset("ANIM", "anim/prince.zip"),
	Asset("ANIM", "anim/prince_happy.zip"),

	Asset("ANIM", "anim/swap_pickaxe.zip"),
	Asset("ANIM", "anim/swap_axe.zip"),

	Asset("ANIM", "anim/player_idles_shiver.zip"),
	Asset("ANIM", "anim/player_actions.zip"),
	Asset("ANIM", "anim/player_actions_axe.zip"),
	Asset("ANIM", "anim/player_actions_pickaxe.zip"),
	Asset("ANIM", "anim/player_actions_shovel.zip"),
	Asset("ANIM", "anim/player_actions_blowdart.zip"),
	Asset("ANIM", "anim/player_actions_eat.zip"),
	Asset("ANIM", "anim/player_actions_item.zip"),
	Asset("ANIM", "anim/player_cave_enter.zip"),
	Asset("ANIM", "anim/player_actions_uniqueitem.zip"),
	Asset("ANIM", "anim/book_uniqueitem_swap.zip"),
	Asset("ANIM", "anim/player_actions_bugnet.zip"),
	Asset("ANIM", "anim/player_actions_fishing.zip"),
	Asset("ANIM", "anim/player_actions_boomerang.zip"),
	Asset("ANIM", "anim/player_bush_hat.zip"),
	Asset("ANIM", "anim/player_attacks.zip"),
	Asset("ANIM", "anim/player_idles.zip"),
	Asset("ANIM", "anim/player_rebirth.zip"),
	Asset("ANIM", "anim/player_jump.zip"),
	Asset("ANIM", "anim/player_amulet_resurrect.zip"),
	Asset("ANIM", "anim/player_teleport.zip"),
	Asset("ANIM", "anim/wilson_fx.zip"),
	Asset("ANIM", "anim/player_one_man_band.zip"),
	Asset("ANIM", "anim/player_slurtle_armor.zip"),
	Asset("ANIM", "anim/player_staff.zip"),

	Asset("ANIM", "anim/shadow_hands.zip"),   

	Asset("ANIM", "anim/player_actions_unsaddle.zip"),

	Asset("ANIM", "anim/goo.zip"),             

	Asset("SOUND", "sound/sfx.fsb"),
	Asset("SOUND", "sound/wilson.fsb"),

	Asset("ANIM", "anim/fish01.zip"),   --These are used for the fishing animations.
	Asset("ANIM", "anim/eel01.zip"),

	--if IsDLCEnabled(PORKLAND_DLC) then
		--Asset("ANIM", "anim/player_idles_groggy.zip"),
		--Asset("ANIM", "anim/player_groggy.zip"),
		--Asset("ANIM", "anim/player_sneeze.zip"),
	--end
}

local prefabs = 
{

}

	--15 slots to simulate player slots!
	local slotpos = {

					--Left slots!                           --Right slots!
                    Vector3(-162, 130 +114, 0),             Vector3(-87, 130 +114, 0), --Slot 2
                    Vector3(-162, 65 +114, 0),              Vector3(-87, 65 +114, 0), --Slot 4
                    Vector3(-162, 114, 0),                  Vector3(-87, 114, 0), --Slot 6
                    Vector3(-162, -65 +114, 0),             Vector3(-87, -65 +114, 0), --Slot 8
                    Vector3(-162, -130 +114, 0),            Vector3(-87, -130 +114, 0), --Slot 10
                    Vector3(-162, -195 +114, 0),            Vector3(-87, -195 +114, 0), --Slot 12
                    Vector3(-162, -260 +114, 0),            Vector3(-87, -260 +114, 0), --Slot 14
                    Vector3(-162, -325 +114, 0),
                    

					--Leftmost 3 slots! The "equippables" slots! Slot number 1 is the hand slot, slot number 6 is the body slot, and slot number 11 is the head slot!
					--Vector3(-(180), 64+32+8+4, -(180)), --HAND SLOT!
					--Vector3(-(180), 32+4, -(180)), --BODY SLOT!
					--Vector3(-(180), -(32+4), -(180)), --HEAD SLOT!

					}


local function ondeath(inst)

	local item = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
	if item and item.prefab == "amulet" then
	inst:DoTaskInTime(5, function()
		inst.sg:GoToState("amulet_rebirth")
	end)
	elseif not (item or (item and item.prefab ~= "amulet")) then

		if IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
			inst.components.container:DropEverything()
		--A workaround to make *A* Prince follow Princess in a new world in saves without Shipwrecked or Hamlet enabled.
		elseif not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
			local player = GetPlayer()
			local resurrector = FindEntity(inst, 10000, function(ent)
                	return ent.components.resurrector and ent.components.resurrector:CanBeUsed()
            	end)
       		if player.prefab == "princess" and player.components.resurrectable:CanResurrect() and not resurrector then
        		--NEVERMIND. Timed to follow after Princess' response in her "Heartbreak" code.
        		inst:DoTaskInTime(0.2, function()
        			print("Prince has been slain here!")
        			inst.components.health:Respawn(20)
            		inst.components.health:SetPercent(0.5)
            		inst.components.health:SetInvincible(true)
        			inst.CaveVictim = true
            		inst:PushEvent("banish")
        			player.PrinceIsAlive = false
        			player.ReplacePrince = true
        		end)
        	end
        	inst.components.container:DropEverything()
		end

	end

end


local function ShouldAcceptItem(inst, item)

	if inst.components.health:IsDead() or inst.components.combat.target or inst.sg:HasStateTag("busy") or inst.sg:HasStateTag("fishing") then
		return false
	end

	if item.components.edible and inst.components.eater:CanEat(item) and item.prefab ~= "petals_evil" then
		return true
	end

	if item.components.equippable then
		if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.HEAD then
			return true
		end

		if item.components.equippable and item.components.equippable.equipslot ~= EQUIPSLOTS.HEAD and item.components.equippable.equipslot ~= EQUIPSLOTS.BODY then
			return true
		end

		if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.BODY then
			return true
		end
	end

	--Just so he can accept grass and twigs!
	if item.components.tradable and item.prefab ~= "petals_evil" then
		return true
	end

end


local function OnGetItemFromPlayer(inst, giver, item)

	if not inst:HasTag("armingprince") then
		inst:AddTag("armingprince")
	end

	--Prince loves flowers, and will "keep" petals instead of eating them!
	if item.prefab == "petals" then
		inst.components.sanity:DoDelta(5, false)

		--Petals are the only non-equippable allowed in Prince's inventory!
        --inst.components.inventory:GiveItem(item)

        inst.sg:GoToState("item_hat")

		if giver.prefab == "princess" and math.random() < 0.75 then
			local princesstalk = {"Here, Prince. It's a flower!", "I know how much you love flowers, Prince.", "I have some flowers for you, Prince."}
        	giver.components.talker:Say(princesstalk[math.random(1, #princesstalk)])

        	inst:DoTaskInTime(0.5, function()
				local princetalk = {"Thank you, Princess!", "It smells so nice...", "I hope to see it someday."}
        		inst.components.talker:Say(princetalk[math.random(1, #princetalk)])
        	end)

        end

        item:Remove()

	end

	print(item.prefab, inst.components.eater:CanEat(item))
	if inst.components.eater:CanEat(item) and item.prefab ~= "petals" then

		if giver.prefab == "princess" then
			--If the food isn't harmful in any way, Princess will feel better feeding Prince if he's not full!
			if (item.components.edible and item.components.edible.healthvalue >= 0 and item.components.edible.hungervalue >= 0 and item.components.edible.sanityvalue >= 0 
			and item.components.perishable and item.components.perishable:IsFresh() and not item.components.perishable:IsStale() and not item.components.perishable:IsSpoiled())
			or (item.components.edible and item.components.edible.healthvalue >= 0 and item.components.edible.hungervalue >= 0 and item.components.edible.sanityvalue >= 0 
			and not item.components.perishable) then
				if not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
					inst.components.container:RemoveItem(item)
				end
				print("This food tastes great!")
				--Sync Princess's sanity gain with the slow eating animation!
				if item.components.edible.foodtype == "MEAT" then
					inst:DoTaskInTime(1, function()
						local sanity = giver.components.sanity
						if sanity then
							sanity:DoDelta((item.components.edible.hungervalue/10) + (item.components.edible.sanityvalue/5), false)
						end
					end)
			else
				--Sync Princess's sanity gain with the fast eating animation!
					inst:DoTaskInTime(0.5, function()
						local sanity = giver.components.sanity
						if sanity then
							sanity:DoDelta((item.components.edible.hungervalue/10) + (item.components.edible.sanityvalue/5), false)
						end
					end)
				end
			--If the food is very harmful to Prince, Princess will lose sanity!
			elseif (item.components.edible and (item.components.edible.healthvalue < 0 or item.components.edible.hungervalue < 0) 
			or (item.components.edible and item.components.edible.sanityvalue < 0))
			or (item.components.edible and item.components.perishable and item.components.perishable:IsSpoiled() and not item.components.perishable:IsStale()) then
				if not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
					inst.components.container:RemoveItem(item)
				end
				print("This food tastes awful!")
				--Sync Princess's sanity drain with the slow eating animation!
				if item.components.edible.foodtype == "MEAT" then
					inst:DoTaskInTime(1, function()
						local sanity = giver.components.sanity
						if sanity then
							if item.components.perishable and item.components.perishable:IsSpoiled() then
								sanity:DoDelta(-TUNING.SANITY_TINY, false)
							else
								sanity:DoDelta((item.components.edible.healthvalue/2) + (item.components.edible.sanityvalue/2), false)
							end
						end
					end)
			else
				--Sync Princess's sanity drain with the fast eating animation!
					inst:DoTaskInTime(0.5, function()
						local sanity = giver.components.sanity
						if sanity then
							if item.components.perishable and item.components.perishable:IsSpoiled() then
								sanity:DoDelta(-TUNING.SANITY_TINY, false)
							else
								sanity:DoDelta((item.components.edible.healthvalue/2) + (item.components.edible.sanityvalue/2), false)
							end
						end
					end)
				end
			end
			if math.random() < 0.50 then
				local princesstalk = {"You look hungry, Prince.", "Here you go, Prince.", "I got something for you, Prince!", "Here Prince, take a bite!"}
        		giver.components.talker:Say(princesstalk[math.random(1, #princesstalk)])

        		--inst:DoTaskInTime(0.5, function()
					--local princetalk = {"Thank you, Princess.", "Thank you!"}
        			--inst.components.talker:Say(princetalk[math.random(1, #princetalk)])
        		--end)
			end
        end

		local action = BufferedAction(inst, item, ACTIONS.EAT)
		inst:PushBufferedAction(action)

		--There's a bug where Prince won't eat something if it can add to a stack in his inventory. This is a workaround for such a case.
		local food = inst.components.container:FindItem(function(edible)
    		return inst.components.eater:CanEat(edible) and edible.prefab == item.prefab
    	end)
    	if food then
    		if not inst.sg:HasStateTag("busy") then
    			local action = BufferedAction(inst, food, ACTIONS.EAT)
				inst:PushBufferedAction(action)
    		end

		end

	end


	--Have Princess and Prince occasionally chat if Princess is arming Prince!
	if item.components.equippable then
		inst.components.container:Close()
		if giver.prefab == "princess" and item.prefab ~= "flowerhat" and item.prefab ~= "grass_umbrella" then
			if math.random() < 0.50 then
				local princesstalk = {"Here, Prince. Put this on.", "Here you are, Prince.", "I have something for you, Prince."}
        		giver.components.talker:Say(princesstalk[math.random(1, #princesstalk)])

        		inst:DoTaskInTime(0.5, function()
					local princetalk = {"Thank you, Princess.", "Thank you."}
        			inst.components.talker:Say(princetalk[math.random(1, #princetalk)])
        		end)
        	end
    --Have Princess and Prince always talk if Princess is giving Prince a flowery wearable!
        elseif giver.prefab == "princess" and (item.prefab == "flowerhat" or item.prefab == "grass_umbrella") then
			local princesstalk = {"Here, Prince. It's made with flowers!", "I know how much you love flowers, Prince.", "I think you'll like this, Prince!"}
        	giver.components.talker:Say(princesstalk[math.random(1, #princesstalk)])

        	inst:DoTaskInTime(0.5, function()
				local princetalk = {"Thank you, Princess!", "It smells so nice...", "It must be very pretty."}
        		inst.components.talker:Say(princetalk[math.random(1, #princetalk)])
        	end)
        end
    end

	--Equip headgear if Princess gives him headgear!
	if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.HEAD then
		local headpiece = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
		if headpiece then
			inst.components.inventory:RemoveItem(headpiece)--:Remove()  
			inst.components.container:GiveItem(headpiece)
		end
		inst.components.inventory:Equip(item)
	end

	--Equip hand tool if Princess gives him a tool!
	if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.HANDS 
	and item.components.equippable.equipslot ~= EQUIPSLOTS.HEAD and item.components.equippable.equipslot ~= EQUIPSLOTS.BODY 
	and item.prefab ~= "tarlamp" then
		local tool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
		if tool then
			inst.components.inventory:RemoveItem(tool)--:Remove()  
			inst.components.container:GiveItem(tool)
		end
		inst.components.inventory:Equip(item)
	end

	--Equip clothes/armor if Princess gives him clothes/armor!
	if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.BODY then
		local body = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
		if body then
			inst.components.inventory:RemoveItem(body)--:Remove()  
			inst.components.container:GiveItem(body)
		end
		inst.components.inventory:Equip(item)
	end


	--If Shipwrecked or Hamlet are available, Prince will try and equip a boat equippable if he's riding on a boat!
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then

		if (item.components.equippable and item.components.equippable.boatequipslot == BOATEQUIPSLOTS.BOAT_LAMP) 
		or (item.components.equippable and item.components.equippable.boatequipslot == BOATEQUIPSLOTS.BOAT_SAIL) then
			if inst.components.driver and inst.components.driver.vehicle and inst.components.driver.vehicle.components.container.hasboatequipslots then
				inst.components.container:Close()
				local boat = inst.components.driver.vehicle
				local equipslot = item.components.equippable.boatequipslot
				local princeboatitem = boat.components.container:GetItemInBoatSlot(equipslot)
				--A different way to put the unequipped item back in Prince's container
    			if princeboatitem then
    				--local boatslot = boatitem.components.equippable.boatequipslot
    				boat.components.container:RemoveItemBySlot(equipslot)--:Remove()  
					inst.components.container:GiveItem(princeboatitem)
    			end
    			inst.components.driver.vehicle.components.container:Equip(item)
			end
		end
		--If Prince isn't riding a boat, just accept the boat equippable into his container!
		if ((item.components.equippable and item.components.equippable.boatequipslot == BOATEQUIPSLOTS.BOAT_LAMP) 
		or (item.components.equippable and item.components.equippable.boatequipslot == BOATEQUIPSLOTS.BOAT_SAIL)) 
		and inst.components.driver and not inst.components.driver.vehicle then
			if item.prefab == "tarlamp" then
				local tool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
				if tool then
					inst.components.inventory:RemoveItem(tool)--:Remove()  
					inst.components.container:GiveItem(tool)
				end
				inst.components.inventory:Equip(item)
			else
				inst.components.container:GiveItem(item)
			end
		end

	end


	--If the item isn't edible or equippable, just leave it in Prince's container!
	if not item.components.edible and not item.components.equippable then
		inst.components.container:GiveItem(item)
	end

	--When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
	if inst:HasTag("armingprince") then
		--inst:DoTaskInTime(0.05, function() 
			inst:RemoveTag("armingprince") 
		--end)
	end

end

--The only item Princess won't want to give Prince is dark petals!
local function OnRefuseItem(inst, item)
	local princess = FindEntity(inst, 3, function(ent)
		return ent:HasTag("liarprincess")
	end)
	if princess then
		local princesstalk = {"(I think I'll keep these away from Prince...)", "(I don't think Prince would like these...)", "(Prince might not like this flower...)"}
    	princess.components.talker:Say(princesstalk[math.random(1, #princesstalk)])
	end
end

--Prince will unequip his gear if asked!
local function UnequipGear(inst)
	--inst:AddTag("armingprince")

	local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
	if handslot then

		--Quit fishing if Prince is getting unequipped!
        if inst.sg:HasStateTag("fishing") then
        	if handslot.components.fishingrod then
            	handslot.components.fishingrod:StopFishing()
            end
            inst.sg:GoToState("fishing_pst")
        end
			--inst.components.inventory:DropItem(handslot, true, true)
		if handslot.prefab == "lantern" or handslot.prefab == "bottlelantern" then
			inst.components.inventory:RemoveItem(handslot)
			inst.components.inventory:DropItem(handslot, true, true) 
		else
			inst.components.inventory:RemoveItem(handslot)--:Remove()  
			inst.components.container:GiveItem(handslot)
			--inst.components.inventory:RemoveItemBySlot(handslot)   
		end
	end


	local bodyslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY) --inst.components.container:RemoveItemBySlot(i)
	if bodyslot then
		--inst.components.inventory:DropItem(bodyslot, true, true)
		inst.components.inventory:RemoveItem(bodyslot)--:Remove()  
		inst.components.container:GiveItem(bodyslot)
		--inst.components.inventory:RemoveItemBySlot(bodyslot)
	end


	local headslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
	if headslot then
		--inst.components.inventory:DropItem(headslot, true, true)
		inst.components.inventory:RemoveItem(headslot)--:Remove()  
		inst.components.container:GiveItem(headslot)
		--inst.components.inventory:RemoveItemBySlot(headslot)
	end


	--If Shipwrecked or Hamlet are available, Prince will try and equip a boat equippable if he's riding on a boat!
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then

		if inst.components.driver and inst.components.driver.vehicle and inst.components.driver.vehicle.components.container.hasboatequipslots and inst.components.driver.vehicle.prefab ~= "woodlegsboat" then
			local boat = inst.components.driver.vehicle
			local princeboatlampslot = boat.components.container:GetItemInBoatSlot(BOATEQUIPSLOTS.BOAT_LAMP)
			local princeboatsailslot = boat.components.container:GetItemInBoatSlot(BOATEQUIPSLOTS.BOAT_SAIL)
				--A different way to put the unequipped item back in Prince's container
    		if princeboatlampslot then
    			--local boatslot = boatitem.components.equippable.boatequipslot
    			boat.components.container:Unequip(princeboatlampslot)--:Remove()  
    			boat.components.container:RemoveItem(princeboatlampslot)--:Remove()  
				inst.components.container:GiveItem(princeboatlampslot)
    		end
    		if princeboatsailslot then
    			boat.components.container:Unequip(princeboatsailslot)--:Remove()  
    			boat.components.container:RemoveItem(princeboatsailslot)--:Remove()  
				inst.components.container:GiveItem(princeboatsailslot)
    		end
		end

	end

	--inst:DoTaskInTime(0.1, function() inst:RemoveTag("armingprince") end)
end

--Prince will arm himself in combat!
local function OnNewTarget(inst, data)

	inst.components.container:Close()
	local container = inst.components.container
	for i = 1, container:GetNumSlots() do
		local item = container:GetItemInSlot(i)

		--This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
		inst:AddTag("armingprince")

		if item then

			print("Prince is arming up for battle!")

				--Weirdly, tools don't label themselves as hand equippables...
				if item.components.equippable and item.components.weapon and inst.components.follower.leader then
					local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
						if ((currenthand and not currenthand.components.weapon) 
						--If Prince is already wielding a weapon class tool, find another weapon in his inventory that does more damage!
						or (currenthand and currenthand.components.weapon and item.components.weapon.damage > currenthand.components.weapon.damage))
						--If Prince is holding a handheld light source, he won't switch to a weapon!
						and currenthand.prefab ~= "torch" and currenthand.prefab ~= "tarlamp" then

								inst.components.inventory:RemoveItem(currenthand)--:Remove()  
								--If what Prince was wielding was a placeable light source like a lantern or bottle lantern, he'll drop it to the floor!
								if currenthand.prefab == "lantern" or currenthand.prefab == "bottlelantern" then
									inst.components.inventory:DropItem(currenthand, true, true) 
								else
									inst.components.container:GiveItem(currenthand)
								end
								inst.components.container:RemoveItemBySlot(i)--:Remove()  
								inst.components.inventory:Equip(item)
						end

						if not currenthand then
							inst.components.container:RemoveItemBySlot(i)--:Remove()  
							inst.components.inventory:Equip(item)
						end
				end

				--Separate hand code for when Prince needs to run away without Princess!
				if inst.components.follower and not inst.components.follower.leader then
					if item.components.equippable and item.prefab == "cane" then
						local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
							if currenthand and currenthand.prefab ~= "cane" then

									inst.components.inventory:RemoveItem(currenthand)--:Remove()  
									inst.components.container:GiveItem(currenthand)
									inst.components.container:RemoveItemBySlot(i)--:Remove()  
									inst.components.inventory:Equip(item)
							end

							if not currenthand then
								inst.components.container:RemoveItemBySlot(i)--:Remove()  
								inst.components.inventory:Equip(item)
							end	
					end
				end

				if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.BODY and item.components.armor then
					local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
						if currentbody and currentbody.prefab ~= "amulet" and not currentbody.components.armor then

							inst.components.inventory:RemoveItem(currentbody)--:Remove()  
							inst.components.container:GiveItem(currentbody)
							inst.components.container:RemoveItemBySlot(i)--:Remove()  
							inst.components.inventory:Equip(item)
						end

						if not currentbody then
							inst.components.container:RemoveItemBySlot(i)--:Remove()  
							inst.components.inventory:Equip(item)
						end
				end

				if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.HEAD and item.components.armor then
					local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
						if (currenthead and not currenthead.components.armor) and (currenthead and currenthead.prefab ~= "gasmaskhat" and currenthead.prefab ~= "earmuffshat") then

							inst.components.inventory:RemoveItem(currenthead)--:Remove()  
							inst.components.container:GiveItem(currenthead)
							inst.components.container:RemoveItemBySlot(i)--:Remove()  
							inst.components.inventory:Equip(item)
						end

						if not currenthead then
							inst.components.container:RemoveItemBySlot(i)--:Remove()  
							inst.components.inventory:Equip(item)
						end
				end
	

		end

		--When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
		inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
	end

end


local function KeepTarget(inst, target)
	local princess = FindEntity(inst, 7, function(ent)
		return ent:HasTag("liarprincess")
	end)
	if inst.components.follower.leader and not princess then
		return false
	end
	if target and target.components.health and not target.components.health:IsDead() then --and target:IsValid()
		local distsq = target:GetDistanceSqToInst(inst)
		return distsq < 2500
	else
		return false
	end
end

local function OnAttacked(inst, data)
    local attacker = data.attacker

    inst.components.container:Close()

    	--Seeing Prince get hurt will make Princess feel bad!
    	local princess = FindEntity(inst, 20, function(ent)
			return ent:HasTag("liarprincess") or ent:HasTag("wolf")
		end)
		if princess then
			local sanity = princess.components.sanity
			if sanity then
				sanity:DoDelta(-1, false)
			end
		end

		inst.components.container:Close()
		local container = inst.components.container
		for i = 1, container:GetNumSlots() do
			local item = container:GetItemInSlot(i)

			--This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
			inst:AddTag("armingprince")

			if item then

				if item.components.equippable and item.components.weapon and inst.components.follower.leader then
					local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
						if ((currenthand and not currenthand.components.weapon) 
						--If Prince is already wielding a weapon class tool, find another weapon in his inventory that does more damage!
						or (currenthand and currenthand.components.weapon and item.components.weapon.damage > currenthand.components.weapon.damage))
						--If Prince is holding a handheld light source, he won't switch to a weapon!
						and currenthand.prefab ~= "torch" and currenthand.prefab ~= "tarlamp" then

								inst.components.inventory:RemoveItem(currenthand)--:Remove()  
								--If what Prince was wielding was a placeable light source like a lantern or bottle lantern, he'll drop it to the floor!
								if currenthand.prefab == "lantern" or currenthand.prefab == "bottlelantern" then
									inst.components.inventory:DropItem(currenthand, true, true) 
								else
									inst.components.container:GiveItem(currenthand)
								end
								inst.components.container:RemoveItemBySlot(i)--:Remove()  
								inst.components.inventory:Equip(item)
						end

						if not currenthand then
							inst.components.container:RemoveItemBySlot(i)--:Remove()  
							inst.components.inventory:Equip(item)
						end
				end

				--Separate hand code for when Prince needs to run away without Princess!
				if inst.components.follower and not inst.components.follower.leader then
					if item.components.equippable and item.prefab == "cane" then
						local currenthand = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
							if currenthand and currenthand.prefab ~= "cane" then

									inst.components.inventory:RemoveItem(currenthand)--:Remove()  
									inst.components.container:GiveItem(currenthand)
									inst.components.container:RemoveItemBySlot(i)--:Remove()  
									inst.components.inventory:Equip(item)
							end

							if not currenthand then
								inst.components.container:RemoveItemBySlot(i)--:Remove()  
								inst.components.inventory:Equip(item)
							end
					end

				end

			end

			inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)

		end

		local body = inst.components.inventory and inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)

        --If someone is assaulting him, then check if the Queen Womant has stunned him.
        if not inst.sanity_stunned then
        --If Prince is not stunned, then decide what to do next if Prince is wearing a snurtle shell.
            if body and body.prefab == "armorsnurtleshell" 
            and not inst.sg:HasStateTag("eating") and not inst.sg:HasStateTag("attack")
            and not inst.components.health:IsDead() 
            --and (inst.components.follower and inst.components.follower.leader)
            and math.random() < 0.33 then 
            	inst.sg:GoToState("shell_enter") 
            end
        end


		--Prince's version of getting thieved!
		if data.attacker.prefab == "frog" or data.attacker.prefab == "grabbing_vine" then
			local item = inst.components.container:FindItem(function(victim) return not victim:HasTag("nosteal") end)
			if item then
				inst.components.container:DropItem(item)
			end
		end


		--Prince will flee from danger when under attack and not in Princess' company!
    	if inst.components.follower and not inst.components.follower.leader then

			if not attacker:HasTag("liarprincess") then
				inst.components.combat:SetTarget(attacker)
			end

		end
end


local function AttractShadows(inst)

	if inst.components.sanity:IsCrazy() then
		local shadow = FindEntity(inst, 20, function(ent)
			return ent:HasTag("shadowcreature")
		end)
		if shadow and not inst.sg:HasStateTag("hiding") then
			shadow.components.combat:SuggestTarget(inst)
		end
	end

	local nightmare = FindEntity(inst, 20, function(ent)
		return ent.prefab == "crawlingnightmare" or ent.prefab == "nightmarebeak"
	end)
	if nightmare and not inst.sg:HasStateTag("hiding") then
		nightmare.components.combat:SuggestTarget(inst)
	end

end

--Prince lightning response! Ripped from playerlightningtarget.
local function StruckByLightning(inst, instant)
    if inst.components.health and not inst.components.health:IsDead() then
        local protected = false
        --Used to be GetPlayer() instead of inst. Stop being selfish, components!
        if inst.components.inventory:IsInsulated() then
            protected = true
        end

        if not protected then

            if not instant then
            	local mult = 1 * inst.components.moisture:GetMoisturePercent()
            	local damage = -(10 + (mult * 10))
            	inst.components.health:DoDelta(damage)
            	inst.components.burnable:Extinguish()
            	inst.sg:GoToState("electrocute")
        	end
        else
        	if not instant then
            	inst:PushEvent("lightningdamageavoided")
        	end

        end
    end
end


local function onlightning(inst, data)
    --if data.rod == inst then
        StruckByLightning(inst)
    --end
end


local function ontrawl(inst, data)
	--If the player is on deep water then try to spawn kraken
	local krakener = inst.components.krakener
	if krakener then
		local chance = krakener.spawn_chance
		if math.random() <= chance then
			local tiles = krakener.valid_tiles
			local tile = GetMap():GetTileAtPoint(inst:GetPosition():Get())
			if table.contains(tiles, tile) then
				krakener:DoKrakenEvent()
			end
		end
	end
end


local function DisposeTool(inst)
	local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
    if handslot 
    and ((handslot.components.perishable and handslot.components.perishable.perishremainingtime <= 0) 
    or (handslot.components.fueled and handslot.components.fueled.currentfuel <= 0)) 
	and handslot.prefab ~= "lantern" and handslot.prefab ~= "bottlelantern" then
		inst.components.container:Close()
        inst.components.inventory:RemoveItem(handslot)
        inst.components.container:GiveItem(handslot)
        handslot:Remove()
        inst.AnimState:Hide("ARM_carry") 
        inst.AnimState:Show("ARM_normal") 
        inst.sg:GoToState("item_out")
    end
end


local function IsPoisonDisabled()
	local world = GetWorld()
	return world and world.components.globalsettings and world.components.globalsettings.settings.poisondisabled and world.components.globalsettings.settings.poisondisabled == true
end


local function BeginGas(inst)
	if inst.gasTask == nil then
		inst.gasTask = inst:DoPeriodicTask(TUNING.GAS_INTERVAL,
			function()				

				local safe = false
				-- check armour
				if inst.components.inventory then
					for k,v in pairs (inst.components.inventory.equipslots) do
						if v.components.equippable and v.components.equippable:IsPoisonGasBlocker() then
							safe = true
						end		
					end
				end

				if inst:HasTag("has_gasmask") then
					safe = true
				end

				if IsPoisonDisabled() then
					safe = true
				end
				
				if not safe then
					inst.components.health:DoGasDamage(TUNING.GAS_DAMAGE_PER_INTERVAL)			
					inst:PushEvent("poisondamage")	
					inst.components.talker:Say(GetString(inst.prefab, "ANNOUNCE_GAS_DAMAGE"))
				end
			end)
	end
end

local function EndGas(inst)
	if inst.gasTask then
		inst.gasTask:Cancel()
		inst.gasTask = nil
	end
end

local function OnGasChange(inst, onGas)
	if not inst.gassources then
		inst.gassources = 0
	end

	if onGas then
		inst.gassources = inst.gassources +1	
		if inst.gassources > 0 and not inst.gasTask then
			BeginGas(inst)
		end
	else
		inst.gassources = math.max(0,inst.gassources - 1)
		if inst.gassources < 1 then
			EndGas(inst)
		end
	end
end

local function OnTileChange(inst, tile, tileInfo)
end


--Code ripped from Hamlet's season manager, heehee.
local function ToggleSneeze(inst)
	--Start sneezing!
	if GetSeasonManager():GetSeason() == SEASONS.LUSH then
		if GetSeasonManager():GetPercentSeason() > 0.1 then
			if inst.components.hayfever and not inst.components.hayfever.enabled then
				inst.components.hayfever:Enable()
			end
		end
	end

	--Stop sneezing!
	if GetSeasonManager():GetSeason() ~= SEASONS.LUSH and GetSeasonManager():GetPercentSeason() > 0.02 then
		if inst.components.hayfever and inst.components.hayfever.enabled then
			inst.components.hayfever:Disable()
		end
	end		

end

--Make Prince drop random items in his container!
local function SneezeDropItems(inst)

	local itemstodrop = 0 
    if math.random() < 0.6 then
        itemstodrop = itemstodrop +1
    end
    if math.random() < 0.1 then
        itemstodrop = itemstodrop +1
    end    

    if itemstodrop > 0 then
        local findItems = inst.components.container:FindItems(
                        function(item)
                            return not item:HasTag("nosteal")
                        end) 

        for i=1,itemstodrop do            
            for i=1, itemstodrop do
                if #findItems > 0 then
                    local itemnum = math.random(1,#findItems)
                    local item = findItems[itemnum]

                    table.remove(findItems,itemnum)

                    if item then
                        local direction = Vector3(math.random(1)-2,0,math.random(1)-2)        
                        print("DROPPING",item.prefab)   
                        inst.components.container:DropItem(item, false, direction:GetNormalized())
                    end                   

                end                      
            end
        end
    end

end


local function onopen(inst) 

	--inst.components.trader:Disable()

	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then

			    local slotpos_cargo = {

                    --Left slots!                           --Right slots!
                    Vector3(-162, 130 +114, 0),             Vector3(-87, 130 +114, 0), --Slot 2
                    Vector3(-162, 65 +114, 0),              Vector3(-87, 65 +114, 0), --Slot 4
                    Vector3(-162, 114, 0),                  Vector3(-87, 114, 0), --Slot 6
                    Vector3(-162, -65 +114, 0),             Vector3(-87, -65 +114, 0), --Slot 8
                    Vector3(-162, -130 +114, 0),            Vector3(-87, -130 +114, 0), --Slot 10
                    Vector3(-162, -195 +114, 0),            Vector3(-87, -195 +114, 0), --Slot 12
                    Vector3(-162, -260 +114, 0),            Vector3(-87, -260 +114, 0), --Slot 14
                    Vector3(-162, -325 +114, 0),
                    
                    Vector3(8, 95 +114, 0), --Cargo Boat Slot 1
                    Vector3(8, 20 +114, 0), --Cargo Boat Slot 2
                    Vector3(8, -55 +114, 0), --Cargo Boat Slot 3
                    Vector3(8, -130 +114, 0), --Cargo Boat Slot 4
                    Vector3(8, -205 +114, 0), --Cargo Boat Slot 5
                    Vector3(8, -280 +114, 0) --Cargo Boat Slot 6

                    --Leftmost 3 slots! The "equippables" slots! Slot number 1 is the hand slot, slot number 6 is the body slot, and slot number 11 is the head slot!
                    --Vector3(-(180), 64+32+8+4, -(180)), --HAND SLOT!
                    --Vector3(-(180), 32+4, -(180)), --BODY SLOT!
                    --Vector3(-(180), -(32+4), -(180)), --HEAD SLOT!

                    }

    			local slotpos_encrusted = {

                    --Left slots!                           --Right slots!
                    Vector3(-162, 130 +114, 0),             Vector3(-87, 130 +114, 0), --Slot 2
                    Vector3(-162, 65 +114, 0),              Vector3(-87, 65 +114, 0), --Slot 4
                    Vector3(-162, 114, 0),                  Vector3(-87, 114, 0), --Slot 6
                    Vector3(-162, -65 +114, 0),             Vector3(-87, -65 +114, 0), --Slot 8
                    Vector3(-162, -130 +114, 0),            Vector3(-87, -130 +114, 0), --Slot 10
                    Vector3(-162, -195 +114, 0),            Vector3(-87, -195 +114, 0), --Slot 12
                    Vector3(-162, -260 +114, 0),            Vector3(-87, -260 +114, 0), --Slot 14
                    Vector3(-162, -325 +114, 0),

                    Vector3(8, -205 +114, 0), --Encrusted Boat Slot 1
                    Vector3(8, -280 +114, 0) --Encrusted Boat Slot 2

                    --Leftmost 3 slots! The "equippables" slots! Slot number 1 is the hand slot, slot number 6 is the body slot, and slot number 11 is the head slot!
                    --Vector3(-(180), 64+32+8+4, -(180)), --HAND SLOT!
                    --Vector3(-(180), 32+4, -(180)), --BODY SLOT!
                    --Vector3(-(180), -(32+4), -(180)), --HEAD SLOT!

                    }

			if inst.boatstate == "CARGO" then
				inst.components.container:SetNumSlots(#slotpos_cargo)

				inst.components.container.widgetslotpos = slotpos_cargo
				--local widget = "ui_chester_shadow_3x4"
				--widget.Transform:SetScale( 1.5, 1.5, 1.5 )
				inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
				inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
				inst.components.container.widgetpos = Vector3(100, 220, 0)
				inst.components.container.side_align_tip = 160
				--inst.components.container.widgetbuttoninfo = widgetbuttoninfo
				local boat = inst.components.driver and inst.components.driver.vehicle
				if boat and boat.components.container then
                	for k,v in pairs(boat.components.container.slots) do
                    	local item = boat.components.container:RemoveItemBySlot(k)
                    	inst.components.container:GiveItem(item)
                	end
            	end
			elseif 
				inst.boatstate == "ENCRUSTED" then
				inst.components.container:SetNumSlots(#slotpos_encrusted)

				inst.components.container.widgetslotpos = slotpos_encrusted
				--local widget = "ui_chester_shadow_3x4"
				--widget.Transform:SetScale( 1.5, 1.5, 1.5 )
				inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
				inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
				inst.components.container.widgetpos = Vector3(100, 220, 0)
				inst.components.container.side_align_tip = 160
				--inst.components.container.widgetbuttoninfo = widgetbuttoninfo
				local boat = inst.components.driver and inst.components.driver.vehicle
				if boat and boat.components.container then
                	for k,v in pairs(boat.components.container.slots) do
                    	local item = boat.components.container:RemoveItemBySlot(k)
                    	inst.components.container:GiveItem(item)
                	end
            	end
			elseif 
			inst.boatstate == "NORMAL" then
			inst.components.container:SetNumSlots(#slotpos)

			inst.components.container.widgetslotpos = slotpos
			--local widget = "ui_chester_shadow_3x4"
			--widget.Transform:SetScale( 1.5, 1.5, 1.5 )
			inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
			inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
			inst.components.container.widgetpos = Vector3(100, 220, 0)
			inst.components.container.side_align_tip = 160
			--inst.components.container.widgetbuttoninfo = widgetbuttoninfo
		end
	end

end


local function onclose(inst)

	--inst.components.trader:Enable()

end


local function entitydeathfn(inst, data)
	if (data.inst:HasTag("liarprincess") or data.inst:HasTag("wolf")) and data.inst.components.resurrectable and data.inst.components.resurrectable:CanResurrect() and not inst.CaveVictim then
		--inst:DoTaskInTime(0.5, function() 
			--if not inst.sg:HasStateTag("dead") then
				inst.components.health:Kill() 
			--end
			inst:DoTaskInTime(0.25, function()
				if (IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC)) and inst.components.follower and not inst.components.follower.leader then
					print("Take my hand, Prince!")
					GetPlayer().components.leader:AddFollower(inst)
				end
			end)
		--end)
	end
end


--Prince should gain sanity in player houses!
local function sanityfn(inst)
    return inst.HouseSanity
end


--If there is another Prince present, BANISH HIM!
local function SeekPrinceToBanish(inst)

	local princetobanish = FindEntity(inst, 20, function(ent)
		return ent:HasTag("blindprince") and ent ~= inst
	end)

	if princetobanish then
		princetobanish:PushEvent("banish")
		inst:DoTaskInTime(6, function()
			inst:PushEvent("banishedaprince")
		end)
	end


end


local function OnStopFollowing(inst)
	inst.components.sanity.dapperness = -TUNING.DAPPERNESS_LARGE -- -(100/(480*4.5)) replicates -DAPPERNESS_MED_LARGE because it doesn't exist in base game.
	if inst.components.follower and inst.components.follower.leader then
		local princessleader = inst.components.follower.leader

		if princessleader == GetPlayer() and princessleader.prefab == "princess" then
			princessleader.AnimState:SetBuild("princess")

			inst.AnimState:SetBuild("prince")
		end

		if princessleader:HasTag("princeleader") then
			princessleader:RemoveTag("princeleader")
		end
		if princessleader.components.sanity then
			princessleader.components.sanity.dapperness = 0
		end
	end
end

local function OnStartFollowing(inst) 
	--Princess calms Prince down!
	inst.components.sanity.dapperness = 0 --TUNING.DAPPERNESS_TINY
	--Prince calms Princess down more!
	if inst.components.follower and inst.components.follower.leader then

		local princessleader = inst.components.follower.leader

		if princessleader == GetPlayer() and princessleader.prefab == "princess" then
			princessleader.AnimState:SetBuild("princess_happy")

			inst.AnimState:SetBuild("prince_happy")
		end

		if not princessleader:HasTag("princeleader") then
			princessleader:AddTag("princeleader")
		end
		if princessleader.components.sanity then
			--480*6 is equal to the dapper vest's sanity regen. 480*4.5 is equal to the hibearnation vest's.
			princessleader.components.sanity.dapperness = (100/(480*4.5)) --(100/(480*6))
		end

		if inst.components.combat and inst.components.combat.target then
			inst.components.combat:SetTarget(nil)
		end



		--Quit fishing if Princess is calling him!
        local fishingrod = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
        if inst.sg:HasStateTag("fishing") then
        	if fishingrod and fishingrod.components.fishingrod then
            	fishingrod.components.fishingrod:StopFishing()
            end
            inst.sg:GoToState("fishing_pst")
        end

        --Prevent Prince's follower component from warping him back to Princess when he's telelocated a far distance!
        inst:RemoveEventCallback("entitysleep", inst.portnearleader)

	end

end


local function OnEntitySleep(inst)

end


local function OnEntityWake(inst)
	print("Prince is back!")
end

--onsave and onload are to keep track of Prince's boat state, to determine how big his container should be.
local function onsave(inst, data)
	data.CaveVictim = inst.CaveVictim

	data.boatstate = inst.boatstate

	data.HasBanished = inst.HasBanished
	data.HasWarped = inst.HasWarped --Helps to let Prince know there's a new Prince to banish.

	data.HouseSanity = inst.HouseSanity

	data.IsIndoors = inst.IsIndoors
end

local function onload(inst, data)
	inst.CaveVictim = data.CaveVictim --For Vanilla or Reign of Giants saves. A workaround for Prince not warping with Princess in/out of caves to resurrect on death!

	inst.boatstate = data.boatstate or "NORMAL" --Status for if Prince's boat has container slots!
	
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
		if not inst:GetIsOnWater() then
			inst.boatstate = "NORMAL" --Status for if Prince's boat has container slots!
		end
	end

	--Important so that Princess doesn't duplicate Princes when she merges worlds with another Princess world!
	inst.HasBanished = data.HasBanished or false
	inst.HasWarped = data.HasWarped or false --Checks whether Prince has used a Seaworthy or Skyworthy.

	inst.HouseSanity = data.HouseSanity or 0 --Check if Prince is in a player house for sanity regen!

	inst.IsIndoors = data.IsIndoors or false --Is Prince indoors or not?

	inst.entity:SetCanSleep(false)

end

local function OnPreLoad(inst, data)
	--local pos = Vector3(inst.Transform:GetWorldPosition())
	--local ents = TheSim:FindEntities(pos.x,pos.y,pos.z, 30)
	--for k,v in pairs(ents) do
		--if v ~= inst then
			--v:RunScript(v.prefab)
		--end
	--end

		--inst.entity:SetCanSleep(false)


		if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then

			if not data or not inst:GetIsOnWater() then return end

    		    local slotpos_cargo = {

                    --Left slots!                           --Right slots!
                    Vector3(-162, 130 +114, 0),             Vector3(-87, 130 +114, 0), --Slot 2
                    Vector3(-162, 65 +114, 0),              Vector3(-87, 65 +114, 0), --Slot 4
                    Vector3(-162, 114, 0),                  Vector3(-87, 114, 0), --Slot 6
                    Vector3(-162, -65 +114, 0),             Vector3(-87, -65 +114, 0), --Slot 8
                    Vector3(-162, -130 +114, 0),            Vector3(-87, -130 +114, 0), --Slot 10
                    Vector3(-162, -195 +114, 0),            Vector3(-87, -195 +114, 0), --Slot 12
                    Vector3(-162, -260 +114, 0),            Vector3(-87, -260 +114, 0), --Slot 14
                    Vector3(-162, -325 +114, 0),
                    
                    Vector3(8, 95 +114, 0), --Cargo Boat Slot 1
                    Vector3(8, 20 +114, 0), --Cargo Boat Slot 2
                    Vector3(8, -55 +114, 0), --Cargo Boat Slot 3
                    Vector3(8, -130 +114, 0), --Cargo Boat Slot 4
                    Vector3(8, -205 +114, 0), --Cargo Boat Slot 5
                    Vector3(8, -280 +114, 0) --Cargo Boat Slot 6

                    --Leftmost 3 slots! The "equippables" slots! Slot number 1 is the hand slot, slot number 6 is the body slot, and slot number 11 is the head slot!
                    --Vector3(-(180), 64+32+8+4, -(180)), --HAND SLOT!
                    --Vector3(-(180), 32+4, -(180)), --BODY SLOT!
                    --Vector3(-(180), -(32+4), -(180)), --HEAD SLOT!

                    }

    			local slotpos_encrusted = {

                    --Left slots!                           --Right slots!
                    Vector3(-162, 130 +114, 0),             Vector3(-87, 130 +114, 0), --Slot 2
                    Vector3(-162, 65 +114, 0),              Vector3(-87, 65 +114, 0), --Slot 4
                    Vector3(-162, 114, 0),                  Vector3(-87, 114, 0), --Slot 6
                    Vector3(-162, -65 +114, 0),             Vector3(-87, -65 +114, 0), --Slot 8
                    Vector3(-162, -130 +114, 0),            Vector3(-87, -130 +114, 0), --Slot 10
                    Vector3(-162, -195 +114, 0),            Vector3(-87, -195 +114, 0), --Slot 12
                    Vector3(-162, -260 +114, 0),            Vector3(-87, -260 +114, 0), --Slot 14
                    Vector3(-162, -325 +114, 0),
                    
                    Vector3(8, -205 +114, 0), --Encrusted Boat Slot 1
                    Vector3(8, -280 +114, 0) --Encrusted Boat Slot 2

                    --Leftmost 3 slots! The "equippables" slots! Slot number 1 is the hand slot, slot number 6 is the body slot, and slot number 11 is the head slot!
                    --Vector3(-(180), 64+32+8+4, -(180)), --HAND SLOT!
                    --Vector3(-(180), 32+4, -(180)), --BODY SLOT!
                    --Vector3(-(180), -(32+4), -(180)), --HEAD SLOT!

                    }

			if data.boatstate == "CARGO" then
				inst.components.container:SetNumSlots(#slotpos_cargo)

				inst.components.container.widgetslotpos = slotpos_cargo
				--local widget = "ui_chester_shadow_3x4"
				--widget.Transform:SetScale( 1.5, 1.5, 1.5 )
				inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
				inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
				inst.components.container.widgetpos = Vector3(100, 220, 0)
				inst.components.container.side_align_tip = 160
				--inst.components.container.widgetbuttoninfo = widgetbuttoninfo
			elseif 
				data.boatstate == "ENCRUSTED" then
				inst.components.container:SetNumSlots(#slotpos_encrusted)

				inst.components.container.widgetslotpos = slotpos_encrusted
				--local widget = "ui_chester_shadow_3x4"
				--widget.Transform:SetScale( 1.5, 1.5, 1.5 )
				inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
				inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
				inst.components.container.widgetpos = Vector3(100, 220, 0)
				inst.components.container.side_align_tip = 160
				--inst.components.container.widgetbuttoninfo = widgetbuttoninfo
			elseif 
				data.boatstate == "NORMAL" then
				inst.components.container:SetNumSlots(#slotpos)

				inst.components.container.widgetslotpos = slotpos
				--local widget = "ui_chester_shadow_3x4"
				--widget.Transform:SetScale( 1.5, 1.5, 1.5 )
				inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
				inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
				inst.components.container.widgetpos = Vector3(100, 220, 0)
				inst.components.container.side_align_tip = 160
				--inst.components.container.widgetbuttoninfo = widgetbuttoninfo
			end
	end

end



local function fn()
	local inst = CreateEntity()

	--Try to keep Prince active while he's absent!
	inst.entity:SetCanSleep(false)

	local trans = inst.entity:AddTransform()
	local anim = inst.entity:AddAnimState()
	local sound = inst.entity:AddSoundEmitter()
	local shadow = inst.entity:AddDynamicShadow()

	local minimap = inst.entity:AddMiniMapEntity()
	minimap:SetIcon("prince_icon.tex")
	--minimap:SetPriority( 9 )

	

	--inst.persists = true

	--For Prince's sanity code!
	local lightwatch = inst.entity:AddLightWatcher()
	lightwatch:SetLightThresh(.075)
	lightwatch:SetDarkThresh(.05)--DarkThresh is typically 0.05 for players, but Prince's is higher to counteract a bug where he won't recognize when he's in the dark after dismounting or mounting a boat.

	shadow:SetSize( 1.3, .6 )

	inst.Transform:SetFourFaced(inst)

	MakeCharacterPhysics(inst, 75, .5)

	anim:SetBank("wilson")
	anim:SetBuild("prince")
	anim:PlayAnimation("idle")


	--Use a different brain if sailing is available!
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
		local sailorbrain = require "brains/princebrain_sailor"
		inst:SetBrain(sailorbrain)
	else
		local brain = require "brains/princebrain"
		inst:SetBrain(brain)
	end


	local widgetbuttoninfo = {
	text = "Unequip",
	position = Vector3(-87, -325 +114, 0),
	fn = function(inst) UnequipGear(inst) end,
	}
	inst:AddComponent("container")
	inst.components.container:SetNumSlots(#slotpos)

	--inst.components.container.canbeopened = CheckSafety
	inst.components.container.onopenfn = onopen
	inst.components.container.onclosefn = onclose

	inst.components.container.widgetslotpos = slotpos
	inst.components.container.widgetanimbank = "ui_krampusbag_2x8"
	inst.components.container.widgetanimbuild = "ui_krampusbag_2x8"
	inst.components.container.widgetpos = Vector3(100, 220, 0)
	inst.components.container.side_align_tip = 160
	inst.components.container.widgetbuttoninfo = widgetbuttoninfo

	


	anim:Hide("ARM_carry")
	anim:Hide("hat")
	anim:Hide("hat_hair")
	anim:OverrideSymbol("fx_wipe", "wilson_fx", "fx_wipe")
    anim:OverrideSymbol("fx_liquid", "wilson_fx", "fx_liquid")
    anim:OverrideSymbol("shadow_hands", "shadow_hands", "shadow_hands")


	inst:AddTag("blindprince")
	inst:AddTag("companion")
	inst:AddTag("character")
	inst:AddTag("scarytoprey")
	inst:AddTag("notraptrigger")
	--To make Prince not get in the way of building and planting!
	inst:AddTag("NOBLOCK")

	inst:AddComponent("locomotor")
	inst.components.locomotor.runspeed = TUNING.WILSON_RUN_SPEED
	inst.components.locomotor:SetSlowMultiplier( 0.6 )
	inst.components.locomotor.pathcaps = { player = true, ignorecreep = true }
	inst.components.locomotor.fasteronroad = true

	inst:AddComponent("combat")
	inst.components.combat.hiteffectsymbol = "torso"
	-- inst.components.combat:SetRetargetFunction(1, Retarget)
	inst.components.combat:SetKeepTargetFunction(KeepTarget)
	inst.components.combat:SetAttackPeriod(1)
	inst.components.combat:SetRange(2)
	inst.components.combat:SetDefaultDamage(10)
	--Prince does not hit too hard!
	if not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
		inst.components.combat.damagemultiplier = .6
	else
		inst.components.combat:AddDamageModifier("prince", -0.4)
	end
	inst:ListenForEvent("attacked", OnAttacked)
	inst:ListenForEvent("newcombattarget", OnNewTarget)


	if (IsDLCEnabled(REIGN_OF_GIANTS) and not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) 
	or (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) 
	or (IsDLCEnabled(PORKLAND_DLC)) then
		inst:AddComponent("moisture")
		inst:ListenForEvent("moisturechange", function(inst, data) 

			if inst.components.moisture.moisture > 0 and not inst:HasTag("moisturetracked") then

				inst:AddTag("moisturetracked")

				local princess = GetPlayer()

                if princess.prefab == "princess" then
        			--if inst.components.driver and inst.components.driver.vehicle then
            			princess:PushEvent("princemoisture", {inst = inst})
        			--end
        		end

        	elseif inst.components.moisture.moisture <= 0 and inst:HasTag("moisturetracked") then

        		inst:RemoveTag("moisturetracked")

        	end	

		end, inst)
		--inst:AddTag("lightningrod") --Blanked out because this makes Prince attract too much lightning onto himself.
		--inst.lightningpriority = -1
		inst:ListenForEvent("lightningstrike", StruckByLightning) --This one may be more accurate. The event is pushed from Prince's brain!
		--Glowing for when Prince is struck by lightning!
		--local light = inst.entity:AddLight()
		--inst.Light:Enable(false)
		--inst.Light:SetRadius(.8)
		--inst.Light:SetFalloff(0.5)
		--inst.Light:SetIntensity(.65)
		--inst.Light:SetColour(255/255,255/255,236/255)
	end

	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
		--inst:AddComponent("poisonable")
		MakePoisonableCharacter(inst, nil, nil, 0, 0, 0)
		inst.components.poisonable.duration = TUNING.TOTAL_DAY_TIME * 3
		inst.components.poisonable.transfer_poison_on_attack = false
		--inst.components.poisonable.show_fx = false
	end

	inst:AddComponent("health")
	inst.components.health:SetMaxHealth(50)
	inst.components.health.nofadeout = true
	inst.components.health.fire_damage_scale = .39 --Designed so Prince doesn't instantly die when struck by lightning.
	inst:ListenForEvent("death", ondeath)

	inst:AddComponent("hunger")
	inst.components.hunger:SetMax(115)
	inst.components.hunger:SetRate(TUNING.WILSON_HUNGER_RATE*1.10)
	inst.components.hunger:SetKillRate(150/TUNING.STARVE_KILL_TIME)
	inst:AddComponent("eater")

	inst:AddComponent("inventory")


	inst:ListenForEvent("equip", function(it, data) 
		Print(VERBOSITY.DEBUG, "equip: ["..data.item.prefab.."]")
		inst.SoundEmitter:PlaySound("dontstarve/wilson/equip_item")

		--Make flowery equippables raise more sanity for Prince!
		if data.item.prefab == "flowerhat" then
			if IsDLCEnabled(REIGN_OF_GIANTS) or IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
        		data.item.components.equippable.dapperness = TUNING.DAPPERNESS_LARGE
    		else
    			--data.item.components.equippable.dapperness = TUNING.DAPPERNESS_MED
    			--RELIC OF AN OLDER VERSION! 
        		--if inst:HasComponent("dapperness") then
        			data.item.components.dapperness.dapperness = TUNING.DAPPERNESS_MED
        		--end
    		end
    	end

    	if data.item.prefab == "grass_umbrella" then
    		if IsDLCEnabled(REIGN_OF_GIANTS) or IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
        		data.item.components.equippable.dapperness = (100/(480*4.5)) -- (100/(480*4.5)) replicates DAPPERNESS_MED_LARGE because it doesn't exist in base game.
    		else
        		--data.item.components.equippable.dapperness = (100/(480*4.5)) -- (100/(480*4.5)) replicates DAPPERNESS_MED_LARGE because it doesn't exist in base game.
        		--if inst:HasComponent("dapperness") then
        			data.item.components.dapperness.dapperness = (100/(480*4.5)) -- (100/(480*4.5)) replicates DAPPERNESS_MED_LARGE because it doesn't exist in base game.
        		--end
    		end
    	end

	end)    
	inst:ListenForEvent("unequip", function(it, data) 
		--Print(VERBOSITY.DEBUG, "equip: ["..data.item.prefab.."]")
		--inst.SoundEmitter:PlaySound("dontstarve/wilson/equip_item")

		--The flowery equippable's dapperness will return to normal when no longer worn by Prince.
		if data.item.prefab == "flowerhat" then
			if IsDLCEnabled(REIGN_OF_GIANTS) or IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
        		data.item.components.equippable.dapperness = TUNING.DAPPERNESS_TINY
    		else
    			--RELIC OF AN OLDER VERSION!
        		--if inst:HasComponent("dapperness") then
        			data.item.components.dapperness.dapperness = TUNING.DAPPERNESS_TINY
        		--end
    		end
    	end
    	if data.item.prefab == "grass_umbrella" then
			if IsDLCEnabled(REIGN_OF_GIANTS) or IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
        		data.item.components.equippable.dapperness = TUNING.DAPPERNESS_SMALL
    		end
    	end

	end)    


	inst.components.inventory.dropondeath = false

	--Prince can talk, but he doesn't have much to say.
	inst:AddComponent("talker")
	--inst.components.talker.special_speech = true

	inst:AddComponent("trader")
	inst.components.trader:Enable()
	inst.components.trader:SetAcceptTest(ShouldAcceptItem)
	inst.components.trader.onaccept = OnGetItemFromPlayer
	--inst.components.trader.onrefuse = OnRefuseItem

	--Most of Prince's dialogue is through here, since he can't examine things due to his blindness.
	inst:AddComponent("wisecracker")


	--FOR SHIPWRECKED AND HAMLET BOATING
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
		inst:AddTag("amphibious")

		inst:AddComponent("driver")
		inst.components.driver.landstategraph = "SGprince_sw_h"
		inst.components.driver.boatingstategraph = "SGprinceboating"

		--Prince can anger Quacken too!
		inst:AddComponent("krakener")

		--This is here because the krakener component only considers players. Selfish.
		inst:ListenForEvent("trawlitem", ontrawl)

		--In case Prince is not on a boat and ends up stuck on water.
		--inst:AddComponent("keeponland") --NEVERMIND! IT KILL PRINCE!

		inst:DoTaskInTime(0.6, function()
			--Push an event for Princess to display Prince's boat's health!
        	local princess = GetPlayer()

            if princess.prefab == "princess" then
        		--if inst.components.driver and inst.components.driver.vehicle then
            		princess:PushEvent("princeboating", {inst = inst})
        		--end
        	end
        end)

	end

	--Put beneath the inventory component to avoid a crash!
	if not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
		inst:AddComponent("prince_sanitymonsterspawner")
	else
		inst:AddComponent("prince_sanitymonsterspawner_sw_h")
	end
	inst:AddComponent("sanity")
	inst.components.sanity:SetMax(TUNING.WILSON_SANITY)

	--Prince has some trauma involving scary monsters!
	inst.components.sanity.neg_aura_mult = 5


	inst:AddComponent("temperature")
	inst:AddComponent("frostybreather")

	inst:AddComponent("pinnable") --For that blasted ewecus.

	if IsDLCEnabled(REIGN_OF_GIANTS) or IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC) then
		MakeHugeFreezableCharacter(inst)
    	inst.components.freezable:SetDefaultWearOffTime(3)
    end


	inst.OnSave = onsave
	inst.OnLoad = onload
	inst.OnPreLoad = OnPreLoad
	--inst.OnEntitySleep = OnEntitySleep
	--inst.OnEntityWake = OnEntityWake


	local princess = GetPlayer()
	if princess.prefab == "princess" then
		inst:DoTaskInTime(0.01, function() 
			if not inst.CaveVictim then
				princess:PushEvent("checkprincestatus", {inst = inst}) 
			end
		end)
	end

	--Handles potential duplicated princes when world merging!
	inst:DoTaskInTime(0.2, function() 
		if inst.HasWarped then
			inst:DoTaskInTime(math.random(3), function()
				inst.PrinceBanisher = inst:DoPeriodicTask(1, SeekPrinceToBanish)
			end)
		end
	end)
	inst:ListenForEvent("banish", function(inst)
	
		if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
			if inst.components.driver and inst.components.driver.vehicle then
				inst.components.driver:OnDismount()
			end
		end
		inst:DoTaskInTime(0.2, function() 
			if not inst.sg:HasStateTag("banished") then
				inst.sg:GoToState("banished")
			end
		end)

	end)
	inst:ListenForEvent("banishedaprince", function(inst)
	
		if inst.HasWarped then
			inst.HasWarped = false
			inst.HasBanished = true
		end

		if inst.PrinceBanisher then
			inst.PrinceBanisher:Cancel()
			inst.PrinceBanisher = nil
		end

		local princess = GetPlayer()
		if princess.prefab == "princess" then
			inst:DoTaskInTime(5, function() princess:PushEvent("checkprincestatus", {inst = inst}) end)
		end

	end)

    inst:ListenForEvent("entity_death", function(world, data) entitydeathfn(inst, data) end, GetWorld())

	--Prince relies on Princess to find his way!
	inst:AddComponent("follower")
	inst:ListenForEvent("stopfollowing", OnStopFollowing)
	inst:ListenForEvent("startfollowing", OnStartFollowing)

	if inst.components.follower and not inst.components.follower.leader then
		inst.components.sanity.dapperness = -TUNING.DAPPERNESS_LARGE -- -(100/(480*4.5)) replicates -DAPPERNESS_MED_LARGE because it doesn't exist in base game.
	else
		inst.components.sanity.dapperness = 0
	end

	--Prince will follow Princess in and out of volcanoes!
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
		--Prince has a problem getting rid of items that wear off if Shipwrecked or Hamlet is enabled, this is to attempt fixing that.
		inst:ListenForEvent("torchranout", DisposeTool)

		inst.components.follower:SetFollowExitDestinations({EXIT_DESTINATION.LAND,EXIT_DESTINATION.WATER})
	end

	--Umbrellas seem to have the same problem as torches.
	inst:ListenForEvent("umbrellaranout", DisposeTool)

	--Mainly to avoid one man band bugs!
	inst:AddComponent("leader")

	--Allows the player to call Prince on command!
	inst:AddComponent("handholder")

	--inst:AddComponent("grue")
    --inst.components.grue:SetSounds("dontstarve/charlie/warn","dontstarve/charlie/attack")

    inst:AddComponent("inspectable")

	--Get shadow creatures to attack Prince when he's insane!
	inst:DoPeriodicTask(math.random( 3,6 ), AttractShadows)


	if IsDLCEnabled(PORKLAND_DLC) then

		inst.gasTask = nil

		--For gaseous rainforests!
		inst:AddComponent("tiletracker")
		inst.components.tiletracker:SetOnGasChangeFn(OnGasChange)
		-- inst.components.tiletracker:SetOnTileChangeFn(OnTileChange)
		inst.components.tiletracker:Start()

		inst.OnGasChange = OnGasChange


		--Make Prince not lose sanity from darkness if he's inside a building on loading a save.
		inst:DoTaskInTime(0.01, function() 
			local pt = inst:GetPosition()
			local x,y,z = inst.Transform:GetWorldPosition() 
    		local ents = TheSim:FindEntities(x,y,z, 10, {"INTERIOR_LIMBO"}, nil)
    		local tile = GetWorld().Map:GetTileAtPoint(pt.x,pt.y,pt.z)
    		if (tile == GROUND.INTERIOR) or (#ents >= 1) or (tile == nil) then
    			inst.components.sanity.night_drain_mult = 0
    		end
    	end)


		GetWorld():ListenForEvent("enterinterior", function(world, data)
			--Ensure loading/unloading an enterior doesn't send Prince back to limbo!
    		local pt = inst:GetPosition()
    		local tile = GetWorld().Map:GetTileAtPoint(pt.x,pt.y,pt.z)
    		if tile == GROUND.INTERIOR then
    			if not inst.IsIndoors then
					inst.IsIndoors = true
				end
				if GetInteriorSpawner() and GetInteriorSpawner():InPlayerRoom() then
    			--Prince should gain sanity in player houses!
				inst.HouseSanity = TUNING.SANITY_PLAYERHOUSE_GAIN --(100/(30*32)) --Replicates SANITY_PLAYERHOUSE_GAIN
    			end
    		end
    	end)

    	GetWorld():ListenForEvent("exitinterior", function(world, data)
    		--Ensure loading/unloading an enterior doesn't send Prince back to limbo!
			--GetInteriorSpawner():ReturnItemToScene(inst)
			--inst.entity:SetCanSleep(false)

    		local pt = inst:GetPosition()
    		local tile = GetWorld().Map:GetTileAtPoint(pt.x,pt.y,pt.z)
    		if tile ~= GROUND.INTERIOR and tile ~= GROUND.IMPASSABLE then
    			if inst.IsIndoors then
					inst.IsIndoors = false
				end
				inst.HouseSanity = 0
    		end
    	end)

    	inst.components.sanity.custom_rate_fn = sanityfn


		--For fog! Currently unused.
		inst:AddComponent("grogginess")
	    inst.components.grogginess:SetResistance(3)


		inst:AddComponent("hayfever")
		inst.components.hayfever.hayfeverimune = false

		--Delay for the sneeze code to prevent a weird crash when Prince is left indoors and Princess is outdoors.
		inst:DoTaskInTime(0.01, function()
			--Release the sneeze!
			if GetSeasonManager():GetSeason() == SEASONS.LUSH then
				if GetSeasonManager():GetPercentSeason() > 0.1 then
					if inst.components.hayfever and not inst.components.hayfever.enabled then
						inst.components.hayfever:Enable()
					end
				end
			end
		end)
		--Check each day to enable sneezing!
		inst:ListenForEvent( "daycomplete", function(i, data) ToggleSneeze(inst) end, GetWorld())
		--Get Prince to sneeze!
		inst:ListenForEvent("sneeze", function(inst, data)
        	if not inst.components.health:IsDead() and not inst.components.health.invincible then
            	if inst.sg:HasStateTag("busy") then
                	inst.wantstosneeze = true
            	else
                	inst.sg:GoToState("sneeze")
                	if (inst.IsIndoors and TheCamera.interior) or not (inst.IsIndoors) then
                		SneezeDropItems(inst)
                	end
            	end
        	end
    	end)    

		inst:AddComponent("infestable")

		--Make Prince bat food if in the Plateau!
		local is_porkland = SaveGameIndex:IsModePorkland()
		if is_porkland then
			inst:AddComponent("prince_batted")
			inst.components.prince_batted.neverattack = GetWorld().components.batted.neverattack
		end

	end





	--Make Prince susceptible to hound attacks!
	if IsDLCEnabled(PORKLAND_DLC) then
		if GetWorld().components.hounded and not SaveGameIndex:IsModePorkland() then
			inst:AddComponent("prince_hounded_h")
			inst.components.prince_hounded_h.spawnmode = GetWorld().components.hounded.spawnmode
		end
	elseif IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
		if GetWorld().components.hounded and not SaveGameIndex:IsModePorkland() then
			inst:AddComponent("prince_hounded_sw")
			inst.components.prince_hounded_sw.spawnmode = GetWorld().components.hounded.spawnmode
		end
	elseif IsDLCEnabled(REIGN_OF_GIANTS) and not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
		if GetWorld().prefab ~= "cave" then
			inst:AddComponent("prince_hounded_rog")
			inst.components.prince_hounded_rog.spawnmode = GetWorld().components.hounded.spawnmode
		end
	elseif not IsDLCEnabled(REIGN_OF_GIANTS) and not IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC) then
		if GetWorld().prefab ~= "cave" then
			inst:AddComponent("prince_hounded")
			inst.components.prince_hounded.spawnmode = GetWorld().components.hounded.spawnmode
		end
	end

	--Make Prince susceptible to depths worm attacks!
	if GetWorld().prefab == "cave" then
		inst:AddComponent("prince_periodicthreat")
    	local threats = require("fa_periodicthreats")
    	inst.components.prince_periodicthreat:AddThreat("WORM", threats["WORM"])
	end


	inst:DoTaskInTime(0.1, function()
        if inst.CaveVictim then
        	inst:PushEvent("banish")
        end
    end)


	--Prince will try to keep track of Princess! This code will also help him follow Princess through the Wooden Thing, and make entities persist when Prince is near them!
	inst:AddComponent("prince_persister")


	MakeMediumBurnableCharacter(inst, "torso")


	--Separate stategraphs because vanilla and Reign of Giants Don't Starve doesn't like the Shipwrecked or Hamlet additions.
	if (IsDLCEnabled(CAPY_DLC) and not IsDLCEnabled(PORKLAND_DLC)) or (IsDLCEnabled(PORKLAND_DLC)) then
		inst:SetStateGraph("SGprince_sw_h")
	else
		inst:SetStateGraph("SGprince")
	end


	inst.soundsname = "blindprince"

	STRINGS.CHARACTERS.PRINCE = require "speech_prince"


	return inst
end


return Prefab("prince", fn, assets, prefabs)