require("stategraphs/commonstates")

local function DoEquipmentFoleySounds(inst)
    for k, v in pairs(inst.components.inventory.equipslots) do
        if v.components.inventoryitem and v.components.inventoryitem.foleysound then
            inst.SoundEmitter:PlaySound(v.components.inventoryitem.foleysound)
        end
    end
end

local function DoFoleySounds(inst)
    DoEquipmentFoleySounds(inst)
end


local function DoRunSounds(inst)
    inst.sg.mem.foosteps = inst.sg.mem.foosteps + 1
    PlayFootstep(inst, inst.sg.mem.foosteps < 5 and 1 or .6)

    local pos = inst:GetPosition()
    if GetWorld().Flooding and GetWorld().Flooding:OnFlood(pos.x, 0, pos.z) then 
        local rot = inst.Transform:GetRotation()
        local splash = SpawnPrefab("splash_footstep")
        local CameraRight = TheCamera:GetRightVec()
        local CameraDown = TheCamera:GetDownVec()
        local displacement = CameraRight:Cross(CameraDown) * .15
        local pos = pos - displacement 
        splash.Transform:SetPosition(pos.x,pos.y, pos.z)
        splash.Transform:SetRotation(rot)
    end     
end

local function ClearStatusAilments(inst)
    if inst.components.freezable ~= nil and inst.components.freezable:IsFrozen() then
        inst.components.freezable:Unfreeze()
    end
    if inst.components.pinnable ~= nil and inst.components.pinnable:IsStuck() then
        inst.components.pinnable:Unstick()
    end
end

local function ConfigureRunState(inst)
    if inst:HasTag("groggy") then
        inst.sg.statemem.groggy = true   
    else
        inst.sg.statemem.normal = true
    end
end

local function GetRunStateAnim(inst)
    return (inst.sg.statemem.groggy and "idle_walk")
        or "run"
end


local actionhandlers =
{ 
    ActionHandler(ACTIONS.CHOP, 
        function(inst, action)
            action.invobject = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if inst.sg:HasStateTag("midchop") then
                return "chop"
            else
                return "chop_start"
            end 
        end),
    ActionHandler(ACTIONS.DIG, 
        function(inst, action) 
            action.invobject = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if inst.sg:HasStateTag("middig") then
                return "dig"
            else
                return "dig_start"
            end 
        end),
    ActionHandler(ACTIONS.MINE, 
        function(inst, action) 
            action.invobject = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if inst.sg:HasStateTag("midmine") then
                return "mine"
            else
                return "mine_start"
            end 
        end),
    ActionHandler(ACTIONS.HACK, 
        function(inst, action) 
            action.invobject = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if inst.sg:HasStateTag("midhack") then
                return "hack"
            else
                return "hack_start"
            end
        end),
    ActionHandler(ACTIONS.FISH, 
    function(inst, action)
        action.invobject = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
        return "fishing_pre"
    end),
    ActionHandler(ACTIONS.EAT, 
        function(inst, action)
            if (inst.sg:HasStateTag("busy") and not inst.sg:HastStateTag("eating")) or inst.components.health:IsDead() then
                return nil
            end

            local obj = action.target or action.invobject
            if not (obj and obj.components.edible) then
                return nil
            end

            if inst:HasTag("hungryprince") then
                if obj.components.edible.foodtype == "MEAT" then
                    return "hungryeat"
                else
                    return "hungryquickeat"
                end
            end

            if obj.components.edible.foodtype == "MEAT" then
                return "eat"
            else
                return "quickeat"
            end
        end),

    ActionHandler(ACTIONS.CUREPOISON, "curepoison"),

    ActionHandler(ACTIONS.MOUNT, "jumponboatstart"),

}

local events = 
{
    CommonHandlers.OnLocomote(true, false),
    CommonHandlers.OnDeath(),
    EventHandler("blocked", function(inst, data)
        if not inst.components.health:IsDead() then
            if inst.sg:HasStateTag("shell") then
                inst.sg:GoToState("shell_hit")
            end
        end
    end),
    EventHandler("attacked", function(inst, data)
        if not inst.components.health:IsDead() then
            if (data.attacker and data.attacker:HasTag("insect")) or inst.sg:HasStateTag("not_hit_stunned") then
                local is_idle = inst.sg:HasStateTag("idle")
                if not is_idle then
                    -- avoid stunlock when attacked by bees/mosquitos
                    -- don't go to full hit state, just play sounds

                    inst.SoundEmitter:PlaySound("dontstarve/wilson/hit")        
                    
                    return
                end
            end
            if inst.components.pinnable and inst.sg:HasStateTag("pinned") then
                inst.sg:GoToState("pinned_hit")            
            elseif inst.sg:HasStateTag("shell") then
                inst.sg:GoToState("shell_hit")
            else
                if data.stimuli and data.stimuli == "electric" and not inst.components.inventory:IsInsulated() then
                    inst.sg:GoToState("electrocute")
                else
                    inst.sg:GoToState("hit")
                end
            end
        end
    end),
    EventHandler("freeze", 
        function(inst)
            if inst.components.health and inst.components.health:GetPercent() > 0 then
                inst.sg:GoToState("frozen")
            end
        end),
    EventHandler("pinned",
        function(inst, data)
            if inst.components.health and not inst.components.health:IsDead() and inst.components.pinnable then
                if inst.components.pinnable.canbepinned then
                    inst.sg:GoToState("pinned_pre", data)
                elseif inst.components.pinnable:IsStuck() then
                    --V2C: Since sg events are queued, it's possible we're no longer pinnable
                    inst.components.pinnable:Unstick()
                end
            end
        end),   
    EventHandler("doattack", function(inst)
        if not inst.components.health:IsDead() and not inst.sg:HasStateTag("attack") and not inst.sg:HasStateTag("sneeze") then
            local weapon = inst.components.combat and inst.components.combat:GetWeapon()
            if weapon then
                print(weapon.prefab )
            else
                print("NO WEAPON")
            end
            inst.sg:GoToState("attack")
        end
    end),
    EventHandler("torchranout",
        function(inst, data)

        if not inst:HasTag("armingprince") then
            inst:AddTag("armingprince")
        end

        inst.components.container:Close()

        if not inst.components.inventory:IsItemEquipped(data.torch) then
            local sameTool = inst.components.container:FindItem(function(item)
                return item.prefab == data.torch.prefab
            end)
            if sameTool then
                inst.components.inventory:Equip(sameTool)
            end
        end

        if inst:HasTag("armingprince") then
            inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
        end

    end),
    EventHandler("umbrellaranout",
        function(inst, data)

            if not inst:HasTag("armingprince") then
                inst:AddTag("armingprince")
            end

            if not inst.components.inventory:IsItemEquipped(data.umbrella) then
                local sameTool = inst.components.container:FindItem(function(item)
                    return item.prefab == "umbrella" or item.prefab == "grass_umbrella" or item.prefab == "palmleaf_umbrella"
            end)

            if sameTool then
                inst.components.inventory:Equip(sameTool)
            end

            if inst:HasTag("armingprince") then
                inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
            end

        end
    end),

    EventHandler("ontalk", function(inst, data)
        if inst.sg:HasStateTag("idle") 
        and not inst.sg:HasStateTag("eating") and not inst.sg:HasStateTag("hiding")
        --Make sure Prince doesn't idle talk when he's offering to help.
        and ((inst.components.follower.leader and not inst.components.follower.leader.sg:HasStateTag("working")) 
        or not inst.components.follower.leader) then
            inst.sg:GoToState("talk", data.noanim)
        end
        
    end),

    EventHandler("equip", function(inst, data)
        if inst.sg:HasStateTag("idle") then
            if data.eslot == EQUIPSLOTS.HANDS then
                inst.sg:GoToState("item_out")
            else
                inst.sg:GoToState("item_hat")
            end
        end
    end),
    
    EventHandler("unequip", function(inst, data)
        
        if inst.sg:HasStateTag("idle") then
        
            if data.eslot == EQUIPSLOTS.HANDS then
                inst.sg:GoToState("item_in")
            else
                inst.sg:GoToState("item_hat")
            end
        end
    end),

    EventHandler("toolbroke",
        function(inst, data)
            inst.sg:GoToState("toolbroke", data.tool)
        end),      

    EventHandler("armorbroke",
        function(inst, data)
            if data.armor and data.armor.prefab ~= "armorvortexcloak" then
                inst.components.inventory:RemoveItem(data.armor)--:Remove()  
                inst.components.container:GiveItem(data.armor)
                data.armor:Remove()
            end
            inst.sg:GoToState("armorbroke")
        end),    

    EventHandler("landboat", function(inst)
        inst.sg:GoToState("jumpboatland")
    end),

    EventHandler("sneeze", function(inst, data)
        if not inst.components.health:IsDead() and not inst.components.health.invincible then
            if inst.sg:HasStateTag("busy") then
                inst.wantstosneeze = true
            else
                inst.sg:GoToState("sneeze")
            end
        end
    end),    

    -- Happens when the Ant Queen uses her song attack
    EventHandler ("sanity_stun", 
        function (inst, data)
            if not inst.components.inventory:IsItemNameEquipped("earmuffshat") then
                inst.sanity_stunned = true
                inst.sg:GoToState("sanity_stun")
                inst.components.sanity:DoDelta(-TUNING.SANITY_LARGE)

                inst:DoTaskInTime(data.duration or 3.5, function()  
                    if inst.sg.currentstate.name == "sanity_stun" then
                        inst.sg:GoToState("idle")
                        inst.sanity_stunned = false
                        inst:PushEvent("sanity_stun_over")
                    end
                end)
            end
        end),

    --A unique event for Prince so that he can take advantage of his snurtle shell.
    EventHandler("disengage", function(inst, data)
    --Some code ripped from the snurtle shell coding because it won't trigger when Prince hides.
        inst:DoTaskInTime(5, function() 
            if inst.sg:HasStateTag("shell") then
                local pt = inst:GetPosition()
                local ents = TheSim:FindEntities(pt.x, pt.y, pt.z, 20)

                for k,v in pairs(ents) do
                    if v.components.combat and v.components.combat.target and v.components.combat.target == inst then
                        v.components.combat:SetTarget(nil)
                    end
                end
            end
        end)

    end),
}

local states =
{

        State{
        name = "jumponboatstart",
        tags = {"doing", "busy", "canrotate"},

        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.components.locomotor:EnableGroundSpeedMultiplier(false)
            inst.AnimState:PlayAnimation("jumpboat")
            inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/boatjump_whoosh")
            local BA = inst:GetBufferedAction()
            inst.sg.statemem.startpos = inst:GetPosition()
            inst.sg.statemem.targetpos = BA.target and BA.target:GetPosition()

            RemovePhysicsColliders(inst)
            inst.components.health:SetInvincible(true)
        end,

        onexit = function(inst)
        --This shouldn't actually be reached
            ChangeToCharacterPhysics(inst)
            inst.components.locomotor:Stop()
            inst.components.locomotor:EnableGroundSpeedMultiplier(true)
            inst.components.health:SetInvincible(false)

        end,

        timeline =
        {
            TimeEvent(7*FRAMES, function(inst)
                inst:ForceFacePoint(inst.sg.statemem.targetpos:Get())
                local dist = inst:GetPosition():Dist(inst.sg.statemem.targetpos)
                local speed = dist / (18/30)
                inst.Physics:SetMotorVelOverride(1 * speed, 0, 0)
            end),
        },

        events =
        {
            EventHandler("animover", function(inst)
                inst.components.health:SetInvincible(false)
                inst.Transform:SetPosition(inst.sg.statemem.targetpos:Get())
                inst.Physics:Stop()
                ChangeToCharacterPhysics(inst)
                inst.components.locomotor:Stop()
                inst.components.locomotor:EnableGroundSpeedMultiplier(true)
                inst:PerformBufferedAction()
            end),
        },
    },

    State{
        name = "jumpoffboatstart",
        tags = {"doing", "busy", "canrotate"},

        onenter = function(inst, pos)
            inst.components.locomotor:Stop()
            inst.components.locomotor:EnableGroundSpeedMultiplier(false)
            inst.AnimState:PlayAnimation("jumpboat")
            inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/boatjump_whoosh")
            inst.sg.statemem.startpos = inst:GetPosition()
            inst.sg.statemem.targetpos = pos

            RemovePhysicsColliders(inst)
            inst.components.health:SetInvincible(true)
        end,

        onexit = function(inst)
        --This shouldn't actually be reached
            ChangeToCharacterPhysics(inst)
            inst.components.locomotor:Stop()
            inst.components.locomotor:EnableGroundSpeedMultiplier(true)
            inst.components.health:SetInvincible(false)
        end,

        timeline =
        {
            TimeEvent(7*FRAMES, function(inst)
                inst:ForceFacePoint(inst.sg.statemem.targetpos:Get())
                local dist = inst:GetPosition():Dist(inst.sg.statemem.targetpos)
                local speed = dist / (18/30)
                inst.Physics:SetMotorVelOverride(1 * speed, 0, 0)
            end),
        },

        events =
        {
            EventHandler("animover", function(inst)
                inst.Transform:SetPosition(inst.sg.statemem.targetpos:Get())
                inst.Physics:Stop()
                inst.components.health:SetInvincible(false)
                inst.sg:GoToState("jumpoffboatland")
            end),
        },
    },

    State{
        name = "jumpoffboatland",
        tags = {"doing", "busy", "canrotate"},

        onenter = function(inst)
            inst.components.health:SetInvincible(true)
            inst.Physics:Stop()
            inst.AnimState:PushAnimation("land", false)
            PlayFootstep(inst)

        end,

        onexit = function(inst)
            inst.components.health:SetInvincible(false)
        end,

        events =
        {
            EventHandler("animqueueover", function(inst)
                inst:PerformBufferedAction()

                --Try to reset Prince's acknowlegement of light!
                inst.LightWatcher:SetLightThresh(.075)
                inst.LightWatcher:SetDarkThresh(.05)

                local princess = GetPlayer()

                if princess.prefab == "princess" then
                --if inst.components.driver and inst.components.driver.vehicle then
                    princess:PushEvent("princeboating", {inst = inst})--This is to tell Princess' HUD to know what to do with Prince's boat HP bar!
                --end
                end

                --Failsafe in case Prince lands on water instead of land!
                if inst:GetIsOnWater() then
                    inst.Transform:SetPosition(GetPlayer():GetPosition():Get())
                    local fx = SpawnPrefab("statue_transition_2")
                    if fx then
                        fx.Transform:SetPosition(inst:GetPosition():Get())
                        fx.AnimState:SetScale(2,3,2)
                    end
                end

                inst.sg:GoToState("idle")
            end),
        },
    },


    State{
        name = "curepoison",
        tags ={"busy"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("quick_eat")

        end,

        timeline=
        {
            TimeEvent(12*FRAMES, function(inst)
                inst:PerformBufferedAction()
                inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/player_drink")
                inst.sg:RemoveStateTag("busy")
            end),
        },

        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("celebrate") end),
        },
    },


   State{
        name = "celebrate",
        tags ={"idle"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("research")
        end,

        timeline = 
        {
            TimeEvent( 8*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/antivenom_whoosh") end),
            TimeEvent(13*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/heelclick") end),
            TimeEvent(21*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/heelclick") end),
        },

        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },
    },


    State{
        name = "yawn",
        tags = { "busy", "yawn", "pausepredict" },

        onenter = function(inst, data)
            ForceStopHeavyLifting(inst)
            inst.components.locomotor:Stop()
            inst:ClearBufferedAction()

            if data ~= nil and
                data.grogginess ~= nil and
                data.grogginess > 0 and
                inst.components.grogginess ~= nil then
                --Because we have the yawn state tag, we will not get
                --knocked out no matter what our grogginess level is.
                inst.sg.statemem.groggy = true
                inst.sg.statemem.knockoutduration = data.knockoutduration
                inst.components.grogginess:AddGrogginess(data.grogginess, data.knockoutduration)
            end

            inst.AnimState:PlayAnimation("yawn")
        end,

        timeline =
        {
            TimeEvent(.1, function(inst)
                local mount = inst.components.rider:GetMount()
                if mount ~= nil and mount.sounds ~= nil and mount.sounds.yell ~= nil then
                    inst.SoundEmitter:PlaySound(mount.sounds.yell)
                end
            end),
            --TimeEvent(15 * FRAMES, DoYawnSound), --Prince doesn't have a voice!
        },

        events =
        {
            EventHandler("animover", function(inst)
                if inst.AnimState:AnimDone() then
                    inst.sg:RemoveStateTag("yawn")
                    inst.sg:GoToState("idle")
                end
            end),
        },

        onexit = function(inst)
            if inst.sg.statemem.groggy and
                not inst.sg:HasStateTag("yawn") and
                inst.components.grogginess ~= nil then
                --Add a little grogginess to see if it triggers
                --knock out now that we don't have the yawn tag
                inst.components.grogginess:AddGrogginess(.01, inst.sg.statemem.knockoutduration)
            end
        end,
    },


    State{
        name = "idle",
        tags = {"idle", "canrotate"},
        onenter = function(inst, pushanim)    
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("idle_loop", true)

            --Failsafe in case Prince lands on water instead of land!
                if inst:GetIsOnWater() and not GetPlayer():GetIsOnWater() then
                    inst.Transform:SetPosition(GetPlayer():GetPosition():Get())
                    local fx = SpawnPrefab("statue_transition_2")
                    if fx then
                        fx.Transform:SetPosition(inst:GetPosition():Get())
                        fx.AnimState:SetScale(2,3,2)
                    end
                end

            if IsDLCEnabled(PORKLAND_DLC) then
                if inst.wantstosneeze then
                    inst.sg:GoToState("sneeze")
                end
            end

            local equippedArmor = inst.components.inventory and inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)


            -- if equippedArmor and equippedArmor:HasTag("shell") then
            --     inst.sg:GoToState("shell_enter")
            --     return
            -- end
            
    --         local equippedHat = inst.components.inventory and inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
    --         if equippedHat and equippedHat:HasTag("hide") then
                -- inst.sg:GoToState("hide")
                -- return
    --         end
            

            if equippedArmor and equippedArmor:HasTag("band") then
                inst.sg:GoToState("enter_onemanband")
                return
            end

            local anims = {}
            
            local anim = "idle_loop"
            
            if not inst.components.sanity:IsSane() then
                    table.insert(anims, "idle_sanity_pre")
                    table.insert(anims, "idle_sanity_loop")
                elseif inst.components.temperature:IsFreezing() then
                    table.insert(anims, "idle_shiver_pre")
                    table.insert(anims, "idle_shiver_loop")
                elseif (IsDLCEnabled(REIGN_OF_GIANTS) and inst.components.temperature:IsOverheating()) or (IsDLCEnabled(CAPY_DLC) and inst.components.temperature:IsOverheating()) or (IsDLCEnabled(PORKLAND_DLC) and inst.components.temperature:IsOverheating()) then
                    table.insert(anims, "idle_hot_pre")
                    table.insert(anims, "idle_hot_loop")
                elseif inst:HasTag("groggy") then
                    table.insert(anims, "idle_groggy_pre")
                    table.insert(anims, "idle_groggy")            
                else
                    table.insert(anims, "idle_loop")
                end
            
            if pushanim then
                for k,v in pairs (anims) do
                    inst.AnimState:PushAnimation(v, k == #anims)
                end
            else
                inst.AnimState:PlayAnimation(anims[1], #anims == 1)
                for k,v in pairs (anims) do
                    if k > 1 then
                        inst.AnimState:PushAnimation(v, k == #anims)
                    end
                end
            end
            
            inst.sg:SetTimeout(math.random()*4+2)
        end,
        
        ontimeout= function(inst)
        local body = inst.components.inventory and inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
        local head = inst.components.inventory and inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)

            if body and body.prefab == "armorsnurtleshell" then 
                inst.sg:GoToState("shell_enter")
            elseif head and head.prefab == "bushhat" then
                inst.sg:GoToState("hide")
            else
                inst.sg:GoToState("funnyidle")
            end

        end,
    },

    State{
        
        name = "funnyidle",
        tags = {"idle", "canrotate"},
        onenter = function(inst)
        
            if (IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC)) then
                if inst.components.poisonable:IsPoisoned() then
                    inst.AnimState:PlayAnimation("idle_poison_pre")
                    inst.AnimState:PushAnimation("idle_poison_loop")
                    inst.AnimState:PushAnimation("idle_poison_pst", false)
                end
            end
            
            if inst.components.temperature:GetCurrent() < 5 then
                inst.AnimState:PlayAnimation("idle_shiver_pre")
                inst.AnimState:PushAnimation("idle_shiver_loop")
                inst.AnimState:PushAnimation("idle_shiver_pst", false)
            elseif inst.components.temperature:GetCurrent() > 60 and (IsDLCEnabled(REIGN_OF_GIANTS) or IsDLCEnabled(CAPY_DLC) or IsDLCEnabled(PORKLAND_DLC)) then
                inst.AnimState:PlayAnimation("idle_hot_pre")
                inst.AnimState:PushAnimation("idle_hot_loop")
                inst.AnimState:PushAnimation("idle_hot_pst", false)
            elseif inst.components.hunger:GetPercent() < TUNING.HUNGRY_THRESH then
                inst.AnimState:PlayAnimation("hungry")
                inst.SoundEmitter:PlaySound("dontstarve/wilson/hungry")    
            elseif inst.components.sanity:GetPercent() < .5 then
                inst.AnimState:PlayAnimation("idle_inaction_sanity")
            elseif inst:HasTag("groggy") then
                inst.AnimState:PlayAnimation("idle_groggy01_pre")
                inst.AnimState:PushAnimation("idle_groggy01_loop")
                inst.AnimState:PushAnimation("idle_groggy01_pst", false)                
            else
                inst.AnimState:PlayAnimation("idle_inaction")
            end
        end,

        events=
        {
            EventHandler("animqueueover", function(inst) inst.sg:GoToState("idle") end ),
        },
        
    },

    State{
        name = "hide",
        tags = {"idle", "hiding"},
        onenter = function(inst)
            
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("hide")
            inst.SoundEmitter:PlaySound("dontstarve/movement/foley/hidebush")
            inst:AddTag("notarget")
        end,
        
        onexit = function(inst)
            inst:RemoveTag("notarget")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("hide_idle") end),
        },
    },

    State{
        name = "hide_idle",
        tags = {"idle", "hiding"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("hide_idle", true)
            inst:AddTag("notarget")
        end,
        
        onexit = function(inst)
            inst:RemoveTag("notarget")
        end,
        
    },

    State{
        name = "shell_enter",
        tags = {"idle", "hiding", "shell"},
        onenter = function(inst)            
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("hideshell")
            inst:PushEvent("disengage")
            --local equippedArmor = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            --if equippedArmor then
                --equippedArmor.components.armor:SetAbsorption(TUNING.FULL_ABSORPTION)
            --end
        end,
        timeline =
        {
            TimeEvent(6*FRAMES, function(inst)
                inst.SoundEmitter:PlaySound("dontstarve/movement/foley/hideshell")    
            end),
        },        
        
        onexit = function(inst)

        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("shell_idle") end),
        },
    },

    State{
        name = "shell_idle",
        tags = {"idle", "hiding", "shell"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("hideshell_idle", true)
        end,
        
        onexit = function(inst)
        end,
        
    },

    State{
        name = "shell_hit",
        tags = {"busy", "hiding", "shell"},
        
        onenter = function(inst)
            inst:ClearBufferedAction()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/hit")        
            inst.AnimState:PlayAnimation("hitshell")
            --local sound_name = inst.soundsname or inst.prefab
            --local sound_event = "dontstarve/characters/"..sound_name.."/hurt"
            --inst.SoundEmitter:PlaySound(sound_event)
        end,
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("shell_idle") 
            end ),
        }, 
        
        timeline =
        {
            TimeEvent(3*FRAMES, function(inst)
                --inst.sg:RemoveStateTag("busy")
            end),
        },        
               
    },


    State{
        name = "enter_onemanband",
        tags = {"doing", "playing", "idle"},

        onenter = function(inst)
        inst.components.locomotor:Stop()
        inst.AnimState:PlayAnimation("idle_onemanband1_pre")
        inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")
        end,

        onexit = function(inst)
        end,

        events = 
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("play_onemanband") end),
        },
    },

    State{
        name = "play_onemanband",
        tags = {"doing", "playing", "idle"},

        onenter = function(inst)

            inst.components.locomotor:Stop()
            --inst.AnimState:PlayAnimation("idle_onemanband1_pre")
            inst.AnimState:PlayAnimation("idle_onemanband1_loop")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")
        end,

        onexit = function(inst)
        end,

        events =
        {
            EventHandler("animover", function(inst)
                if math.random() <= 0.15 then
                    inst.sg:GoToState("play_onemanband_stomp") -- go into stomp
                else
                    inst.sg:GoToState("play_onemanband")-- loop state again
                end
            end),
        },
    },

    State{
        name = "play_onemanband_stomp",
        tags = {"doing", "playing", "idle"},

        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("idle_onemanband1_pst")
            inst.AnimState:PushAnimation("idle_onemanband2_pre")
            inst.AnimState:PushAnimation("idle_onemanband2_loop")
            inst.AnimState:PushAnimation("idle_onemanband2_pst", false)  
            inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband") 
        end,

        onexit = function(inst)
        end,

        timeline=
        {
            TimeEvent(20*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),

            TimeEvent(25*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),
            
            TimeEvent(30*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),

            TimeEvent(35*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),
        },

        events = 
        {
            EventHandler("animqueueover", function(inst)
                inst.sg:GoToState("idle") 
            end),
        },
    },


    State{
        name = "sneeze",
        tags = {"busy","sneeze"},
        
        onenter = function(inst)
            inst.wantstosneeze = false
            inst.SoundEmitter:PlaySound("dontstarve/wilson/hit",nil,.02)
            inst.AnimState:PlayAnimation("sneeze")
            inst.SoundEmitter:PlaySound("dontstarve_DLC003/characters/sneeze")
            inst:ClearBufferedAction()
            
            inst.components.locomotor:Stop()  

            --inst.components.talker:Say(GetString(inst.prefab, "ANNOUNCE_SNEEZE"))        
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        }, 
        
        timeline =
        {
            TimeEvent(10*FRAMES, function(inst)
                if inst.components.hayfever then
                    inst.components.hayfever:DoSneezeEffects()
                end
                inst.sg:RemoveStateTag("busy")
            end),
        },        
               
    },


    State{
        name = "wakeup",
        tags = {"waking_up", "busy"},

        onenter = function(inst)
            inst.AnimState:PlayAnimation("wakeup")
            inst.components.health:SetInvincible(true)
        end,
        
        onexit = function(inst)
            inst.components.health:SetInvincible(false)
        end,
        
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },

    },

    State{
        name = "sleep",
        tags = {"asleep", "idle", "busy", "nointerrupt"},
        
        onenter = function(inst)
            if inst:HasTag("teleporting") then
                inst:RemoveTag("teleporting")
            end
            if inst:HasTag("registered") then
                inst:RemoveTag("registered")
            end

            inst.AnimState:PlayAnimation("sleep")
            inst.components.health:SetInvincible(true)
            inst.components.talker:IgnoreAll()
            --This is to prevent him from dying from lightning strikes from touch stones!
            inst:DoTaskInTime(4.5, function() 
                inst.components.burnable:Extinguish()
                inst.components.health:SetInvincible(false) 
            end)
        end,

        onexit=function(inst)
            inst.components.talker:StopIgnoringAll()
            --inst.components.health:SetInvincible(false)
        end,

    },

    State{
        name = "run_start",
        tags = {"moving", "running", "canrotate"},
        
        onenter = function(inst)
            ConfigureRunState(inst)
			inst.components.locomotor:RunForward()

            if IsDLCEnabled(PORKLAND_DLC) then
                inst.AnimState:PlayAnimation(GetRunStateAnim(inst).."_pre")
            else
                inst.AnimState:PlayAnimation("run_pre")
            end

            inst.sg.mem.foosteps = 0
        end,

        onupdate = function(inst)
            inst.components.locomotor:RunForward()
        end,

        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("run") end ),        
        },
        
        timeline=
        {        
            TimeEvent(4*FRAMES, function(inst)
                PlayFootstep(inst)
                DoFoleySounds(inst)
            end),
        },        
        
    },

    State{
        name = "run",
        tags = {"moving", "running", "canrotate"},
        
        onenter = function(inst) 
            ConfigureRunState(inst)
            inst.components.locomotor:RunForward()

            if IsDLCEnabled(PORKLAND_DLC) then
                local anim = GetRunStateAnim(inst)
                if anim == "run" then
                    if inst.prefab == "wilba" and inst.were then
                        anim = "run_werewilba_loop"
                    else
                        anim = "run_loop"
                    end
                end        
                inst.AnimState:PlayAnimation(anim)
            end

            if not IsDLCEnabled(PORKLAND_DLC) then
                inst.AnimState:PlayAnimation("run_loop")
            end
            
        end,
        
        onupdate = function(inst)
            inst.components.locomotor:RunForward()
        end,

        timeline=
        {
            --groggy
            TimeEvent(1 * FRAMES, function(inst)
                if inst.sg.statemem.groggy and IsDLCEnabled(PORKLAND_DLC) then
                    DoRunSounds(inst)
                    DoFoleySounds(inst)
                end
            end),
            TimeEvent(12 * FRAMES, function(inst)
                if inst.sg.statemem.groggy and IsDLCEnabled(PORKLAND_DLC) then
                    DoRunSounds(inst)
                    DoFoleySounds(inst)
                end
            end),

            TimeEvent(7*FRAMES, function(inst)
                if not inst.sg.statemem.groggy then
				    inst.sg.mem.foosteps = inst.sg.mem.foosteps + 1
                    PlayFootstep(inst, inst.sg.mem.foosteps < 5 and 1 or .6)
                    DoFoleySounds(inst)
                end
            end),
            TimeEvent(15*FRAMES, function(inst)
                if not inst.sg.statemem.groggy then
				    inst.sg.mem.foosteps = inst.sg.mem.foosteps + 1
                    PlayFootstep(inst, inst.sg.mem.foosteps < 5 and 1 or .6)
                    DoFoleySounds(inst)
                end
            end),
        },
        
        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("run") end ),        
        },
    },
    
    State{
    
        name = "run_stop",
        tags = {"canrotate", "idle"},
        
        onenter = function(inst) 
            ConfigureRunState(inst)
            inst.Physics:Stop()

            if IsDLCEnabled(PORKLAND_DLC) then
                inst.AnimState:PlayAnimation(GetRunStateAnim(inst).."_pst")
            else
                inst.AnimState:PlayAnimation("run_pst")
            end
        
        end,
        
        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),        
        },
        
    },

    State{
        name = "teleportato_teleport",
        tags = {"busy", "nointerrupt", "teleporting"},
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            if not inst:HasTag("teleporting") then
                inst:AddTag("teleporting")
            end
            inst.components.health:SetInvincible(true)
            inst.AnimState:PlayAnimation("teleport")
        end,

        onexit = function(inst)
            inst.components.health:SetInvincible(false)
            --SaveGameIndex:GetSaveFollowers(GetPlayer()) 
        end,

        timeline = {
            TimeEvent(0, function(inst)
                inst.SoundEmitter:PlaySound("dontstarve/common/teleportato/teleportato_pulled")
            end),
            --TimeEvent(20, function(inst)
                --This is here because Prince's cloak's sprite acts funny during this animation!
                --inst.AnimState:ClearOverrideSymbol("swap_body")
            --end),
            TimeEvent(82*FRAMES, function(inst)
                inst.SoundEmitter:PlaySound("dontstarve/common/teleportato/teleportato_under")
            end),
        },

        events=
        {   
            EventHandler("animover", function(inst) 
                --Get Prince to follow Princess through the Wooden Thing!
                --SaveGameIndex:GetSaveFollowers(GetPlayer()) 
            end ),        
        },
    },

    State{
        name = "banished",
        tags = {"busy", "nointerrupt", "teleporting", "banished"},
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            if not inst:HasTag("teleporting") then
                inst:AddTag("teleporting")
            end

            if inst.HasWarped then
                inst.HasWarped = false
            end

            inst.components.container.canbeopened = false
            inst.components.container:Close()
            local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if handslot then
                inst.components.inventory:RemoveItem(handslot)--:Remove()  
                inst.components.container:GiveItem(handslot)
            end

            local bodyslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            if bodyslot then
                inst.components.inventory:RemoveItem(bodyslot)--:Remove()  
                inst.components.container:GiveItem(bodyslot)
            end

            local headslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
            if headslot then
                inst.components.inventory:RemoveItem(headslot)--:Remove()  
                inst.components.container:GiveItem(headslot)
            end

            inst.components.container:DropEverything()


            inst.components.health:SetInvincible(true)
            inst.AnimState:PlayAnimation("teleport")
        end,

        onexit = function(inst)
            inst.components.health:SetInvincible(false)
            --SaveGameIndex:GetSaveFollowers(GetPlayer()) 
        end,

        timeline = {
            TimeEvent(0, function(inst)
                inst.SoundEmitter:PlaySound("dontstarve/common/teleportato/teleportato_pulled")
            end),
            --TimeEvent(30, function(inst)
                --This is here because Prince's cloak's sprite acts funny during this animation!
                --inst.AnimState:ClearOverrideSymbol("swap_body")
            --end),
            TimeEvent(82*FRAMES, function(inst)
                inst.SoundEmitter:PlaySound("dontstarve/common/teleportato/teleportato_under")
            end),
        },

        events=
        {   
            EventHandler("animover", function(inst) 
                inst:Remove()
            end ),        
        },
    },


    State{
        name = "attack",
        tags = {"attack", "notalking", "abouttoattack", "busy"},
        
        onenter = function(inst)
            --print(debugstack())
            inst.sg.statemem.target = inst.components.combat.target
            inst.components.combat:StartAttack()
            inst.components.locomotor:Stop()
            local weapon = inst.components.combat:GetWeapon()
            local otherequipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)

            if weapon then
                inst.AnimState:PlayAnimation("atk")
                if weapon:HasTag("icestaff") then
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_icestaff")
                elseif weapon:HasTag("shadow") then
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_nightsword")
                elseif weapon:HasTag("firestaff") then
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_firestaff")
                elseif weapon:HasTag("halberd") then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC003/common/items/weapon/halberd")
                elseif weapon:HasTag("cutlass") then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/swordfish_sword")
                elseif weapon:HasTag("pegleg") then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/pegleg_weapon")  
                elseif weapon:HasTag("corkbat") then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC003/common/items/weapon/corkbat")  
                else
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_weapon")
                end
            elseif otherequipped and (otherequipped:HasTag("light") or otherequipped:HasTag("nopunch")) then
                inst.AnimState:PlayAnimation("atk")

                inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_weapon") 
            else
                inst.sg.statemem.slow = true
                inst.AnimState:PlayAnimation("punch")
                inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_whoosh",nil,.5)
            end
            
            if inst.components.combat.target then
                inst.components.combat:BattleCry()
                if inst.components.combat.target and inst.components.combat.target:IsValid() then
                    inst:FacePoint(Point(inst.components.combat.target.Transform:GetWorldPosition()))
                end
            end
            
        end,
        
        timeline=
        {
            TimeEvent(8*FRAMES, function(inst) 
                inst.components.combat:DoAttack(inst.sg.statemem.target) 
                inst.sg:RemoveStateTag("abouttoattack") 
                local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                if handslot and handslot.components.finiteuses then
                    if handslot.components.finiteuses.current <= 0 then
                        inst.components.inventory:RemoveItem(handslot)--:Remove()  
                        inst.components.container:GiveItem(handslot)
                        --inst.AnimState:Hide("ARM_carry") 
                        --inst.AnimState:Show("ARM_normal") 
                    end
                end
            end),
            TimeEvent(12*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("busy")
            end),               
            TimeEvent(13*FRAMES, function(inst)
                if not inst.sg.statemem.slow then
                    inst.sg:RemoveStateTag("attack")
                end
            end),
            TimeEvent(24*FRAMES, function(inst)
                if inst.sg.statemem.slow then
                    inst.sg:RemoveStateTag("attack")
                end
            end),
            
            
        },

        ontimeout = function(inst)
            inst.sg:RemoveStateTag("attack")
            --inst.sg:AddStateTag("idle")
        end,
        events=
        {
            EventHandler("animqueueover", function(inst)
                inst.sg:GoToState("idle")
            end ),
        },
    }, 

    State{
        name = "death",
        tags = {"busy", "nointerrupt", "dead"},
        
        onenter = function(inst)
            inst.Physics:Stop()
            --For his equipped gear!
            local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if handslot then
                inst.components.inventory:RemoveItem(handslot)
                inst.components.inventory:DropItem(handslot, true, true)
            end
            local bodyslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            if bodyslot then
                if bodyslot.prefab ~= "amulet" then
                    inst.components.inventory:RemoveItem(bodyslot)
                    inst.components.inventory:DropItem(bodyslot, true, true)
                end
            end
            local headslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
            if headslot then
                inst.components.inventory:RemoveItem(headslot)
                inst.components.inventory:DropItem(headslot, true, true)
            end
            --inst.components.inventory:DropEverything()
            --For the rest of his contents!
            inst.components.container:DropEverything()
            inst.AnimState:Hide("swap_arm_carry")
            inst.AnimState:PlayAnimation("death")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/death")
        end,

        events =
        {
            EventHandler("animover", function(inst)
                inst:DoTaskInTime(1, function() 
                    --inst:Remove()
                end)
            end ),
        },
    },  
   
    State{
        name = "hit",
        tags = {"busy"},
        
        onenter = function(inst)
            inst:ClearBufferedAction()
            inst.components.container:Close()
            inst.AnimState:PlayAnimation("hit")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/hit")
            inst.Physics:Stop()            
        end,
        
        events=
        {
            EventHandler("animover", function(inst) 

                --If Prince is not stunned by Queen Womant, then just idle afterwards.
                if not inst.sanity_stunned then
                    inst.sg:GoToState("idle")
                --If Queen Womant stunned him, get stunned!
                elseif inst.sanity_stunned then
                    inst.sg:GoToState("sanity_stun")
                end
            end),
        }, 
        
        timeline =
        {
            TimeEvent(3*FRAMES, function(inst)
                inst.sg:RemoveStateTag("busy")
            end),
        },               
    },


    State{
        name = "frozen",
        tags = {"busy", "frozen"},
        
        onenter = function(inst)
            if inst.components.locomotor then
                inst.components.locomotor:StopMoving()
            end
            inst.AnimState:PlayAnimation("frozen")
            inst.SoundEmitter:PlaySound("dontstarve/common/freezecreature")
            inst.AnimState:OverrideSymbol("swap_frozen", "frozen", "frozen")
        end,
        
        onexit = function(inst)
            inst.AnimState:ClearOverrideSymbol("swap_frozen")
        end,
        
        events=
        {   
            EventHandler("onthaw", function(inst) inst.sg:GoToState("thaw") end ),        
        },
    },

    State{
        name = "thaw",
        tags = {"busy", "thawing"},
        
        onenter = function(inst) 
            if inst.components.locomotor then
                inst.components.locomotor:StopMoving()
            end
            inst.AnimState:PlayAnimation("frozen_loop_pst", true)
            inst.SoundEmitter:PlaySound("dontstarve/common/freezethaw", "thawing")
            inst.AnimState:OverrideSymbol("swap_frozen", "frozen", "frozen")
        end,
        
        onexit = function(inst)
            inst.SoundEmitter:KillSound("thawing")
            inst.AnimState:ClearOverrideSymbol("swap_frozen")
        end,

        events =
        {   
            EventHandler("unfreeze", function(inst)
                if inst.sg.sg.states.hit then
                    inst.sg:GoToState("hit")
                else
                    inst.sg:GoToState("idle")
                end
            end),
        },
    },


    State{
        name = "electrocute",
        tags = {"busy"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("shock")
            inst.AnimState:SetBloomEffectHandle("shaders/anim.ksh")
            inst.fx = SpawnPrefab("shock_fx")

            inst.fx.Transform:SetRotation(inst.Transform:GetRotation())

            local pos = inst:GetPosition()
            inst.fx.Transform:SetPosition(pos.x, pos.y, pos.z)

            --inst.Light:Enable(true)
        end,

        onexit = function(inst)
            --inst.Light:Enable(false)
            inst.AnimState:ClearBloomEffectHandle()
            inst.fx:Remove()
        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("electrocute_pst")
            end),
        },
    },
    State{
        name = "electrocute_pst",
        tags = {"busy"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("shock_pst")
        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle") 
            end),
        },
    },


        State{ name = "chop_start",
        tags = {"prechop", "chopping", "working", "wantstowork"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("chop_pre")

        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("chop") end),
        },
    },
    
    State{
        name = "chop",
        tags = {"prechop", "chopping", "working", "midchop", "wantstowork"},
        onenter = function(inst)
            inst.sg.statemem.action = inst:GetBufferedAction()
            inst.AnimState:PlayAnimation("chop_loop")        
        end,

        timeline=
        {
            TimeEvent(5*FRAMES, function(inst) 
                    inst:PerformBufferedAction() 
            end),

            TimeEvent(9*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("prechop")
                inst.sg:RemoveStateTag("chopping")
                inst.sg:RemoveStateTag("working")
            end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("idle")
            end ),            
        },        
    },

    State{ 
        name = "mine_start",
        tags = {"premine", "working", "wantstowork"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("pickaxe_pre")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("mine") end),
        },
    },
    
    State{
        name = "mine",
        tags = {"premine", "mining", "working", "midmine", "wantstowork"},
        onenter = function(inst)
            inst.sg.statemem.action = inst:GetBufferedAction()
            inst.AnimState:PlayAnimation("pickaxe_loop")
        end,

        timeline=
        {
            TimeEvent(9*FRAMES, function(inst) 
                if inst.sg.statemem.action and inst.sg.statemem.action.target then
                    local fx = SpawnPrefab("mining_fx")
                    fx.Transform:SetPosition(inst.sg.statemem.action.target:GetPosition():Get())
                end
                inst:PerformBufferedAction() 
                inst.sg:RemoveStateTag("premine") 
                if inst.sg.statemem.action and inst.sg.statemem.action.target and inst.sg.statemem.action.target.prefab == "rock_ice" then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC001/common/iceboulder_hit")
                else
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/use_pick_rock")
                end

            end),

            TimeEvent(11*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("mining")
                inst.sg:RemoveStateTag("working")
            end),
            
            -- TimeEvent(14*FRAMES, function(inst)
            --     if  inst.sg.statemem.action and 
            --         inst.sg.statemem.action.target and 
            --         inst.sg.statemem.action.target:IsActionValid(inst.sg.statemem.action.action) and 
            --         inst.sg.statemem.action.target.components.workable then
            --             inst:ClearBufferedAction()
            --             inst:PushBufferedAction(inst.sg.statemem.action)
            --     end
            -- end),  
            --TimeEvent(14*FRAMES, function(inst) 
                --inst.sg:RemoveStateTag("working")
            --end),          
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.AnimState:PlayAnimation("pickaxe_pst") 
                inst.sg:GoToState("idle", true)
            end ),            
        },        
    },

State{ name = "dig_start",
        tags = {"predig", "working", "wantstowork"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("shovel_pre")
        end,
        
        events=
        {
            EventHandler("unequip", function(inst) inst.sg:GoToState("idle") end ),
            EventHandler("animover", function(inst) inst.sg:GoToState("dig") end),
        },
    },
    
    State{
        name = "dig",
        tags = {"predig", "digging", "working", "middig", "wantstowork"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("shovel_loop")
            inst.sg.statemem.action = inst:GetBufferedAction()
        end,

        timeline=
        {
            TimeEvent(15*FRAMES, function(inst) 
--[[                if inst.sg.statemem.action and inst.sg.statemem.action.target then
                    local fx = SpawnPrefab("shovel_dirt")
                    fx.Transform:SetPosition( inst.sg.statemem.action.target.Transform:GetWorldPosition() )
                end
--]]                
                inst:PerformBufferedAction() 
                inst.sg:RemoveStateTag("predig") 
                inst.SoundEmitter:PlaySound("dontstarve/wilson/dig")
                
            end),

            TimeEvent(32*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("digging") 
                inst.sg:RemoveStateTag("working") 
            end),
            
            TimeEvent(35*FRAMES, function(inst)
                if  inst.sg.statemem.action and 
                    inst.sg.statemem.action.target and 
                    inst.sg.statemem.action.target:IsActionValid(inst.sg.statemem.action.action, true) and
                    inst.sg.statemem.action.target.components.workable then
                        inst:ClearBufferedAction()
                        inst:PushBufferedAction(inst.sg.statemem.action)
                end
            end),
            
        },
        
        events=
        {
            --EventHandler("unequip", function(inst) inst.sg:GoToState("idle") end ),
            EventHandler("animover", function(inst) 
                inst.AnimState:PlayAnimation("shovel_pst") 
                inst.sg:GoToState("idle", true)
            end ),
            
        },        
    },      


    State{ 
        name = "hack_start",
        tags = {"prehack", "working", "wantstowork"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("chop_pre")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("hack") end),
        },
    },
    
    State{
        name = "hack",
        tags = {"prehack", "hacking", "working", "midhack", "wantstowork"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("chop_loop")
        end,

        timeline=
        {
            TimeEvent(9*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),

            TimeEvent(11*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
                inst.sg:RemoveStateTag("prehack") 
                inst.sg:RemoveStateTag("hacking") 
                inst.sg:RemoveStateTag("working") 
            end),
        },
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.AnimState:PlayAnimation("chop_pst") 
                inst.sg:GoToState("idle", true)
            end ),            
        },        
    },

    State{ name = "fishing_pre",
        tags = {"prefish", "fishing"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:Show("RIPPLES")
            inst.AnimState:PlayAnimation("fishing_pre")
        end,
        
        timeline = 
        {
            TimeEvent(13*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_cast") end),
            TimeEvent(15*FRAMES, function(inst) inst:PerformBufferedAction() end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_baitsplash")
                inst.sg:GoToState("fishing")
            end ),
        },        
    },
    

    State{
        name = "fishing",
        tags = {"fishing"},
        onenter = function(inst, pushanim)
            if pushanim then
                if type(pushanim) == "string" then
                    inst.AnimState:PlayAnimation(pushanim)
                end
                inst.AnimState:PushAnimation("fishing_idle", true)
            else
                inst.AnimState:PlayAnimation("fishing_idle", true)
            end
            inst.sg:SetTimeout(60 + math.random())
            local equippedTool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if equippedTool and equippedTool.components.fishingrod then
                equippedTool.components.fishingrod:WaitForFish()
            end

        end,

        --Quit fishing if nothing is happening.
        ontimeout = function(inst)
            local fishingrod = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if fishingrod and fishingrod.components.fishingrod then
                fishingrod.components.fishingrod:StopFishing()
            end
            inst.sg:GoToState("fishing_pst")
        end,
        
        events=
        {
            EventHandler("fishingnibble", function(inst) inst.sg:GoToState("fishing_nibble") end ),
        },
    },
    
    State{ name = "fishing_pst",
        tags = {"fishing"}, --Tag is mainly there so Prince doesn't interrupt himself if Princess goes back to fishing.
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("fishing_pst")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },
    
    State{
        name = "fishing_nibble",
        tags = {"fishing", "nibble"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("bite_light_pre")
            inst.AnimState:PushAnimation("bite_light_loop", true)
            inst.sg:SetTimeout(1 + math.random())
            inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishinwater", "splash")
            inst:DoTaskInTime(0.25 + math.random(), function()
                local fishingrod = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                --Code ripped from ACTIONS.REEL because there's an oversight I'm too impatient to fix.
                if fishingrod and fishingrod.components.fishingrod and fishingrod.components.fishingrod:IsFishing() then
                    if fishingrod.components.fishingrod:HasHookedFish() then
                        fishingrod.components.fishingrod:Reel()
                    elseif fishingrod.components.fishingrod:FishIsBiting() then
                        fishingrod.components.fishingrod:Hook()
                    else
                    fishingrod.components.fishingrod:StopFishing()
                    end
                end
            end)
        end,
        
        onexit = function(inst)
            inst.SoundEmitter:KillSound("splash")
        end,
        
        ontimeout = function(inst)
            inst.sg:GoToState("fishing", "bite_light_pst")
        end,
        
        events = 
        {
            EventHandler("fishingstrain", function(inst) inst.sg:GoToState("fishing_strain") end),
        },
    }, 
    
    State{
        name = "fishing_strain",
        tags = {"fishing"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("bite_heavy_pre")
            inst.AnimState:PushAnimation("bite_heavy_loop", true)
            inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishinwater", "splash")
            inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_strain", "strain")
            inst:DoTaskInTime(0.25 + math.random(), function()
                local fishingrod = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                --Code ripped from ACTIONS.REEL because there's an oversight I'm too impatient to fix.
                if fishingrod and fishingrod.components.fishingrod and fishingrod.components.fishingrod:IsFishing() then
                    if fishingrod.components.fishingrod:HasHookedFish() then
                        fishingrod.components.fishingrod:Reel()
                    elseif fishingrod.components.fishingrod:FishIsBiting() then
                        fishingrod.components.fishingrod:Hook()
                    else
                    fishingrod.components.fishingrod:StopFishing()
                    end
                end
            end)
        end,
        
        onexit = function(inst)
            inst.SoundEmitter:KillSound("splash")
            inst.SoundEmitter:KillSound("strain")
        end,
        
        events = 
        {
            EventHandler("fishingcatch", function(inst, data)
                inst.sg:GoToState("catchfish", data.build)
            end),
            EventHandler("fishingloserod", function(inst)
                inst.sg:GoToState("loserod")
            end),

        },
    },
    
    State{
        name = "catchfish",
        tags = {"fishing", "catchfish", "busy"},
        onenter = function(inst, build)
            inst.AnimState:PlayAnimation("fish_catch")
           -- print("Using ", build, " to swap out fish01")
            inst.AnimState:OverrideSymbol("fish01", build, "fish01")
            
            -- inst.AnimState:OverrideSymbol("fish_body", build, "fish_body")
            -- inst.AnimState:OverrideSymbol("fish_eye", build, "fish_eye")
            -- inst.AnimState:OverrideSymbol("fish_fin", build, "fish_fin")
            -- inst.AnimState:OverrideSymbol("fish_head", build, "fish_head")
            -- inst.AnimState:OverrideSymbol("fish_mouth", build, "fish_mouth")
            -- inst.AnimState:OverrideSymbol("fish_tail", build, "fish_tail")
        end,
        
        onexit = function(inst)
            inst.AnimState:ClearOverrideSymbol("fish01")
            -- inst.AnimState:ClearOverrideSymbol("fish_body")
            -- inst.AnimState:ClearOverrideSymbol("fish_eye")
            -- inst.AnimState:ClearOverrideSymbol("fish_fin")
            -- inst.AnimState:ClearOverrideSymbol("fish_head")
            -- inst.AnimState:ClearOverrideSymbol("fish_mouth")
            -- inst.AnimState:ClearOverrideSymbol("fish_tail")
        end,

        timeline=
        {
            TimeEvent(8*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishcaught") end),
            TimeEvent(10*FRAMES, function(inst)
                --This is mainly blanked out so Prince doesn't interrupt himself if Princess goes back to fishing.
                --inst.sg:RemoveStateTag("fishing") 
            end),
            TimeEvent(23*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishland") end),
            TimeEvent(24*FRAMES, function(inst)
                local equippedTool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                if equippedTool and equippedTool.components.fishingrod then
                    equippedTool.components.fishingrod:Collect()
                end
            end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("idle")
            end ),
        },        
    },       
    
    State{
        name = "loserod",
        tags = {"busy"},
        onenter = function(inst)
            local equippedTool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if equippedTool and equippedTool.components.fishingrod then
                equippedTool.components.fishingrod:Release()
                inst.components.inventory:RemoveItem(equippedTool)
                equippedTool:Remove()
            end
            inst.AnimState:PlayAnimation("fish_nocatch")
        end,

        timeline=
        {
            TimeEvent(4*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_lostrod") end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("idle")
            end ),
        },        
    },


    State{
        name = "toolbroke",
        tags = {"busy", "toolbroken"},
        onenter = function(inst, tool)
            inst.components.container:Close()

            --if not inst:HasTag("toolbrokeprince") then
                --inst:AddTag("toolbrokeprince")
                --end

            if inst:HasTag("canwork") then
                --Prevent Prince from attempting to keep working with a broken tool!
                inst:RemoveTag("canwork")
            end

            inst.AnimState:PlayAnimation("hit")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/use_break")
            inst.AnimState:Hide("ARM_carry") 
            inst.AnimState:Show("ARM_normal") 
            local brokentool = SpawnPrefab("brokentool")
            brokentool.Transform:SetPosition(inst.Transform:GetWorldPosition() )
            inst.sg.statemem.tool = tool
        end,
        
        onexit = function(inst)
            local losttool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if losttool then
                inst.components.inventory:RemoveItem(losttool):Remove()  
                --inst.components.container:GiveItem(losttool)
            end

            local sameTool = inst.components.container:FindItem(function(item)
                return item.prefab == inst.sg.statemem.tool.prefab
            end)
            if sameTool then
                inst.components.inventory:Equip(sameTool)
            end

            if inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS) then
                inst.AnimState:Show("ARM_carry") 
                inst.AnimState:Hide("ARM_normal")
            end

        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle")
            end ),
        },
    },


    State{
        name = "armorbroke",
        tags = {"busy"},
        onenter = function(inst, armor)
            inst.AnimState:PlayAnimation("hit")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/use_armour_break")
            --inst.sg.statemem.armor = armor
        end,
        
        onexit = function(inst)
            inst.components.container:Close()
            local container = inst.components.container
            for i = 1, container:GetNumSlots() do
                local item = container:GetItemInSlot(i)

                --This tag is a workaround so that Prince's container can give him the tools he needs, while preventing items from getting lost in his inventory.
                inst:AddTag("armingprince")

                if item then

                    if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.BODY and item.components.armor then
                        local currentbody = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
                            if currentbody and currentbody.components.armor and currentbody.components.armor.condition <= 0 then

                                inst.components.inventory:RemoveItem(currentbody)--:Remove()  
                                inst.components.container:GiveItem(currentbody)
                                currentbody:Remove()
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                inst.components.inventory:Equip(item)
                            end
                            if not currentbody then
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                inst.components.inventory:Equip(item)
                            end
                    end

                    if item.components.equippable and item.components.equippable.equipslot == EQUIPSLOTS.HEAD and item.components.armor then
                        local currenthead = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                            if currenthead and currenthead.components.armor and currenthead.components.armor.condition <= 0 then

                                inst.components.inventory:RemoveItem(currenthead)--:Remove()  
                                inst.components.container:GiveItem(currenthead)
                                currenthead:Remove()
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                inst.components.inventory:Equip(item)
                            end

                            if not currenthead then
                                inst.components.container:RemoveItemBySlot(i)--:Remove()  
                                inst.components.inventory:Equip(item)
                            end
                    end
    
                end

             --When the tag is removed, it should keep the player from accidentally transporting items into his inaccessible inventory.
            inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
            end
        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle")
            end ),
        },
    },



       State{
        name = "eat",
        tags ={"busy", "eating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            
            inst.AnimState:PlayAnimation("eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            
            TimeEvent(28*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),
            
            TimeEvent(30*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("busy")
            end),
            
            TimeEvent(70*FRAMES, function(inst) 
                inst.SoundEmitter:KillSound("eating")    
            end),
            
        },        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },    
    
   State{
        name = "quickeat",
        tags ={"busy", "eating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            inst.AnimState:PlayAnimation("quick_eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            TimeEvent(12*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),
            TimeEvent(14*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("busy")
            end),
        },        
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },    

       State{
        name = "hungryeat",
        tags ={"busy", "eating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            
            inst.AnimState:PlayAnimation("eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            
            TimeEvent(28*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),
            
            TimeEvent(30*FRAMES, function(inst) 
                --inst.sg:RemoveStateTag("busy")
            end),
            
            TimeEvent(70*FRAMES, function(inst) 
                inst.SoundEmitter:KillSound("eating")    
            end),
            
        },        
        events=
        {
            EventHandler("animover", function(inst) 
            if inst.components.hunger and inst.components.hunger:GetPercent() >= 0.80 then
                if inst:HasTag("hungryprince") then
                    inst:RemoveTag("hungryprince")
                end
            end
            inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },    
    
   State{
        name = "hungryquickeat",
        tags ={"busy", "eating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            inst.AnimState:PlayAnimation("quick_eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            TimeEvent(12*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
                --inst.sg:RemoveStateTag("busy")
            end),
            TimeEvent(14*FRAMES, function(inst) 
                --inst.sg:RemoveStateTag("busy")
            end),
        },        
        
        events=
        {
            EventHandler("animover", function(inst)
            if inst.components.hunger and inst.components.hunger:GetPercent() >= 0.80 then
                if inst:HasTag("hungryprince") then
                    inst:RemoveTag("hungryprince")
                end
            end
            inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },    


    State{
        name = "talk",
        tags = {"idle", "talking"},
        
        onenter = function(inst, noanim)
            inst.components.locomotor:Stop()
            if not noanim then
                inst.AnimState:PlayAnimation("dial_loop", true)
            end
            
            local sound_name = inst.soundsname or inst.prefab
            local path = inst.talker_path_override or "dontstarve/characters/"
            local equippedHat = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)

            if equippedHat and equippedHat:HasTag("muffler") then
                inst.SoundEmitter:PlaySound(path..sound_name.."/gasmask_talk", "talk")

            elseif inst.talksoundoverride then
                inst.SoundEmitter:PlaySound(inst.talksoundoverride, "talk")
            else
                inst.SoundEmitter:PlaySound(path..sound_name.."/talk_LP", "talk")
            end

            if inst:HasTag("hasvoiceintensity_health") then
                local percent = inst.components.health:GetPercent()
                inst.SoundEmitter:SetParameter( "talk", "intensity", percent )                
            end
            
            inst.sg:SetTimeout(1.5 + math.random()*.5)
        end,
        
        ontimeout = function(inst)
            inst.SoundEmitter:KillSound("talk")
            inst.sg:GoToState("idle")
            if inst.components.talker.endspeechsound then
                inst.SoundEmitter:PlaySound(inst.components.talker.endspeechsound)
            end               
        end,
        
        onexit = function(inst)
            inst.SoundEmitter:KillSound("talk")
            if inst.components.talker.endspeechsound then
                inst.SoundEmitter:PlaySound(inst.components.talker.endspeechsound)
            end               
        end,
        
        events=
        {
            EventHandler("donetalking", function(inst) inst.sg:GoToState("idle") end),
        },
    }, 


        State{
        name = "amulet_rebirth",
        tags = {"busy"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("amulet_rebirth")
            --inst.AnimState:OverrideSymbol("FX", "player_amulet_resurrect", "FX")
        end,
        
        onexit= function(inst)
            inst.components.hunger:SetPercent(2/3)
            inst.components.health:Respawn(TUNING.RESURRECT_HEALTH)
            
            if inst.components.sanity then
                inst.components.sanity:SetPercent(.5)
            end
            
            local item = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            if item and item.prefab == "amulet" then
                item = inst.components.inventory:RemoveItem(item)
                if item then
                    item:Remove()
                    item.persists = false
                end
            end
            --SaveGameIndex:SaveCurrent()
            --inst.AnimState:ClearOverrideSymbol("FX")
            
        end,
        
        timeline =
        {
            TimeEvent(0*FRAMES, function(inst)
                inst.stafflight = SpawnPrefab("staff_castinglight")
                local pos = inst:GetPosition()
                local colour = {150/255, 46/255, 46/255}
                inst.stafflight.Transform:SetPosition(pos.x, pos.y, pos.z)
                inst.stafflight.setupfn(inst.stafflight, colour, 1.7, 1)           

            end),

            TimeEvent(0, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/rebirth_amulet_raise") end),
            TimeEvent(60*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/rebirth_amulet_poof") end),
        
            TimeEvent(80*FRAMES, function(inst)
                local pos = Vector3(inst.Transform:GetWorldPosition())
                local ents = TheSim:FindEntities(pos.x,pos.y,pos.z, 10)
                for k,v in pairs(ents) do
                    if v ~= inst and v.components.sleeper and v.prefab ~= "princess" then
                        v.components.sleeper:GoToSleep(20)
                    end
                end
                
                
            end),
        },        
                   
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },  

    State{
        name="item_hat",
        tags = {"idle"},
        
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            inst.AnimState:PlayAnimation("item_hat")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },    
    State{
        name="item_in",
        tags = {"idle"},
        
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            inst.AnimState:PlayAnimation("item_in")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },    
    State{
        name="item_out",
        tags = {"idle"},
        
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            inst.AnimState:PlayAnimation("item_out")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },      


    State{
        name = "pinned_pre",
        tags = { "busy", "pinned"},

        onenter = function(inst)
            if inst.components.freezable and inst.components.freezable:IsFrozen() then
                inst.components.freezable:Unfreeze()
            end
            if  inst.components.pinnable then
                inst.components.pinnable:Stick()
            end
            if not inst.components.pinnable or not inst.components.pinnable:IsStuck() then
                inst.sg:GoToState("breakfree")
                return
            end

            inst.components.locomotor:Stop()
            inst:ClearBufferedAction()
            inst.AnimState:OverrideSymbol("swap_goosplat", "goo", "swap_goosplat")
            inst.AnimState:PlayAnimation("hit")            
            --inst.components.inventory:Hide()
            --inst:PushEvent("ms_closepopups") 
        end,

        events =
        {
            EventHandler("onunpin", function(inst, data)
                inst.sg:GoToState("breakfree")
            end),
            EventHandler("animover", function(inst)
                inst.sg.statemem.isstillpinned = true
                inst.sg:GoToState("pinned")
            end),
        },

        onexit = function(inst)
            inst.AnimState:ClearOverrideSymbol("swap_goosplat")
        end,
    },

    State{
        name = "pinned",
        tags = { "busy", "pinned" },

        onenter = function(inst)
            if inst.components.pinnable == nil or not inst.components.pinnable:IsStuck() then
                inst.sg:GoToState("breakfree")
                return
            end

            inst.components.locomotor:Stop()
            inst:ClearBufferedAction()

            inst.AnimState:PlayAnimation("distress_loop", true)
             -- TODO: struggle sound
            inst.SoundEmitter:PlaySound("dontstarve/creatures/spat/spit_playerstruggle", "struggling")

           -- inst.components.inventory:Hide()
           -- inst:PushEvent("ms_closepopups")
        end,

        events =
        {
            EventHandler("onunpin", function(inst, data) inst.sg:GoToState("breakfree") end),
        },

        onexit = function(inst)

            --inst.components.inventory:Show()
            inst.SoundEmitter:KillSound("struggling")
        end,
    },

    State{
        name = "pinned_hit",
        tags = { "busy", "pinned" },

        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst:ClearBufferedAction()

            inst.AnimState:PlayAnimation("hit_goo")

            inst.SoundEmitter:PlaySound("dontstarve/wilson/hit")
            --DoHurtSound(inst)

            --inst.components.inventory:Hide()
            --inst:PushEvent("ms_closepopups")
      
        end,

        events =
        {
            EventHandler("onunpin", function(inst, data)
                inst.sg:GoToState("breakfree")
            end),
            EventHandler("animover", function(inst)
                inst.sg.statemem.isstillpinned = true
                inst.sg:GoToState("pinned")
            end),
        },

        onexit = function(inst)
        end,
    },

    State{
        name = "breakfree",
        tags = { "busy" },

        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst:ClearBufferedAction()

            inst.AnimState:PlayAnimation("distress_pst")
            inst.SoundEmitter:PlaySound("dontstarve/creatures/spat/spit_playerunstuck")

            --inst.components.inventory:Hide()
            --inst:PushEvent("ms_closepopups")

        end,

        events =
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },

        onexit = function(inst)

        end,
    },    

    State{
        name = "sanity_stun",
        tags = { "busy" },
        
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("idle_sanity_pre", false)
            inst.AnimState:PushAnimation("idle_sanity_loop", true)
        end,

        events=
        {
            EventHandler("animqueueover", function(inst) 
            --Here to prevent Prince from being perma-stunned.
            inst.sanity_stunned = false
            inst.sg:GoToState("idle") end ),
        },
    },


}

return StateGraph("prince", states, events, "idle", actionhandlers)