require("stategraphs/commonstates")

local actionhandlers =
{    
    ActionHandler(ACTIONS.CHOP, 
        function(inst)
            if not inst.sg:HasStateTag("prechop") then 
                if inst.sg:HasStateTag("chopping") then
                    return "chop"
                else
                    return "chop_start"
                end
            end 
        end),
    ActionHandler(ACTIONS.MINE, 
        function(inst) 
            if not inst.sg:HasStateTag("premine") then 
                if inst.sg:HasStateTag("mining") then
                    return "mine"
                else
                    return "mine_start"
                end
            end 
        end),
    ActionHandler(ACTIONS.HACK, 
        function(inst) 
            if not inst.sg:HasStateTag("prehack") then
                if inst.sg:HasStateTag("hacking") then
                    return "hack"
                else
                    return "hack_start"
                end
            end
        end),
    ActionHandler(ACTIONS.FISH, 
    function(inst, action)
        if action.invobject and action.invobject.components.fishingrod then
            return "fishing_pre"
        else
            action.invobject = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            return "fishing_pre"
        end
    end),
    ActionHandler(ACTIONS.EAT, 
        function(inst, action)
            if (inst.sg:HasStateTag("busy") and not inst.sg:HastStateTag("eating")) or inst.components.health:IsDead() then
                return nil
            end

            local obj = action.target or action.invobject
            if not (obj and obj.components.edible) then
                return nil
            end

            if inst:HasTag("hungryprince") then
                if obj.components.edible.foodtype == "MEAT" then
                    return "hungryeat"
                else
                    return "hungryquickeat"
                end
            end

            if obj.components.edible.foodtype == "MEAT" then
                return "eat"
            else
                return "quickeat"
            end
        end),

    ActionHandler(ACTIONS.TOGGLEOFF, "give"),
    ActionHandler(ACTIONS.TOGGLEON, "give"),

    ActionHandler(ACTIONS.CUREPOISON, "curepoison"),
    ActionHandler(ACTIONS.REPAIRBOAT, "dolongaction"),

    ActionHandler(ACTIONS.MOUNT, "mount"),
    ActionHandler(ACTIONS.DISMOUNT, "dismount"),
}

local function OnAttackFn(inst, data)
    local vehicle = inst.components.driver.vehicle
    if not inst.components.health:IsDead() and not (vehicle and vehicle.components.boathealth and vehicle.components.boathealth:IsDead()) then
        
        if not vehicle.components.drivable or not vehicle.components.drivable:CanDoHit() then 
            return 
        end

        if data.attacker and (data.attacker:HasTag("insect") or data.attacker:HasTag("twister"))then
            local is_idle = inst.sg:HasStateTag("idle")
            if not is_idle then
                return
            end
        end

        vehicle.components.drivable:GetHit()
        
        if data.stimuli and data.stimuli == "electric" and not inst.components.inventory:IsInsulated() then
            inst.sg:GoToState("electrocute")
        elseif data.damage > 0 then
            inst.sg:GoToState("hit")
        end

    end
end

local events = 
{
        EventHandler("hitbywave",
        function(inst, data)
           -- local currentSpeed = inst.Physics:GetMotorSpeed()
            --inst.Physics:SetMotorVel(currentSpeed/3.0, 0, 0)
        end), 

        EventHandler("boostbywave",
        function(inst, data)
            if inst.sg:HasStateTag("running") then 
                
                local boost = TUNING.WAVEBOOST
                if inst.components.driver and inst.components.driver.vehicle then

                    if inst.components.driver.vehicle.waveboost then
                        boost = data.boost or inst.components.driver.vehicle.waveboost
                    end
                    -- sanity boost, walani's surfboard mainly
                    if inst.components.driver.vehicle.wavesanityboost and inst.components.sanity then
                        inst.components.sanity:DoDelta(inst.components.driver.vehicle.wavesanityboost)
                    end
                end

                local currentSpeed = inst.Physics:GetMotorSpeed()

                inst.Physics:SetMotorVel(currentSpeed + boost , 0, 0)
                --local x,y,z = inst.Transform:GetWorldPosition()
                --print("position is ", y)

                --inst.Transform:SetPosition(x, 1, z)


                --x,y,z = inst.Physics:GetVelocity() 
                --y = 3
                --inst.Physics:SetVel(x, y, z)
            end 

        end),

        EventHandler("sailequipped", 
            function(inst)
                if inst.sg:HasStateTag("rowing") then 
                    inst.sg:GoToState("sail")
                    local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                    if equipped then 
                        equipped:PushEvent("stoprowing", {owner = inst})
                    end
                end 

            end 
        ), 

        EventHandler("sailunequipped", 
            function(inst)
                if inst.sg:HasStateTag("sailing") then 
                    inst.sg:GoToState("row")
                    local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                    if equipped then 
                        equipped:PushEvent("startrowing", {owner = inst})
                    end
                end 
            end 
        ),
        EventHandler("locomote", function(inst)

        local is_attacking = inst.sg:HasStateTag("attack")
        local is_busy = inst.sg:HasStateTag("busy")
     
        local is_moving = inst.sg:HasStateTag("moving")
        local is_running = inst.sg:HasStateTag("running")
        local should_move = inst.components.locomotor:WantsToMoveForward()
        if inst.components.driver and inst.components.driver.vehicle and not inst.components.driver.vehicle.components.drivable then
            should_move = false
        end

        local should_run = inst.components.locomotor:WantsToRun()
        local hasSail = inst.components.driver:GetIsSailing()
        if not should_move then
            if inst.components.driver and inst.components.driver.vehicle then
                inst.components.driver.vehicle:PushEvent("boatstopmoving")
            end
        end 
        if should_move  then 
            if inst.components.driver and inst.components.driver.vehicle then
                inst.components.driver.vehicle:PushEvent("boatstartmoving")
            end
        end 
        if is_attacking or is_busy then return end
        if is_moving and not should_move then
            if hasSail then
                inst.sg:GoToState("sail_stop")
            else
                inst.sg:GoToState("row_stop")
            end
        elseif (not is_moving and should_move) or (is_moving and should_move and is_running ~= should_run) then
            if hasSail then
                inst.sg:GoToState("sail_start")
            else
                inst.sg:GoToState("row_start")
            end
        end 
    end),

    EventHandler("attacked", function(inst, data)
        OnAttackFn(inst, data)
    end),

    EventHandler("boatattacked", function(inst, data)
        OnAttackFn(inst, data)
    end),

    EventHandler("death", function(inst, data)

        if data.cause == "drowning" or (inst.components.driver and inst.components.driver.vehicle and inst.components.driver.vehicle.components.boathealth and inst.components.driver.vehicle.components.boathealth:GetPercent() <= 0 and inst:GetIsOnWater()) then
            inst.sg:GoToState("death_boat")
            local sound_name = inst.soundsname or inst.prefab
            local path = inst.talker_path_override or "dontstarve_DLC002/characters/"
            inst.SoundEmitter:PlaySound(path..sound_name.."/sinking_death")
        else
            inst.sg:GoToState("death")
            local sound_name = inst.soundsname or inst.prefab
            local path = inst.talker_path_override or "dontstarve/characters/"
            if inst.prefab ~= "wes" then
                inst.SoundEmitter:PlaySound(path..sound_name.."/death_voice")
            end
        end

        inst.SoundEmitter:PlaySound("dontstarve/wilson/death")
    end),

    EventHandler("torchranout",
        function(inst, data)

        if not inst:HasTag("armingprince") then
            inst:AddTag("armingprince")
        end

        inst.components.container:Close()

        if not inst.components.inventory:IsItemEquipped(data.torch) then
            local sameTool = inst.components.container:FindItem(function(item)
                return item.prefab == data.torch.prefab
            end)
            if sameTool then
                inst.components.inventory:Equip(sameTool)
            end
        end

        if inst:HasTag("armingprince") then
            inst:DoTaskInTime(0.01, function() inst:RemoveTag("armingprince") end)
        end

    end),

    EventHandler("ontalk", function(inst, data)
        if inst.sg:HasStateTag("idle") then
            inst.sg:GoToState("talk", data.noanim)
        end
        
    end),

    EventHandler("equip", function(inst, data)
        if inst.sg:HasStateTag("idle") then
            if data.eslot == EQUIPSLOTS.HANDS then
                inst.sg:GoToState("item_out")
            else
                inst.sg:GoToState("item_hat")
            end
        end
    end),
    
    EventHandler("unequip", function(inst, data)
        
        if inst.sg:HasStateTag("idle") then
        
            if data.eslot == EQUIPSLOTS.HANDS then
                inst.sg:GoToState("item_in")
            else
                inst.sg:GoToState("item_hat")
            end
        end
    end),

    EventHandler("toolbroke",
        function(inst, data)
            inst.sg:GoToState("toolbroke", data.tool)
        end),      

    EventHandler("landboat", function(inst)
            inst.sg:GoToState("jumpboatland")
        end),

    EventHandler("sneeze", function(inst, data)
        if not inst.components.health:IsDead() and not inst.components.health.invincible then
            if inst.sg:HasStateTag("busy") then
                inst.wantstosneeze = true
            else
                inst.sg:GoToState("sneeze")
            end
        end
    end),    

    CommonHandlers.OnAttack(),
}

local states =
{

    State{
    --[[
        How does this work? Here is how.
        -> Enter state, call OnDismount
        -> StateGraph changes to SGWilson in that function
        -> OnDismount tells SGWilson to enter a new state (jumponboatstart) which handles the jump
        -> At the end of that state BufferedAction is called, which mounts the new boat & changes SG back to this one.
    --]]
        name = "mount",
        tags = {"canrotate", "boating"},
        onenter = function(inst)
            inst.components.driver:OnDismount(false, nil, true)
        end,
    },

     State{
        name = "dismount",
        onenter = function(inst)
            inst:PerformBufferedAction()
        end, 

        onexit = function(inst)
        end, 
    },


    State{
        name = "jumpboatland",
        tags = {"doing", "busy", "canrotate", "invisible"},

        onenter = function(inst)
            inst.components.health:SetInvincible(true)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("landboat")

            --Try to reset Prince's acknowlegement of light!
            inst.LightWatcher:SetLightThresh(.075)
            inst.LightWatcher:SetDarkThresh(.05)

            local boat = inst.components.driver.vehicle 
            
            if boat.landsound then
                inst.SoundEmitter:PlaySound(boat.landsound)
            end

            --Push an event for Princess to display Prince's boat's health!
            local princess = GetPlayer()

            if princess.prefab == "princess" then
            --if inst.components.driver and inst.components.driver.vehicle then
                princess:PushEvent("princeboating", {inst = inst})
            --end
            end

        end,

        onexit = function(inst)
            inst.components.health:SetInvincible(false)
        end,

        events =
        {
            EventHandler("animqueueover", function(inst)
                inst.sg:GoToState("idle")
            end),
        },
    },


    State{
        name = "jumpboatstart",
        tags = {"doing", "busy", "canrotate"},

        onenter = function(inst)
            inst:PerformBufferedAction()
            inst.components.health:SetInvincible(true)
            inst.components.locomotor:Stop()
            inst.components.locomotor:EnableGroundSpeedMultiplier(false)
            inst.AnimState:PlayAnimation("jumpboat")
            inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/boatjump_whoosh")
            local BA = inst:GetBufferedAction()
            inst.sg.statemem.startpos = inst:GetPosition()
            inst.sg.statemem.targetpos = (BA.target and BA.target:GetPosition()) or BA.pos

            RemovePhysicsColliders(inst)
        end,

        onexit = function(inst)
            --This shouldn't actually be reached
            inst.components.health:SetInvincible(false)
            ChangeToCharacterPhysics(inst)
            inst.components.locomotor:Stop()
            inst.components.locomotor:EnableGroundSpeedMultiplier(true)
        end,

        timeline =
        {
            TimeEvent(7*FRAMES, function(inst)
                inst:ForceFacePoint(inst.sg.statemem.targetpos:Get())
                local dist = inst:GetPosition():Dist(inst.sg.statemem.targetpos)
                local speed = dist / (18/30)
                inst.Physics:SetMotorVelOverride(1 * speed, 0, 0)
            end),
        },

        events =
        {
            EventHandler("animover", function(inst)
                inst.components.health:SetInvincible(false)
                inst.Transform:SetPosition(inst.sg.statemem.targetpos:Get())
                inst.Physics:Stop()
                ChangeToCharacterPhysics(inst)
                inst.components.locomotor:Stop()
                inst.components.locomotor:EnableGroundSpeedMultiplier(true)
            end),
        },
    },

    State{
        name = "curepoison",
        tags ={"busy"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("quick_eat")

        end,

        timeline=
        {
            TimeEvent(12*FRAMES, function(inst)
                inst:PerformBufferedAction()
                inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/player_drink")
                inst.sg:RemoveStateTag("busy")
            end),
        },

        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("celebrate") end),
        },
    },

    State{
        name = "celebrate",
        tags ={"idle"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("research")
        end,

        timeline = 
        {
            TimeEvent( 8*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/antivenom_whoosh") end),
            TimeEvent(13*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/heelclick") end),
            TimeEvent(21*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/heelclick") end),
        },

        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },
    },


    State{
        name = "give",
        tags = {"busy", "nointerrupt", "boating"},
        
        onenter = function(inst)
            --inst.components.driver:SplitFromVehicle()
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("give") 
        end,
        
        timeline =
        {
            TimeEvent(13*FRAMES, function(inst)
                inst:PerformBufferedAction()
            end),
        },        

        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },   


    State{
        name = "idle",
        tags = {"idle", "canrotate", "boating"},
        onenter = function(inst, pushanim)
            
            inst.components.locomotor:Stop()

            if IsDLCEnabled(PORKLAND_DLC) then
                if inst.wantstosneeze then
                    inst.sg:GoToState("sneeze")
                end
            end

            local equippedArmor = inst.components.inventory and inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)

            if equippedArmor and equippedArmor:HasTag("band") then
                inst.sg:GoToState("enter_onemanband")
                return
            end

            local anims = {}
            
            local anim = "idle_loop"
            
            if not inst.components.sanity:IsSane() then
                table.insert(anims, "idle_sanity_pre")
                table.insert(anims, "idle_sanity_loop")
            elseif inst.components.temperature:IsFreezing() then
                table.insert(anims, "idle_shiver_pre")
                table.insert(anims, "idle_shiver_loop")
            elseif inst.components.temperature:IsOverheating() then
                table.insert(anims, "idle_hot_pre")
                table.insert(anims, "idle_hot_loop")

            else
                table.insert(anims, "idle_loop")
            end
            
            if pushanim then
                for k,v in pairs (anims) do
                    inst.AnimState:PushAnimation(v, k == #anims)
                end
            else
                inst.AnimState:PlayAnimation(anims[1], #anims == 1)
                for k,v in pairs (anims) do
                    if k > 1 then
                        inst.AnimState:PushAnimation(v, k == #anims)
                    end
                end
            end
            
            inst.sg:SetTimeout(math.random()*4+2)
        end,
        
        ontimeout= function(inst)
            if inst.components.temperature:GetCurrent() > 70 then
                return 
            end
            inst.sg:GoToState("funnyidle")
        end,
    },

    State{
        
        name = "funnyidle",
        tags = {"idle", "canrotate", "boating"},
        onenter = function(inst)
        
            if inst.components.poisonable:IsPoisoned() then
                inst.AnimState:PlayAnimation("idle_poison_pre")
                inst.AnimState:PushAnimation("idle_poison_loop")
                inst.AnimState:PushAnimation("idle_poison_pst", false)
            elseif inst.components.temperature:GetCurrent() < 10 then
                inst.AnimState:PlayAnimation("idle_shiver_pre")
                inst.AnimState:PushAnimation("idle_shiver_loop")
                inst.AnimState:PushAnimation("idle_shiver_pst", false)
            elseif inst.components.temperature:GetCurrent() > 60 then
                --plug in overheats once they're done
                inst.AnimState:PlayAnimation("idle_hot_pre")
                inst.AnimState:PushAnimation("idle_hot_loop")
                inst.AnimState:PushAnimation("idle_hot_pst", false)
            elseif inst.components.hunger:GetPercent() < TUNING.HUNGRY_THRESH then
                inst.AnimState:PlayAnimation("hungry")
                inst.SoundEmitter:PlaySound("dontstarve/wilson/hungry")    
            elseif inst.components.sanity:GetPercent() < .5 then
                inst.AnimState:PlayAnimation("idle_inaction_sanity")
            else
                inst.AnimState:PlayAnimation("idle_inaction")
            end
        end,

        events=
        {
            EventHandler("animqueueover", function(inst) inst.sg:GoToState("idle") end ),
        },
        
    },


    State{
        name = "enter_onemanband",
        tags = {"doing", "playing", "idle", "boating"},

        onenter = function(inst)
        inst.components.locomotor:Stop()
        inst.AnimState:PlayAnimation("idle_onemanband1_pre")
        inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")
        end,

        onexit = function(inst)
        end,

        events = 
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("play_onemanband") end),
        },
    },

    State{
        name = "play_onemanband",
        tags = {"doing", "playing", "idle", "boating"},

        onenter = function(inst)

            inst.components.locomotor:Stop()
            --inst.AnimState:PlayAnimation("idle_onemanband1_pre")
            inst.AnimState:PlayAnimation("idle_onemanband1_loop")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")
        end,

        onexit = function(inst)
        end,

        events =
        {
            EventHandler("animover", function(inst)
                if math.random() <= 0.15 then
                    inst.sg:GoToState("play_onemanband_stomp") -- go into stomp
                else
                    inst.sg:GoToState("play_onemanband")-- loop state again
                end
            end),
        },
    },

    State{
        name = "play_onemanband_stomp",
        tags = {"doing", "playing", "idle", "boating"},

        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("idle_onemanband1_pst")
            inst.AnimState:PushAnimation("idle_onemanband2_pre")
            inst.AnimState:PushAnimation("idle_onemanband2_loop")
            inst.AnimState:PushAnimation("idle_onemanband2_pst", false)  
            inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband") 
        end,

        onexit = function(inst)
        end,

        timeline=
        {
            TimeEvent(20*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),

            TimeEvent(25*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),
            
            TimeEvent(30*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),

            TimeEvent(35*FRAMES, function( inst )
                inst.SoundEmitter:PlaySound("dontstarve/wilson/onemanband")                
            end),
        },

        events = 
        {
            EventHandler("animqueueover", function(inst)
                inst.sg:GoToState("idle") 
            end),
        },
    },


    State{
        name = "sneeze",
        tags = {"busy","sneeze"},
        
        onenter = function(inst)
            inst.wantstosneeze = false
            --inst.SoundEmitter:PlaySound("dontstarve/wilson/hit",nil,.02)
            inst.AnimState:PlayAnimation("sneeze")
            inst.SoundEmitter:PlaySound("dontstarve_DLC003/characters/sneeze")
            inst:ClearBufferedAction()
            
            if inst.prefab ~= "wes" then
                local sound_name = inst.soundsname or inst.prefab
                local path = inst.talker_path_override or "dontstarve/characters/"
                --local equippedHat = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
                --if equippedHat and equippedHat:HasTag("muffler") then
                    --inst.SoundEmitter:PlaySound(path..sound_name.."/gasmask_hurt")
                --else
                    local sound_event = path..sound_name.."/hurt"
                    inst.SoundEmitter:PlaySound(inst.hurtsoundoverride or sound_event)
                --end
            end
            inst.components.locomotor:Stop()  

            --inst.components.talker:Say(GetString(inst.prefab, "ANNOUNCE_SNEEZE"))        
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        }, 
        
        timeline =
        {
            TimeEvent(10*FRAMES, function(inst)
                if inst.components.hayfever then
                    inst.components.hayfever:DoSneezeEffects()
                end
                inst.sg:RemoveStateTag("busy")
            end),
        },        
               
    },
    

    State{
        name = "eat",
        tags ={"busy", "boating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            --inst.components.driver:SplitFromVehicle()
            local is_gear = inst:GetBufferedAction() and inst:GetBufferedAction().invobject and inst:GetBufferedAction().invobject.components.edible and inst:GetBufferedAction().invobject.components.edible.foodtype == "GEARS"

            if not is_gear then
                inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            end
            
            inst.AnimState:PlayAnimation("eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            
            TimeEvent(28*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),
            
            TimeEvent(30*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("busy")
            end),
            
            TimeEvent(70*FRAMES, function(inst) 
                inst.SoundEmitter:KillSound("eating")    
            end),
            
        },        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    }, 


    State{
        name = "quickeat",
        tags ={"busy", "boating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            --inst.components.driver:SplitFromVehicle()
            local is_gear = inst:GetBufferedAction() and inst:GetBufferedAction().invobject and inst:GetBufferedAction().invobject.components.edible and inst:GetBufferedAction().invobject.components.edible.foodtype == "GEARS"
            if not is_gear then
                inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            end
            inst.AnimState:PlayAnimation("quick_eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            TimeEvent(12*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),
            TimeEvent(14*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("busy")
            end),
        },        
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },

    State{
        name = "hungryeat",
        tags ={"busy", "eating", "boating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            
            inst.AnimState:PlayAnimation("eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            
            TimeEvent(28*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
            end),
            
            TimeEvent(30*FRAMES, function(inst) 
                --inst.sg:RemoveStateTag("busy")
            end),
            
            TimeEvent(70*FRAMES, function(inst) 
                inst.SoundEmitter:KillSound("eating")    
            end),
            
        },        
        events=
        {
            EventHandler("animover", function(inst) 
            if inst.components.hunger and inst.components.hunger:GetPercent() >= 0.80 then
                if inst:HasTag("hungryprince") then
                    inst:RemoveTag("hungryprince")
                end
            end
            inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },    
    
   State{
        name = "hungryquickeat",
        tags ={"busy", "eating", "boating"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/eat", "eating")    
            inst.AnimState:PlayAnimation("quick_eat")
            inst.components.hunger:Pause()
        end,

        timeline=
        {
            TimeEvent(12*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
                --inst.sg:RemoveStateTag("busy")
            end),
            TimeEvent(14*FRAMES, function(inst) 
                --inst.sg:RemoveStateTag("busy")
            end),
        },        
        
        events=
        {
            EventHandler("animover", function(inst)
            if inst.components.hunger and inst.components.hunger:GetPercent() >= 0.80 then
                if inst:HasTag("hungryprince") then
                    inst:RemoveTag("hungryprince")
                end
            end
            inst.sg:GoToState("idle") end),
        },
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("eating")    
            inst.components.hunger:Resume()
        end,
    },    


    State{
        name = "talk",
        tags = {"idle", "talking"},

        onenter = function(inst, noanim)
            inst.components.locomotor:Stop()
            if not noanim then
                inst.AnimState:PlayAnimation("dial_loop", true)
            end

            local sound_name = inst.soundsname or inst.prefab
            local path = inst.talker_path_override or "dontstarve/characters/"
            local equippedHat = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)

            if equippedHat and equippedHat:HasTag("muffler") then
                inst.SoundEmitter:PlaySound(path..sound_name.."/gasmask_talk", "talk")
                
            elseif inst.talksoundoverride then
                 inst.SoundEmitter:PlaySound(inst.talksoundoverride, "talk")
            else
                local sound_name = inst.soundsname or inst.prefab
                local path = inst.talker_path_override or "dontstarve/characters/"
                inst.SoundEmitter:PlaySound(path..sound_name.."/talk_LP", "talk")
            end

            inst.sg:SetTimeout(1.5 + math.random()*.5)
        end,

        ontimeout = function(inst)
            inst.SoundEmitter:KillSound("talk")
            if inst.components.talker.endspeechsound then
                inst.SoundEmitter:PlaySound(inst.components.talker.endspeechsound)
            end            
            inst.sg:GoToState("idle")
        end,

        onexit = function(inst)
            inst.SoundEmitter:KillSound("talk")
            if inst.components.talker.endspeechsound then
                inst.SoundEmitter:PlaySound(inst.components.talker.endspeechsound)
            end               
        end,

        events=
        {
            EventHandler("donetalking", function(inst) inst.sg:GoToState("idle") end),
        },
    },


    State{
        name = "run_start",
        tags = {"moving", "running", "canrotate"},
        
        onenter = function(inst)
			inst.components.locomotor:RunForward()
            inst.AnimState:PlayAnimation("run_pre")
            inst.sg.mem.foosteps = 0
        end,

        onupdate = function(inst)
            inst.components.locomotor:RunForward()
        end,

        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("run") end ),        
        },
        
        timeline=
        {        
            TimeEvent(4*FRAMES, function(inst)
                --inst.SoundEmitter:PlaySound("dontstarve/maxwell/shadowmax_step")
            end),
        },        
        
    },

    State{
        name = "run",
        tags = {"moving", "running", "canrotate"},
        
        onenter = function(inst) 
            inst.components.locomotor:RunForward()
            inst.AnimState:PlayAnimation("run_loop")
            
        end,
        
        onupdate = function(inst)
            inst.components.locomotor:RunForward()
        end,

        timeline=
        {
            TimeEvent(7*FRAMES, function(inst)
				inst.sg.mem.foosteps = inst.sg.mem.foosteps + 1
                --inst.SoundEmitter:PlaySound("dontstarve/maxwell/shadowmax_step")
            end),
            TimeEvent(15*FRAMES, function(inst)
				inst.sg.mem.foosteps = inst.sg.mem.foosteps + 1
                --inst.SoundEmitter:PlaySound("dontstarve/maxwell/shadowmax_step")
            end),
        },
        
        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("run") end ),        
        },
    },
    
    State{
    
        name = "run_stop",
        tags = {"canrotate", "idle"},
        
        onenter = function(inst) 
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("run_pst")
        end,
        
        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),        
        },
        
    },

    State{
        name = "attack",
        tags = {"attack", "notalking", "abouttoattack", "busy"},
        
        onenter = function(inst) 

            inst.sg.statemem.target = inst.components.combat.target
            inst.components.combat:StartAttack()
            inst.Physics:Stop()
            local weapon = inst.components.combat:GetWeapon()
            local otherequipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)

            if weapon then
                inst.AnimState:PlayAnimation("atk")
                if weapon:HasTag("icestaff") then
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_icestaff")
                elseif weapon:HasTag("shadow") then
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_nightsword")
                elseif weapon:HasTag("firestaff") then
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_firestaff")
                elseif weapon:HasTag("cutlass") then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/swordfish_sword")
                elseif weapon:HasTag("pegleg") then
                    inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/pegleg_weapon")
                else
                    inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_weapon")
                end
            elseif otherequipped and (otherequipped:HasTag("light") or otherequipped:HasTag("nopunch")) then
                inst.AnimState:PlayAnimation("atk")
                inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_weapon")
            else
                inst.sg.statemem.slow = true
                inst.AnimState:PlayAnimation("punch")
                inst.SoundEmitter:PlaySound("dontstarve/wilson/attack_whoosh")
            end
            
            if inst.components.combat.target then
                if inst.components.combat.target and inst.components.combat.target:IsValid() then
                    inst:FacePoint(Point(inst.components.combat.target.Transform:GetWorldPosition()))
                end
            end
            
        end,
        
        timeline=
        {
            TimeEvent(8*FRAMES, function(inst) 
                inst.components.combat:DoAttack(inst.sg.statemem.target) 
                inst.sg:RemoveStateTag("abouttoattack") 
                local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                if handslot and handslot.components.finiteuses then
                    if handslot.components.finiteuses.current <= 0 then
                        inst.components.inventory:RemoveItem(handslot)--:Remove()  
                        inst.components.container:GiveItem(handslot)
                        --inst.AnimState:Hide("ARM_carry") 
                        --inst.AnimState:Show("ARM_normal") 
                    end
                end
            end),
            TimeEvent(12*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("busy")
            end),               
            TimeEvent(13*FRAMES, function(inst)
                if not inst.sg.statemem.slow then
                    inst.sg:RemoveStateTag("attack")
                end
            end),
            TimeEvent(24*FRAMES, function(inst)
                if inst.sg.statemem.slow then
                    inst.sg:RemoveStateTag("attack")
                end
            end),           
        },
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle")
            end ),
        },
    },  

    State{
        name = "death",
        tags = {"busy", "nointerrupt", "dead"},

        onenter = function(inst)
            --inst.components.driver:SplitFromVehicle()
            --inst.components.driver:CombineWithVehicle()
            inst.components.locomotor:Stop()
            inst.last_death_position = inst:GetPosition()
            inst.AnimState:Hide("swap_arm_carry")
            inst.AnimState:PlayAnimation("death")
            --For his equipped gear!
            local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if handslot then
                inst.components.inventory:RemoveItem(handslot)
                inst.components.inventory:DropItem(handslot, true, true)
            end
            local bodyslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            if bodyslot then
                if bodyslot.prefab ~= "amulet" and bodyslot.prefab ~= "armor_lifejacket" then
                    inst.components.inventory:RemoveItem(bodyslot)
                    inst.components.inventory:DropItem(bodyslot, true, true)
                end
            end
            local headslot = inst.com
            local headslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
            if headslot then
                inst.components.inventory:RemoveItem(headslot)
                inst.components.inventory:DropItem(headslot, true, true)
            end
            --inst.components.inventory:DropEverything()
            --For the rest of his contents!
            inst.components.container:DropEverything()
        end,
    },

    State{
        name = "death_boat",
        tags = {"busy", "nointerrupt", "dead"},

        onenter = function(inst)
            --inst.components.driver:SplitFromVehicle()
            --inst.components.driver:CombineWithVehicle()
            inst.components.locomotor:Stop()
            inst.last_death_position = inst:GetPosition()
            inst.AnimState:Hide("swap_arm_carry")
            inst.AnimState:PlayAnimation("boat_death")
            --For his equipped gear!
            local handslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if handslot then
                inst.components.inventory:RemoveItem(handslot)
                inst.components.inventory:DropItem(handslot, true, true)
            end
            local bodyslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            if bodyslot then
                if bodyslot.prefab ~= "amulet" and bodyslot.prefab ~= "armor_lifejacket" then
                    inst.components.inventory:RemoveItem(bodyslot)
                    inst.components.inventory:DropItem(bodyslot, true, true)
                end
            end
            local headslot = inst.com
            local headslot = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HEAD)
            if headslot then
                inst.components.inventory:RemoveItem(headslot)
                inst.components.inventory:DropItem(headslot, true, true)
            end
            --inst.components.inventory:DropEverything()
            --For the rest of his contents!
            inst.components.container:DropEverything()
        end,

        onexit= function(inst) inst.DynamicShadow:Enable(true) end,

        timeline=
        {
            TimeEvent(70*FRAMES, function(inst)
                inst.DynamicShadow:Enable(false)
            end),
        },
    },
   
    State{
        name = "hit",
        tags = {"busy"},
        
        onenter = function(inst)
            inst:ClearBufferedAction()
            inst.AnimState:PlayAnimation("hit")
            inst.Physics:Stop()            
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        }, 
        
        timeline =
        {
            TimeEvent(3*FRAMES, function(inst)
                inst.sg:RemoveStateTag("busy")
            end),
        },               
    },


    State{
        name = "electrocute",
        tags = {"busy"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("shock")
            inst.AnimState:SetBloomEffectHandle("shaders/anim.ksh")
            inst.fx = SpawnPrefab("shock_fx")

            inst.fx.Transform:SetRotation(inst.Transform:GetRotation())

            local pos = inst:GetPosition()
            inst.fx.Transform:SetPosition(pos.x, pos.y, pos.z)

            --inst.Light:Enable(true)
        end,

        onexit = function(inst)
            --inst.Light:Enable(false)
            inst.AnimState:ClearBloomEffectHandle()
            inst.fx:Remove()
        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("electrocute_pst")                
            end),
        },
    },
    State{
        name = "electrocute_pst",
        tags = {"busy"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("shock_pst")
        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle") 
            end),
        },
    },



    State{
        name = "stunned",
        tags = {"busy", "canrotate"},

        onenter = function(inst)
            inst:ClearBufferedAction()
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("idle_sanity_pre")
            inst.AnimState:PushAnimation("idle_sanity_loop", true)
            inst.sg:SetTimeout(5)
        end,

        ontimeout = function(inst)
            inst.sg:GoToState("idle")
        end,
    },

        State{ name = "chop_start",
        tags = {"prechop", "chopping", "working"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("chop_pre")

        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("chop") end),
        },
    },
    
    State{
        name = "chop",
        tags = {"prechop", "chopping", "working"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("chop_loop")        
        end,

        timeline=
        {
            TimeEvent(5*FRAMES, function(inst) 
                    inst:PerformBufferedAction() 
            end),

            TimeEvent(9*FRAMES, function(inst)
                    inst.sg:RemoveStateTag("prechop")
            end),

            TimeEvent(16*FRAMES, function(inst) 
                inst.sg:RemoveStateTag("chopping")
            end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("idle")
            end ),            
        },        
    },

    State{ 
        name = "mine_start",
        tags = {"premine", "working"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("pickaxe_pre")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("mine") end),
        },
    },
    
    State{
        name = "mine",
        tags = {"premine", "mining", "working"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("pickaxe_loop")
        end,

        timeline=
        {
            TimeEvent(9*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
                inst.sg:RemoveStateTag("premine") 
                inst.SoundEmitter:PlaySound("dontstarve/wilson/use_pick_rock")
            end),
            
            -- TimeEvent(14*FRAMES, function(inst)
            --     if  inst.sg.statemem.action and 
            --         inst.sg.statemem.action.target and 
            --         inst.sg.statemem.action.target:IsActionValid(inst.sg.statemem.action.action) and 
            --         inst.sg.statemem.action.target.components.workable then
            --             inst:ClearBufferedAction()
            --             inst:PushBufferedAction(inst.sg.statemem.action)
            --     end
            -- end),            
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.AnimState:PlayAnimation("pickaxe_pst") 
                inst.sg:GoToState("idle", true)
            end ),            
        },        
    },


    State{ 
        name = "hack_start",
        tags = {"prehack", "working"},
        onenter = function(inst)
            inst.Physics:Stop()
            inst.AnimState:PlayAnimation("chop_pre")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("hack") end),
        },
    },
    
    State{
        name = "hack",
        tags = {"prehack", "hacking", "working"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("chop_loop")
        end,

        timeline=
        {
            TimeEvent(9*FRAMES, function(inst) 
                inst:PerformBufferedAction() 
                --local currenttool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                --if currenttool.components.finiteuses then
                    --currenttool.components.finiteuses:Use(1)
                --end
                inst.sg:RemoveStateTag("prehack") 
                inst.sg:RemoveStateTag("working") 
            end),
        },
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.AnimState:PlayAnimation("chop_pst") 
                inst.sg:GoToState("idle", true)
            end ),            
        },        
    },

    State{ name = "fishing_pre",
        tags = {"prefish", "fishing"},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:Show("RIPPLES")
            inst.AnimState:PlayAnimation("fishing_pre")
        end,
        
        timeline = 
        {
            TimeEvent(13*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_cast") end),
            TimeEvent(15*FRAMES, function(inst) inst:PerformBufferedAction() end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_baitsplash")
                inst.sg:GoToState("fishing")
            end ),
        },        
    },
    

    State{
        name = "fishing",
        tags = {"fishing"},
        onenter = function(inst, pushanim)
            if pushanim then
                if type(pushanim) == "string" then
                    inst.AnimState:PlayAnimation(pushanim)
                end
                inst.AnimState:PushAnimation("fishing_idle", true)
            else
                inst.AnimState:PlayAnimation("fishing_idle", true)
            end
            local equippedTool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if equippedTool and equippedTool.components.fishingrod then
                equippedTool.components.fishingrod:WaitForFish()
            end
        end,
        
        events=
        {
            EventHandler("fishingnibble", function(inst) inst.sg:GoToState("fishing_nibble") end ),
        },
    },
    
    State{ name = "fishing_pst",
        tags = {},
        onenter = function(inst)
            inst.components.locomotor:Stop()
            inst.AnimState:PlayAnimation("fishing_pst")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },
    
    State{
        name = "fishing_nibble",
        tags = {"fishing", "nibble"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("bite_light_pre")
            inst.AnimState:PushAnimation("bite_light_loop", true)
            inst.sg:SetTimeout(1 + math.random())
            inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishinwater", "splash")
            inst:DoTaskInTime(0.25 + math.random(), function()
                local fishingrod = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                --Code ripped from ACTIONS.REEL because there's an oversight I'm too impatient to fix.
                if fishingrod and fishingrod.components.fishingrod and fishingrod.components.fishingrod:IsFishing() then
                    if fishingrod.components.fishingrod:HasHookedFish() then
                        fishingrod.components.fishingrod:Reel()
                    elseif fishingrod.components.fishingrod:FishIsBiting() then
                        fishingrod.components.fishingrod:Hook()
                    else
                    fishingrod.components.fishingrod:StopFishing()
                    end
                end
            end)
        end,
        
        onexit = function(inst)
            inst.SoundEmitter:KillSound("splash")
        end,
        
        ontimeout = function(inst)
            inst.sg:GoToState("fishing", "bite_light_pst")
        end,
        
        events = 
        {
            EventHandler("fishingstrain", function(inst) inst.sg:GoToState("fishing_strain") end),
        },
    }, 
    
    State{
        name = "fishing_strain",
        tags = {"fishing"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("bite_heavy_pre")
            inst.AnimState:PushAnimation("bite_heavy_loop", true)
            inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishinwater", "splash")
            inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_strain", "strain")
            inst:DoTaskInTime(0.25 + math.random(), function()
                local fishingrod = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                --Code ripped from ACTIONS.REEL because there's an oversight I'm too impatient to fix.
                if fishingrod and fishingrod.components.fishingrod and fishingrod.components.fishingrod:IsFishing() then
                    if fishingrod.components.fishingrod:HasHookedFish() then
                        fishingrod.components.fishingrod:Reel()
                    elseif fishingrod.components.fishingrod:FishIsBiting() then
                        fishingrod.components.fishingrod:Hook()
                    else
                    fishingrod.components.fishingrod:StopFishing()
                    end
                end
            end)
        end,
        
        onexit = function(inst)
            inst.SoundEmitter:KillSound("splash")
            inst.SoundEmitter:KillSound("strain")
        end,
        
        events = 
        {
            EventHandler("fishingcatch", function(inst, data)
                inst.sg:GoToState("catchfish", data.build)
            end),
            EventHandler("fishingloserod", function(inst)
                inst.sg:GoToState("loserod")
            end),

        },
    },
    
    State{
        name = "catchfish",
        tags = {"fishing", "catchfish", "busy"},
        onenter = function(inst, build)
            inst.AnimState:PlayAnimation("fish_catch")
           -- print("Using ", build, " to swap out fish01")
            inst.AnimState:OverrideSymbol("fish01", build, "fish01")
            
            -- inst.AnimState:OverrideSymbol("fish_body", build, "fish_body")
            -- inst.AnimState:OverrideSymbol("fish_eye", build, "fish_eye")
            -- inst.AnimState:OverrideSymbol("fish_fin", build, "fish_fin")
            -- inst.AnimState:OverrideSymbol("fish_head", build, "fish_head")
            -- inst.AnimState:OverrideSymbol("fish_mouth", build, "fish_mouth")
            -- inst.AnimState:OverrideSymbol("fish_tail", build, "fish_tail")
        end,
        
        onexit = function(inst)
            inst.AnimState:ClearOverrideSymbol("fish01")
            -- inst.AnimState:ClearOverrideSymbol("fish_body")
            -- inst.AnimState:ClearOverrideSymbol("fish_eye")
            -- inst.AnimState:ClearOverrideSymbol("fish_fin")
            -- inst.AnimState:ClearOverrideSymbol("fish_head")
            -- inst.AnimState:ClearOverrideSymbol("fish_mouth")
            -- inst.AnimState:ClearOverrideSymbol("fish_tail")
        end,

        timeline=
        {
            TimeEvent(8*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishcaught") end),
            TimeEvent(10*FRAMES, function(inst) inst.sg:RemoveStateTag("fishing") end),
            TimeEvent(23*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_fishland") end),
            TimeEvent(24*FRAMES, function(inst)
                local equippedTool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                if equippedTool and equippedTool.components.fishingrod then
                    equippedTool.components.fishingrod:Collect()
                end
            end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("idle")
            end ),
        },        
    },       
    
    State{
        name = "loserod",
        tags = {"busy"},
        onenter = function(inst)
            local equippedTool = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if equippedTool and equippedTool.components.fishingrod then
                equippedTool.components.fishingrod:Release()
                equippedTool:Remove()
            end
            inst.AnimState:PlayAnimation("fish_nocatch")
        end,

        timeline=
        {
            TimeEvent(4*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/fishingpole_lostrod") end),
        },
        
        events=
        {
            EventHandler("animover", function(inst) 
                inst.sg:GoToState("idle")
            end ),
        },        
    },


    State{
        name = "toolbroke",
        tags = {"canrotate", "busy"},
        onenter = function(inst, tool)
            inst.AnimState:PlayAnimation("hit")
            inst.SoundEmitter:PlaySound("dontstarve/wilson/use_break")
            inst.AnimState:Hide("ARM_carry") 
            inst.AnimState:Show("ARM_normal") 
            local brokentool = SpawnPrefab("brokentool")
            brokentool.Transform:SetPosition(inst.Transform:GetWorldPosition() )
            inst.sg.statemem.tool = tool
        end,
        
        onexit = function(inst)
            local sameTool = inst.components.inventory:FindItem(function(item)
                return item.prefab == inst.sg.statemem.tool.prefab
            end)
            if sameTool then
                inst.components.inventory:Equip(sameTool)
            end

            if inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS) then
                inst.AnimState:Show("ARM_carry") 
                inst.AnimState:Hide("ARM_normal")
            end

        end,
        
        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle")
            end ),
        },
    },

    State{
        name = "tool_slip",
        tags = {"busy"},
        onenter = function(inst, tool)
            inst.AnimState:PlayAnimation("hit")
            inst.SoundEmitter:PlaySound("dontstarve_DLC002/common/tool_slip")
            inst.AnimState:Hide("ARM_carry")
            inst.AnimState:Show("ARM_normal")
            local brokentool = SpawnPrefab("brokentool")
            brokentool.Transform:SetPosition(inst.Transform:GetWorldPosition() )
            inst.sg.statemem.tool = tool
        end,

        events=
        {
            EventHandler("animover", function(inst)
                inst.sg:GoToState("idle")
            end ),
        },
    },


    State{
        name = "row_start",
        tags = {"moving", "running", "canrotate", "rowing"},
        
        onenter = function(inst)
            --inst.components.driver:CombineWithVehicle()
            inst.components.locomotor:RunForward()
            local anim = inst.components.driver.vehicle.components.drivable.prerunanimation
            inst.AnimState:PlayAnimation(anim)

            -- unequip whatever the player is holding and store it somewhere 
            local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if equipped then 
                equipped:PushEvent("startrowing", {owner = inst})
            end
      
           -- print("currently equipped ", equipped.prefab)
        end,

        onupdate = function(inst)
            inst.components.locomotor:RunForward()
        end,

        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("row") end ),
        },
    },

    State{
        name = "row",
        tags = {"canrotate", "moving", "running", "boating", "rowing"},
        onenter = function(inst)
       
            --inst.components.driver:CombineWithVehicle()
            inst.SoundEmitter:PlaySound(inst.components.driver.vehicle.components.drivable.creaksound)
            inst.SoundEmitter:PlaySound( "dontstarve_DLC002/common/boat_paddle")
            local anim = inst.components.driver.vehicle.components.drivable.runanimation
            inst.AnimState:PlayAnimation(anim, false)
            if inst.components.driver.vehicle.components.rowboatwakespawner then 
                inst.components.driver.vehicle.components.rowboatwakespawner:StartSpawning()
            end 

        end,
        
        onexit = function(inst, nextState)
           if inst.components.driver.vehicle.components.rowboatwakespawner then 
                inst.components.driver.vehicle.components.rowboatwakespawner:StopSpawning()
            end 
            if nextState ~= "row" and nextState ~= "sail" then 
                inst.components.locomotor:Stop(true)
                if nextState == "dismount" then --Make sure equipped items are pulled back out (only really for items with flames right now)
                    local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                    if equipped then
                        equipped:PushEvent("stoprowing", {owner = inst})
                    end
                end 
            end 
        end,

         events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("row") end ),
        },
             
    },  

     State{

        name = "row_stop",
        tags = {"canrotate", "idle"},

        onenter = function(inst) 
            --inst.components.driver:CombineWithVehicle()
            inst.components.locomotor:Stop()
            local anim = inst.components.driver.vehicle.components.drivable.postrunanimation
            inst.AnimState:PlayAnimation(anim)
            --If the player had something in their hand before starting to row, put it back.

            if inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS) then
                inst.AnimState:PushAnimation("item_out", false)
            end

        end,

        events=
        {
            EventHandler("animqueueover", function(inst)
                local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                if equipped then
                    equipped:PushEvent("stoprowing", {owner = inst})
                end
                inst.sg:GoToState("idle")
            end),
            EventHandler("trawlover", function(inst) inst.AnimState:PlayAnimation("trawlover") end),
        },
    },

    State{
        name = "sail_start",
        tags = {"moving", "running", "canrotate", "sailing"},
        
        onenter = function(inst)
            --inst.components.driver:CombineWithVehicle()

            
            inst.components.locomotor:RunForward()
            local anim = inst.components.driver.vehicle.components.drivable.sailstartanim or "sail_pre"
            inst.AnimState:PlayAnimation(anim)

            local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
            if equipped then 
                equipped:PushEvent("startsailing", {owner = inst})
            end
      
           -- print("currently equipped ", equipped.prefab)
        end,

        onupdate = function(inst)
            inst.components.locomotor:RunForward()
        end,

        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("sail") end ),
            EventHandler("trawlover", function(inst) inst.AnimState:PlayAnimation("trawlover") end),
        },
               
        
    },

    State{
        name = "sail",
        tags = {"canrotate", "moving", "running", "boating", "sailing"},
        onenter = function(inst)
            --inst.components.driver:CombineWithVehicle()

            local loopsound = nil 
            local flapsound = nil 

            local boat = inst.components.driver.vehicle
            if boat then 
                if boat.components.container and boat.components.container.hasboatequipslots then 
                    local sail = boat.components.container:GetItemInBoatSlot(BOATEQUIPSLOTS.BOAT_SAIL)
                    if sail then 
                        loopsound = sail.loopsound
                        flapsound = sail.flapsound
                    end 
                elseif boat.sailsound then
                    loopsound = boat.sailsound
                end 
            end
           
            if not inst.SoundEmitter:PlayingSound("sail_loop") and loopsound then 
                  inst.SoundEmitter:PlaySound( "dontstarve_DLC002/" .. loopsound, "sail_loop")
              end 
            if flapsound then 
                inst.SoundEmitter:PlaySound( "dontstarve_DLC002/" .. flapsound) 
            end 
            inst.SoundEmitter:PlaySound(inst.components.driver.vehicle.components.drivable.creaksound)
            
            if inst.components.driver.vehicle.components.drivable then
                local anim = inst.components.driver.vehicle.components.drivable.sailloopanim or "sail_loop"
                inst.AnimState:PlayAnimation(anim, false)
            end
            if inst.components.driver.vehicle.components.rowboatwakespawner then 
                inst.components.driver.vehicle.components.rowboatwakespawner:StartSpawning()
            end 

        end,
        
        onexit = function(inst, nextState)

           if inst.components.driver.vehicle.components.rowboatwakespawner then 
                inst.components.driver.vehicle.components.rowboatwakespawner:StopSpawning()
            end 
            
            local loopsound
            local boat = inst.components.driver.vehicle
            if boat then 
                if boat.components.container and boat.components.container.hasboatequipslots then 
                    local sail = boat.components.container:GetItemInBoatSlot(BOATEQUIPSLOTS.BOAT_SAIL)
                    if sail then 
                        loopsound = sail.loopsound
                    end 
                end 
            end

            if nextState ~= "sail" then 
                inst.SoundEmitter:KillSound( "sail_loop")
                if nextState ~= "row" then 
                    inst.components.locomotor:Stop(true)
                end 
            end

            
           -- inst.components.playercontroller:Enable(true)
        end,

        events=
        {   
            EventHandler("animover", function(inst) inst.sg:GoToState("sail") end ),
            EventHandler("trawlover", function(inst) inst.AnimState:PlayAnimation("trawlover") end),
        },
             
    },  

     State{
    
        name = "sail_stop",
        tags = {"canrotate", "idle"},
        
        onenter = function(inst) 
            --inst.components.driver:CombineWithVehicle()
            inst.components.locomotor:Stop()
            local anim = (inst.components.driver.vehicle.components.drivable and inst.components.driver.vehicle.components.drivable.sailstopanim) or "sail_pst"
            inst.AnimState:PlayAnimation(anim)
           
            --If the player had something in their hand before starting to row, put it back. 
            
        end,

        onexit = function(inst, nextState)
            
            if inst.components.driver and inst.components.driver.vehicle and inst.components.driver.vehicle.components.boathealth then
                
            end
        end,
        
        events=
        {   
            EventHandler("animover", function(inst) 
                local equipped = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.HANDS)
                if equipped then 
                    equipped:PushEvent("stopsailing", {owner = inst})
                end
                inst.sg:GoToState("idle") 
            end ),
            EventHandler("trawlover", function(inst) inst.AnimState:PlayAnimation("trawlover") end),
        },
    },

    State{
        name = "dolongaction",
        tags = {"doing", "busy", "boating", "repairing"},
        
        timeline=
        {
            TimeEvent(4*FRAMES, function( inst )
                inst.sg:RemoveStateTag("busy")
            end),
        },
        
        onenter = function(inst, timeout)
            --inst.components.driver:SplitFromVehicle()
            local targ = inst:GetBufferedAction() and inst:GetBufferedAction().target or nil
            if targ then targ:PushEvent("startlongaction") end
            
            inst.sg:SetTimeout(timeout or 1)
            inst.components.locomotor:Stop()
            inst.SoundEmitter:PlaySound("dontstarve/wilson/make_trap", "make")
            
            inst.AnimState:PlayAnimation("build_pre")
            inst.AnimState:PushAnimation("build_loop", true)
        end,
        
        ontimeout= function(inst)
            inst.AnimState:PlayAnimation("build_pst")
            inst.sg:GoToState("idle", false)
            inst:PerformBufferedAction()

            local vehicle = inst.components.driver.vehicle
            if vehicle and vehicle.components.boathealth:GetPercent() >= 0.8 then
                if inst:HasTag("repairmanprince") then
                    inst:RemoveTag("repairmanprince")
                end
            end

        end,
        
        onexit= function(inst)
            inst.SoundEmitter:KillSound("make")
        end,
    }, 

    State{
        name = "amulet_rebirth",
        tags = {"busy"},
        onenter = function(inst)
            inst.AnimState:PlayAnimation("amulet_rebirth")
            --inst.AnimState:OverrideSymbol("FX", "player_amulet_resurrect", "FX")
        end,
        
        onexit= function(inst)
            inst.components.hunger:SetPercent(2/3)
            inst.components.health:Respawn(TUNING.RESURRECT_HEALTH)
            
            if inst.components.sanity then
                inst.components.sanity:SetPercent(.5)
            end
            
            local item = inst.components.inventory:GetEquippedItem(EQUIPSLOTS.BODY)
            if item and item.prefab == "amulet" then
                item = inst.components.inventory:RemoveItem(item)
                if item then
                    item:Remove()
                    item.persists = false
                end
            end
            --SaveGameIndex:SaveCurrent()
            --inst.AnimState:ClearOverrideSymbol("FX")
            
        end,
        
        timeline =
        {
            TimeEvent(0*FRAMES, function(inst)
                inst.stafflight = SpawnPrefab("staff_castinglight")
                local pos = inst:GetPosition()
                local colour = {150/255, 46/255, 46/255}
                inst.stafflight.Transform:SetPosition(pos.x, pos.y, pos.z)
                inst.stafflight.setupfn(inst.stafflight, colour, 1.7, 1)           

            end),

            TimeEvent(0, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/rebirth_amulet_raise") end),
            TimeEvent(60*FRAMES, function(inst) inst.SoundEmitter:PlaySound("dontstarve/common/rebirth_amulet_poof") end),
        
            TimeEvent(80*FRAMES, function(inst)
                local pos = Vector3(inst.Transform:GetWorldPosition())
                local ents = TheSim:FindEntities(pos.x,pos.y,pos.z, 10)
                for k,v in pairs(ents) do
                    if v ~= inst and v.components.sleeper and v.prefab ~= "princess" then
                        v.components.sleeper:GoToSleep(20)
                    end
                end
                
                
            end),
        },        
                   
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },  

    State{
        name="item_hat",
        tags = {"idle"},
        
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            inst.AnimState:PlayAnimation("item_hat")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },    
    State{
        name="item_in",
        tags = {"idle"},
        
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            inst.AnimState:PlayAnimation("item_in")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },    
    State{
        name="item_out",
        tags = {"idle"},
        
        onenter = function(inst)
            inst.components.locomotor:StopMoving()
            inst.AnimState:PlayAnimation("item_out")
        end,
        
        events=
        {
            EventHandler("animover", function(inst) inst.sg:GoToState("idle") end ),
        },
    },      


}

return StateGraph("prince", states, events, "idle", actionhandlers)
